// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"

SgStatement *whichProcAmI(SgStatement *where)
{
    SgProgHedrStmt *fh;

    if (! where) { 
	return ((SgStatement *)0); 
    }
    
    if (where->variant() == GLOBAL) {
	return ((SgStatement *)0); 
    }
    
    if (where->variant() == PROG_HEDR ||
	where->variant() == PROC_HEDR ||
	where->variant() == PROS_HEDR ||
	where->variant() == FUNC_HEDR) {
	return (where);
    }

    return whichProcAmI(where->controlParent());
}

int lhsId(SgExpression *e){
    if(e->lhs()) return (e->lhs())->id();
    else return -1;
}
int rhsId(SgExpression *e){
    if(e->rhs()) return (e->rhs())->id();
    else return -1;
}
int symbId(SgExpression *e){
    if(e->symbol()) return (e->symbol())->id();
    else return -1;
}
char * symbName(SgExpression *e){
    if (isSgValueExp(e)) {
	return(e->unparse());
    } else {
      if(e->symbol()) 
	return (e->symbol())->identifier();
      else return "";
    }
}
int typeId(SgExpression *e){
    if(e->type()) return (e->type())->id();
    else return -1;
}

SgExpression& exprRef(SgExpression *e)
{ return (*e); }

char *ppString(char *prefix, char *name, char *postfix)
{
    char *theString;

    if (! prefix || ! name || ! postfix) {
	fm_internal_error("\
Expect non-null strings in ppString at %s:%d", __FILE__, __LINE__);
    }
    theString = (char *)malloc(strlen(prefix)+strlen(name)+strlen(postfix)+1);
    if (! theString) {
	fm_internal_error("\
Unable to allocate memory in ppString() at %s:%d", __FILE__, __LINE__);
    }
    sprintf(theString, "%s%s%s", prefix, name, postfix);
    return theString;
}


void declare_auto_symb(char *name, SgStatement *scope)
{
    SgSymbol *symb;

    symb = new SgVariableSymb(name, *auto_t);
    symb->declareTheSymbol(*scope);
}

void declare_symb_with_type(char *name, SgType *t, SgStatement *scope)
{
    SgSymbol *symb;

    symb = new SgVariableSymb(name, *t);
    symb->declareTheSymbol(*scope);
}

char *num2str(int n)
{
    char string[1024];

    sprintf(string, "%d", n);
    return (strdup(string));
}

void copyLabelFromOriginalStmt(SgStatement *origStmt, SgStatement *newStmt)
{
    if (origStmt->hasLabel()) {
	newStmt->setLabel(*origStmt->label());
    }
}

void doSaveStmt(SgStatement *s)
{
    /*
     * Give warning to the user that
     * save statement makes the subroutine
     * non-reentrant.  Also, remove the 
     * automatic specifiers from the 
     * variable declaration.
     *
     */

    SgVarListDeclStmt *save_stat;
    int i;

    save_stat = isSgVarListDeclStmt(s);

    fm_warning(s, "\
Save statement makes this subprogram a non-reentrant one and this cannot be used with threadable processes\n");

    for (i = 0; i < save_stat->numberOfVars(); i++) {
	SgExpression *var;

	var = save_stat->var(i);
	exprMapSet(staticVarMap, symbName(var), var);
    }
}    

void get_LS_stmts(SgStatement *scope)
{
    SgStatement *line;

    for (line = scope->lastDeclaration(); 
	 line != scope->lastNodeOfStmt(); line = line->lexNext())
    {
	if (line->variant() == RETURN_STAT ||
	    line->variant() == EXIT_STMT)
	    LS_stmts.append(line);
    }

    line = scope->lastNodeOfStmt()->lexPrev();

    if (line->variant() != RETURN_STAT ||
	line->variant() != EXIT_STMT)
    {
	SgStatement *last = scope->lastNodeOfStmt();
        LS_stmts.append(last);
    }
}

void insert_port_LS_calls(SgCallStmt *call)
{
    for (Pix ind = LS_stmts.first(); ind; LS_stmts.next(ind)) {
	SgStatement *s = LS_stmts(ind);

	s->insertStmtBefore(*call);
    }
}

void reset_LS_stmts(void)
{
    LS_stmts.clear();
}


	
