// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>

void doChannelStmt(SgStatement *s)
{
    SgChannelStmt            *channelStmt;
    SgExpression             *inport;
    SgExpression             *outport;
    SgExpression             *ioStore;
    SgExpression             *errLabel;
    SgCallStmt               *callStmt;
    SgExpression             *ioArg;
    SgDeclarationStatement   *extern_proc;
    SgExternalSymb           *extern_symb;
    SgVarRefExp              *extern_var;
    char                     *port_type_string;
    char                     *extern_full_func_name;
    char                     *extern_mg_func_name;
    SgValueExp               *extern_mg_func_exp;
    SgStatement              *scopeForDeclare;
    SgExpression             *inport_reference;
    SgExpression             *outport_reference;

    char                     *ident;

    channelStmt              = isSgChannelStmt(s);
    fm_assert(channelStmt);
    inport                   = channelStmt->inport();
    outport                  = channelStmt->outport();
    ioStore                  = channelStmt->ioStore();
    errLabel                 = channelStmt->errLabel();

    //
    // Consistency Checks
    //
    if (! inport) {
	fm_error(s, "missing inport specifier");
	return;
    }
    if (! outport) {
	fm_error(s, "missing outport specifier");
	return;
    }
    if (getPortVariant(inport) != INPORT_DECL) {
        fm_error(s, "invalid variable for in= specifier");
	return;
    }
    if (getPortVariant(outport) != OUTPORT_DECL) {
	fm_error(s, "invalid variable for out= specifier");
	return;
    }
    if (! eqPortTypeCiList(inport,outport)) {
	fm_error(s, "inport and outport are not the same type");
    }
    
    port_type_string = portSetKeyEncode(getPortStatement(outport));
    extern_full_func_name = ppString("fm_",port_type_string, "_init");
    extern_mg_func_exp = isSgValueExp(exprMapGet(portMangleMap, 
						 extern_full_func_name));
    if (! extern_mg_func_exp) {
	extern_mg_func_name = ppString(ppString("fm_",currentFileBase,""),
				       "_init", encodeStr(EC_MANGLE));
	exprMapSet(portMangleMap, extern_full_func_name,
		   new SgValueExp(extern_mg_func_name));
    } else {
	extern_mg_func_name = extern_mg_func_exp->stringValue();
    }
						 
    extern_symb = new SgExternalSymb(extern_mg_func_name,
				     *currentProcedure);

    extern_var  = new SgVarRefExp(*extern_symb);

    if (! stmtMapGet(prosMap, extern_mg_func_name)) {
	extern_proc = new SgDeclarationStatement(EXTERN_STAT);
	extern_proc->addVar(*extern_var);
	scopeForDeclare = currentProcedure->lastDeclaration();
	scopeForDeclare->insertStmtBefore(*extern_proc);
	stmtMapSet(prosMap, extern_mg_func_name,  extern_proc);
    }

    // 
    // generate fm_channel() calls 
    //
    inport_reference  = portExp(inport);
    outport_reference = portExp(outport);

    inport_reference  = EXPR_LIST_VALUE(inport_reference);
    outport_reference = EXPR_LIST_VALUE(outport_reference);
    
//    printf("inport_reference = %s\n", getVariantByName(inport_reference->variant()));
//    printf("outport_reference = %s\n", getVariantByName(outport_reference->variant()));

    if (EXPR_HAS_F90_ARRAY_SECTION(inport_reference) &&
	! EXPR_HAS_F90_ARRAY_SECTION(outport_reference) ||
	! EXPR_HAS_F90_ARRAY_SECTION(inport_reference) &&
	EXPR_HAS_F90_ARRAY_SECTION(outport_reference))
    {
	fm_error(s, "\
incompatible inport and outport shapes: inport=%s, outport=%s",
		 inport_reference->unparse(),
		 outport_reference->unparse());
    } else if (EXPR_HAS_F90_ARRAY_SECTION(inport_reference)) {
	SgContinueStmt *cont_statement;
	SgForStmt      *for_statement;
	SgSymbol       *loop_index_symbol;
	SgVarRefExp    *loop_ip_idx_ref, *loop_op_idx_ref;
	SgExpression   *new_inport_reference;
	SgExpression   *new_outport_reference;
	SgSubscriptExp *ip_dd_sub, *op_dd_sub;
	SgExpression   *ip_l_bound, *ip_u_bound, *ip_step;
	SgExpression   *op_l_bound, *op_u_bound, *op_step;
	int            ip_dd_i, op_dd_i;
	SgAssignStmt   *assign_statement;
	SgWhileStmt    *do_while_stmt;
	SgExpression   *cond_ip, *cond_op;
	//
	// both of them has f90 array section, i.e.
	//
	//     channel(in=ip(...,m_i:n_i,...),out=op(...,m_o:n_o,...))
	//
	//     generate a do loop
	//
	// 
	ip_dd_i   = get_DDOT_position(isSgArrayRefExp(inport_reference),1);
	ip_dd_sub = get_DDOT_in_aref(isSgArrayRefExp(inport_reference),
				     ip_dd_i);

	ip_l_bound = ip_dd_sub->lbound();
        ip_u_bound = ip_dd_sub->ubound();
	ip_step    = ip_dd_sub->step();
	
	op_dd_i   = get_DDOT_position(isSgArrayRefExp(outport_reference),1);
	op_dd_sub = get_DDOT_in_aref(isSgArrayRefExp(outport_reference),
				     op_dd_i);
	
	op_l_bound = op_dd_sub->lbound();
	op_u_bound = op_dd_sub->ubound();
	op_step    = op_dd_sub->step();

	if (! ip_l_bound) {
	    ip_l_bound = get_array_lbound(isSgArrayRefExp(inport_reference),
					  ip_dd_i);
	}
	if (! ip_u_bound) {
	    ip_u_bound = get_array_ubound(isSgArrayRefExp(outport_reference),
					  ip_dd_i);
	}
	if (! ip_step) {
	    ip_step = c1;
	}

	if (! op_l_bound) {
	    op_l_bound = get_array_lbound(isSgArrayRefExp(outport_reference),
					  op_dd_i);
	}
	if (! op_u_bound) {
	    op_u_bound = get_array_ubound(isSgArrayRefExp(outport_reference),
					  op_dd_i);
	}
	if (! op_step) {
	    op_step = c1;
	}

	//
	// generate loop initialization assignment
	// loop_ip_ref = ip_l_bound
	// loop_op_ref = op_l_bound
	//

	ident = encodeStr(EC_DOVAR);
//	declare_symb_with_type(ident, integer_t, currentProcedure);
	

	loop_index_symbol = new SgVariableSymb(ident, *integer_t);
	loop_index_symbol->declareTheSymbol(*currentProcedure);

	if (opt_use_auto_vars) {
	    declare_auto_symb(ident, currentProcedure);
	}

	loop_ip_idx_ref = new SgVarRefExp(*loop_index_symbol);

	ident = encodeStr(EC_DOVAR);

	loop_index_symbol = new SgVariableSymb(ident, *integer_t);

	loop_index_symbol->declareTheSymbol(*currentProcedure);

	if (opt_use_auto_vars) {
	    declare_auto_symb(ident, currentProcedure);
	}

	loop_op_idx_ref = new SgVarRefExp(*loop_index_symbol);

	assign_statement = new SgAssignStmt(*loop_ip_idx_ref, *ip_l_bound);
	s->insertStmtBefore(*assign_statement);
	assign_statement = new SgAssignStmt(*loop_op_idx_ref, *op_l_bound);
	s->insertStmtBefore(*assign_statement);

	cont_statement = new SgContinueStmt;
	
	cond_ip = &(*loop_ip_idx_ref < *ip_u_bound+*c1);
	cond_op = &(*loop_op_idx_ref < *op_u_bound+*c1);

	do_while_stmt = new SgWhileStmt( *cond_ip && *cond_op,
					  *cont_statement);
	
	s->insertStmtBefore(*do_while_stmt);
	
	//
	// generate fm_channel() call
	//
	new_inport_reference 
	  = replace_DDOT_with_index(isSgArrayRefExp(inport_reference),
				    ip_dd_i, *loop_ip_idx_ref);
	new_outport_reference
	  = replace_DDOT_with_index(isSgArrayRefExp(outport_reference),
				    ip_dd_i, *loop_op_idx_ref);
	    
	callStmt    = new SgCallStmt(*fmChannelCallSymb);
	fm_assert(callStmt);
	
	callStmt->addArg(*new_inport_reference);
	callStmt->addArg(*new_outport_reference);
	callStmt->addArg(*extern_var);
	ioArg = fmIOVarExp(ioStore);
	
	callStmt->addArg(*ioArg);
	cont_statement->insertStmtBefore(*callStmt);
	//
	// generate fm_message() call
	// 
	cont_statement->insertStmtBefore(*checkIOStat(1,ioStore,ioArg,errLabel));
	//
	// generate loop_index update statements
	//
	// loop_ip_idx_ref = loop_ip_idx_ref + ip_step
	// loop_op_idx_ref = loop_op_idx_ref + op_step
	//
	assign_statement = new SgAssignStmt(*loop_ip_idx_ref, 
					    *loop_ip_idx_ref + *ip_step);
	cont_statement->insertStmtBefore(*assign_statement);

	assign_statement = new SgAssignStmt(*loop_op_idx_ref, 
					    *loop_op_idx_ref + *op_step);
	cont_statement->insertStmtBefore(*assign_statement);

	cont_statement->deleteStmt();
    } else {
	callStmt    = new SgCallStmt(*fmChannelCallSymb);
	fm_assert(callStmt);
	
	callStmt->addArg(*portExp(inport));
	callStmt->addArg(*portExp(outport));
	callStmt->addArg(*extern_var);
	ioArg = fmIOVarExp(ioStore);
	
	callStmt->addArg(*ioArg);
	s->insertStmtBefore(*callStmt);
	s->insertStmtBefore(*checkIOStat(1,ioStore,ioArg,errLabel));
    }
    fmStmtSLL.append(s);
}

