// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <stdarg.h>

void print_statement(SgStatement *);

void fm_error(SgStatement *s, char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    fprintf(stderr, "\
%s:%d: ", 
	    currentFile,
	    s->lineNumber());
    print_statement(s);

    fprintf(stderr, "Error: ");
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    n_errors++;
}

void fm_warning(SgStatement *s, char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    fprintf(stderr, "\
%s:%d: ", 
	    currentFile,
	    s->lineNumber());
    print_statement(s);

    fprintf(stderr, "Warning: ");
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    n_warnings++;
}

void fm_internal_error(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    fprintf(stderr, "Internal error: ");
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    exit(1);
}

void print_hedr_statement(SgStatement *s)
{
    int i;

    fm_assert(s->symbol());
    if (s->variant() == PROG_HEDR) 
	printf("      program");
    else if (s->variant() == PROC_HEDR) 
        printf("      subroutine");
    else if (s->variant() == PROS_HEDR)
        printf("      process");
    else 
        printf("      function");

    printf(" %s(",  s->symbol()->identifier());
    
    if (s->variant() == PROG_HEDR) {
	SgProgHedrStmt *hedr;

	hedr           = isSgProgHedrStmt(s);
	for(i = 0; i < hedr->numberOfParameters(); i++) {
	    printf("%s", hedr->parameter(i)->identifier());
	    if (i == hedr->numberOfParameters()-1) {
		printf(")\n");
	    } else {
		printf(", ");
	    }
	}
    } else if (s->variant() == PROC_HEDR) {
	SgProcHedrStmt *hedr;

	hedr           = isSgProcHedrStmt(s);
	for(i = 0; i < hedr->numberOfParameters(); i++) {
	    printf("%s", hedr->parameter(i)->identifier());
	    if (i == hedr->numberOfParameters()-1) {
		printf(")\n");
	    } else {
		printf(", ");
	    }
	}
    } else if (s->variant() == PROS_HEDR) {
	SgProsHedrStmt *hedr;

	hedr           = isSgProsHedrStmt(s);
	for(i = 0; i < hedr->numberOfParameters(); i++) {
	    printf("%s", hedr->parameter(i)->identifier());
	    if (i == hedr->numberOfParameters()-1) {
		printf(")\n");
	    } else {
		printf(", ");
	    }
	}
    } else if (s->variant() == FUNC_HEDR) {
	SgFuncHedrStmt *hedr;

	hedr           = isSgFuncHedrStmt(s);
	for(i = 0; i < hedr->numberOfParameters(); i++) {
	    printf("%s", hedr->parameter(i)->identifier());
	    if (i == hedr->numberOfParameters()-1) {
		printf(")\n");
	    } else {
		printf(", ");
	    }
	}
    } else {
	fm_internal_error("\
Expect program, procedure, process or function header statement but get %s",
			  getVariantByName(s->variant()));
    }
}
   
/*
 * print_statement()
 *
 * Some statements print a whole body we only want 
 * to print a single line of the body in the error
 * messages.
 * 
 */

void print_statement(SgStatement *s)
{
    switch (s->variant()) {
      case PROG_HEDR:
      case PROC_HEDR:
      case PROS_HEDR:
      case FUNC_HEDR:
	print_hedr_statement(s);
	break;
      default:
	s->unparsestdout();
	break;
    }
}
