// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <math.h>
#include <assert.h>
#include <string.h>

#define EC_SEQUENCE_NUMBER_LENGTH            4

char *seqnumToString(int);

/*
 * encodeStr()
 *
 * create temparay variables names
 *
 */

char *encodeStr(int key)
{
    int ecStrLen = 3+strlen(currentProcedureName)+2+EC_SEQUENCE_NUMBER_LENGTH;

    char *ecStr    = (char *)malloc(ecStrLen+1);

    if (! ecStr) {
	fm_internal_error("\
Unable to allocate memory at %s:%d",  __FILE__, __LINE__);
    }

    // Encoded String = ProcedureName + prefix + Sequence Number
    switch (key) {
      case EC_IOVAL:
	sprintf(ecStr, "%s%s%s%s",
		FM_PREFIX,
		currentProcedureName, 
		EC_IOVAL_PREFIX, 
		seqnumToString(iovalSN));
	iovalSN++;
	break;
      case EC_EMPTY:
	sprintf(ecStr, "%s%s%s%s",
		FM_PREFIX,
		currentProcedureName, 
		EC_EMPTY_PREFIX, 
		seqnumToString(emptySN));
	emptySN++;
	break;
      case EC_CHANNEL:
	sprintf(ecStr, "%s%s%s%s",
		FM_PREFIX,
		currentProcedureName,
		EC_CHANNEL_PREFIX,
		seqnumToString(channelSN));
	channelSN++;
	break;
      case EC_PORT:
	sprintf(ecStr, "%s%s%s%s",
		FM_PREFIX,
		currentProcedureName,
		EC_PORT_PREFIX,
		seqnumToString(portSN));
	portSN++;
	break;
      case EC_PARBLOCK:
	sprintf(ecStr, "%s%s%s%s",
		FM_PREFIX,
		currentProcedureName,
		EC_PARBLOCK_PREFIX,
		seqnumToString(pblockSN));
	pblockSN++;
	break;
      case EC_BUFFER:
	sprintf(ecStr, "%s%s%s%s",
		FM_PREFIX,
		currentProcedureName,
		EC_BUFFER_PREFIX,
		seqnumToString(bufferSN));
	bufferSN++;
	break;
      case EC_DOVAR:
	sprintf(ecStr,"%s%s%s%s",
		FM_PREFIX,
		currentProcedureName,
		EC_DOVAR_PREFIX,
		seqnumToString(doVarSN));
	doVarSN++;
	break;
      case EC_VEC:
	sprintf(ecStr, "%s%s%s%s",
		FM_PREFIX,
		currentProcedureName,
		EC_VEC_PREFIX,
		seqnumToString(vecSN));
	vecSN++;
	break;
      case EC_MPC:
	sprintf(ecStr, "p%s",
		seqnumToString(mpcSN));
	mpcSN++;
	break;
      case EC_MANGLE:
	sprintf(ecStr, "%s", seqnumToString(mangleSN));
	mangleSN++;
	break;
      default:
	fm_internal_error("unknown EC_Key: %d", key);
    }
    return(ecStr);
}

char *seqnumToString(int n)
{
    char strBuffer[EC_SEQUENCE_NUMBER_LENGTH+2];
    char *strReturned = (char *)malloc(EC_SEQUENCE_NUMBER_LENGTH+1);
    char *Str;
    int i;

    if (! strReturned) {
	fm_internal_error("\
malloc failed at %s:%d", __FILE__, __LINE__);
    }

    // make sure that n could fit in strBuffer
    if (n > 0) {
	if (log10((double)n) >= EC_SEQUENCE_NUMBER_LENGTH) {
	    fm_internal_error("Cannot fit into strBuff, sequence too large");
	}
    }
    sprintf(strBuffer,"%d", n);
    Str = strBuffer + strlen(strBuffer) + 1;
    for (i = 0; i<EC_SEQUENCE_NUMBER_LENGTH-strlen(strBuffer); i++) {
	Str[i] = '0';
    }
    Str[i] = '\000';
    sprintf(strReturned, "%s%s", Str, strBuffer);
    return (strReturned);
}


/*
 * portTypeEncodeChar()
 *
 * return the character abbreviation which form the
 * the encoded port type string
 *
 */

char portTypeEncodeChar(SgType *t)
{
    char typeStr;

    switch (t->variant()) {
      case T_CHAR:
	typeStr = 'c';
	break;
      case T_STRING:
	typeStr = 's';
	break;
      case T_INT:
	typeStr = 'i';
	break;
      case T_SHORT:
	typeStr = 's';
	break;
      case T_BOOL:
	typeStr = 'l';
	break;
      case T_FLOAT:
	typeStr = 'f';
	break;
      case T_DOUBLE:
	typeStr = 'd';
	break;
      case T_COMPLEX:
	typeStr = 'x';
	break;
      default:
	fprintf(stderr, "Warning: % type not supported in port decl stmt\n",
		getVariantByName(t->variant()));
	typeStr = 0;
	break;
    }
    return (typeStr);
}

char * popTypeEncode(SgType *port_t, SgExpression *port_ref)
{
    if (port_t->variant() == T_INPORT) {
	SgInportStmt *pop_inport;
	SgPortTypeExp *pop_t = isSgPortTypeExp(port_ref);
	SgSymbol      *pop_symb = new SgVariableSymb("pop_port", 
						     *integer_t);
	SgVarRefExp   *pop_var = new SgVarRefExp(*pop_symb);
	SgExprListExp *pop_names = new SgExprListExp(*pop_var);

	fm_assert(pop_t);
	pop_inport = new SgInportStmt(*pop_names, *pop_t);
	
	return(portSetKeyEncode(pop_inport));
    } else if (port_t->variant() == T_OUTPORT) {
	SgOutportStmt *pop_outport;
	SgPortTypeExp *pop_t = isSgPortTypeExp(port_ref);
	SgSymbol      *pop_symb = new SgVariableSymb("pop_port", 
						     *integer_t);
	SgVarRefExp   *pop_var = new SgVarRefExp(*pop_symb);
	SgExprListExp *pop_names = new SgExprListExp(*pop_var);

	fm_assert(pop_t);
	pop_outport = new SgOutportStmt(*pop_names, *pop_t);
	
	return(portSetKeyEncode(pop_outport));
    } else {
	return("non_port_type");
    }
    return ("");
}



/*
 * portSetKeyEncode()
 *
 * return the encoded string of a port type in the port declaration 
 * statement.  Two different type of ports should have two different
 * encoded strings. 
 *
 */
    
char *portSetKeyEncode(SgStatement *s)
{
    //
    // key = filebase + [i|o] + "pt" + port type encoding
    //
    SgInportStmt  *inportStmt  = isSgInportStmt(s);
    SgOutportStmt *outportStmt = isSgOutportStmt(s);
    int ntypes = (inportStmt) ? inportStmt->numberOfPortTypes() : outportStmt->numberOfPortTypes();
//    int strLen  = strlen(currentFileBase) + 5 + 2*ntypes;
//    char *ECstr = (char *)malloc(strLen+1);
    char *cp;
    SgPortTypeExp *portType;
    SgType        *t;
    SgExpression  *ref;
    SgVarRefExp   *vRef;
    SgArrayRefExp *aRef;
    char *pt_prefix;

    char nameBuf[1024];

#if 0
    if (! ECstr) {
	fm_internal_error("\
malloc failure at %s:%d", __FILE__, __LINE__);
    }
#endif

    if (inportStmt) {
	pt_prefix = strdup("ipt");
    } else {
	pt_prefix = strdup("opt");
    }

    sprintf(nameBuf, "%s_%s_", currentFileBase, pt_prefix);
    cp = &nameBuf[strlen((const char*)nameBuf)];
    for (int i = 0; i < ntypes; i++) {
	portType = (inportStmt)? inportStmt->porttype(i) : outportStmt->porttype(i);
	t = portType->type();
	ref = portType->ref();
	if (! t) {
	    // we have a nested port over port
	    // case here, just treat it the same
	    // as T_INPORT and T_OUTPORT
	    if (portType->variant() == INPORT_TYPE_OP) {
		portType->setType(*inport_t);
	    } else if (portType->variant() == OUTPORT_TYPE_OP) {
		portType->setType(*outport_t);
	    }
	}

	t = portType->type();
	
	if (t->variant() == T_INPORT) {
	    *cp++ = 'p';
	    *cp++ = 'i';
	    *cp++ = '1';
//	    ref = 0;
	} else if (t->variant() == T_OUTPORT) {
	    *cp++ = 'p';
	    *cp++ = 'o';
	    *cp++ = '1';
//	    ref = 0;
	} else {
	    *cp++ = portTypeEncodeChar(t);
	}

	if (ref) {
	    if (vRef = isSgVarRefExp(ref)) {
//		*cp++ = 'v';
	    } else if (aRef = isSgArrayRefExp(ref)) {
		*cp++ = 'a';
	    } else if (t->variant() == T_INPORT)  {
		SgInportStmt  *pop_inport;
		SgPortTypeExp *pop_t;
		SgSymbol      *pop_symb;
		SgVarRefExp   *pop_var;
		SgExprListExp *pop_names;
		char *subStr;

		pop_t = isSgPortTypeExp(ref);
		pop_symb = new SgVariableSymb("pop_port", 
					      *integer_t);

		pop_var = new SgVarRefExp(*pop_symb);
		pop_names = new SgExprListExp(*pop_var);

		fm_assert(pop_t);
		pop_inport = new SgInportStmt(*pop_names, *pop_t);

		subStr = portSetKeyEncode(pop_inport);
	        strcpy(cp, subStr+strlen(currentFileBase)+4);
		free(subStr);
		cp += strlen(cp);
		delete pop_inport;
		delete pop_names;
		delete pop_var;
		delete pop_symb;
	    } else if (t->variant() == T_OUTPORT) {
		SgOutportStmt  *pop_outport;
		SgPortTypeExp *pop_t;
		SgSymbol      *pop_symb;
		SgVarRefExp   *pop_var;
		SgExprListExp *pop_names;
		char *subStr;

		pop_t = isSgPortTypeExp(ref);
		pop_symb = new SgVariableSymb("pop_port", 
					      *integer_t);

		pop_var = new SgVarRefExp(*pop_symb);
		pop_names = new SgExprListExp(*pop_var);

		fm_assert(pop_t);
		pop_outport = new SgOutportStmt(*pop_names, *pop_t);
		
		subStr = portSetKeyEncode(pop_outport);
		strcpy(cp, subStr+strlen(currentFileBase)+4);
		free(subStr);
		cp += strlen(cp);
		delete pop_outport;
		delete pop_names;
		delete pop_var;
		delete pop_symb;
	    } else {
		fm_error(s,"Illegal port type %s",
			 getVariantByName(ref->variant()));
	    }
	}
    }
    *cp = '\000';
    return(strdup(nameBuf));
}

/*
 * portDictKeyEncode()
 *
 * return the encoded string of the port name for accessing
 * port dictionary.  Each entry in a port dictionary is identified
 * by a encoded port variable name.
 *
 */ 

char *portDictKeyEncode(SgSymbol *symb)
{
    int strLen = strlen(currentProcedureName) 
      + strlen(currentProcedureName) + strlen(symb->identifier());
    char *ECstr = (char *)malloc(strLen+1);
    
    if (! ECstr) {
	fm_internal_error("\
malloc failure at %s:%d", __FILE__, __LINE__);
    }

    sprintf(ECstr, "%s%s%s", currentProcedureName, 
	    currentProcedureName, symb->identifier());
    return(ECstr);
}

/*
 * processorsKeyEncode()
 *
 * Return the encoded string for accessing the per-procedure Map
 * processorsMap
 *
 */

char *processorsKeyEncode()
{
    int strLen = strlen(currentProcedureName);
    char *s = (char *)malloc(strLen+1);
    
    if (! s) {
	fm_internal_error("\
malloc failure at %s:%d", __FILE__, __LINE__);
    }

    sprintf(s, "%s", currentProcedureName);
    return(s);
}

/*
 * prosMapKeyEncode()
 *
 * Return the encoded string for accessing the prosMap.
 *
 */

char *prosMapKeyEncode(char *prosName, char *varName)
{
    int ECStrLen = strlen(prosName) + 1 + strlen(varName);
    char *ECStr   = (char *)malloc(ECStrLen + 1);

    if (! ECStr) {
	fm_internal_error("\
unable to allocate memory in %s:%d", __FILE__, __LINE__);
    }
    sprintf(ECStr, "%s_%s", prosName, varName);
    return(ECStr);
}

/*
 * prosHedrMapKeyEncode()
 *
 * Return the encoded string as key to access prosHedrMap
 *
 */

char *prosHedrMapKeyEncode(char *prosName, char *varName)
{
    int ECStrLen = strlen(prosName) + 1 + strlen(varName);
    char *ECStr  = (char *)malloc(ECStrLen + 1);

    if (! ECStr) {
	fm_internal_error("\
unable to allocate memory in %s:%d", __FILE__, __LINE__);
    }
    sprintf(ECStr, "%s_%s", prosName, varName);
    return (ECStr);
}

/*
 * prosDefMapKeyEncode()
 * 
 * Return the encoded string as key for accessing prosDefMap
 *
 */

char *prosDefMapKeyEncode(char *prosName)
{
    int ECStrLen = strlen(currentFileBase) + 1 + strlen(prosName);
    char *ECStr  = (char *)malloc(ECStrLen + 1);

    if (! ECStr) {
	fm_internal_error("\
unable to allocate memory in %s:%d", __FILE__, __LINE__);
    }
    sprintf(ECStr, "%s_%s", currentFileBase, prosName);
    return (ECStr);
}

/* 
 * parameterMapKeyEncode()
 *
 * Return the encoded string as key for accessing parameter map
 *
 */

char *parameterMapKeyEncode(SgSymbol *symb)
{
    int ECStrLen = strlen(currentFile) 
      + strlen(currentProcedureName) 
      + strlen(symb->identifier());
    char *ECStr = (char *)malloc(ECStrLen+1);

    if (! ECStr) {
	fm_internal_error("\
fail to allocate space in %s:%d", __FILE__, __LINE__);
    }

    sprintf(ECStr, "%s%s%s", 
	    currentFile, 
	    currentProcedureName, 
	    symb->identifier());

    return (ECStr);
}

char *intentMapKeyEncode(char *str)
{
    return(str);
}

char *directiveMapKeyEncode(char *prosName, char *directive_name)
{
    return ppString(prosName,"_", directive_name);
}

char *commonMapKeyEncode(char *var_name)
{
    return ppString("common", "_", var_name);
}
