// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"

/*
 * arraySizeExp_r()
 * 
 * Return the size of the indicated array
 *   SgExpression &arrayLenExp = arraySizeExp(1,a->numberOfSubscripts(),a);
 * Note: _r stands for recursion. This is called by arraySizeExp().
 *
 */

SgExpression& arraySizeExp_r(int c, int n, SgArrayRefExp *a)
{
    SgExpression *e = a->subscript(c);
    if (c == n-1) {
	return ( (*e) );
    } else {
	return( (*e) * arraySizeExp_r(c+1,n,a) );
    }
}

/*
 * arraySizeExp()
 *
 * Return the size of the indicated array
 *
 */

SgExpression& arraySizeExp(SgArrayRefExp *a)
{
    SgExpression *array_size_e;

    return(arraySizeExp_r(0,a->numberOfSubscripts(),a));
}

/*
 * portExp()
 *
 * Return the port reference buried in the port expressions
 * in send, receive, probe, moveport, channel, merger and 
 * endchannel statements
 *
 */

SgExpression *portExp(SgExpression *e)
{
    SgDataImpliedDoExp *dataimpldo;
    SgExpression *ptExp;
    
    switch(NODE_CODE(e->thellnd))
    {
      case INPORT_NAME:
      {
	SgInportExp *in = isSgInportExp(e);
	ptExp = in->exp();
	break;
      }
      case OUTPORT_NAME:
      {
	SgOutportExp *out = isSgOutportExp(e);
	ptExp = out->exp();
	break;
      }
      case FROMPORT_NAME:
      {
	SgFromportExp *from = isSgFromportExp(e);
	ptExp = from->exp();
	break;
      }
      case TOPORT_NAME:
      {
	SgToportExp *to = isSgToportExp(e);
	ptExp = to->exp();
	break;
      }
      case VAR_REF:
      case ARRAY_REF:
	return (e);
      case EXPR_LIST:
	return(isSgExprListExp(e)->value());
      default:
	printf("Huh? what is: %d\n", NODE_CODE(e->thellnd));
	return (SgExpression *)0;
	break;
    }
    return (ptExp);
}

int is_possible_port_expression(SgExpression *e)
{
    SgDataImpliedDoExp *dataimpldo;
    SgExpression *ptExp;
    
    switch(NODE_CODE(e->thellnd))
    {
      case INPORT_NAME:
      case OUTPORT_NAME:
      case FROMPORT_NAME:
      case TOPORT_NAME:
      case VAR_REF:
      case ARRAY_REF:
      case EXPR_LIST:
        return 1;
	break;
      default:
	return 0;
	break;
    }
    //
    // impossible to get here: 
    //   only to avoid compiler complaints.
    //
    return (0);
}

/*
 * ioStoreExp()
 * 
 * Return the IOSTAT reference buried in iostore expressions
 * in send, receive, probe, moveport, channel, merger and 
 * endchannel statements.
 *
 */
 
SgExpression *ioStoreExp(SgExpression *e)
{
    SgExpression *ioExp;
    if (! e) {
	return ((SgExpression *) 0);
    }

    switch(NODE_CODE(e->thellnd))
    {
      case IOSTAT_STORE:
      {
	SgIO_statStoreExp *io = isSgIO_statStoreExp(e);
	ioExp = io->exp();
	break;
      }
      case EMPTY_STORE:
      {
	SgEmptyStoreExp *empty = isSgEmptyStoreExp(e);
	ioExp = empty->exp();
	break;
      }
      default:
	printf("Huh, what is: %d\n", NODE_CODE(e->thellnd));
    }
    return (ioExp);
}

/*
 * labelExp()
 * 
 * Return the label reference buried in the label expressions 
 * in send, receive, probe, moveport, channel, merger and 
 * endchannel statements.
 *
 */

SgExpression *labelExp(SgExpression *e)
{
    SgExpression *lbExp;
    if (! e) {
	return ((SgExpression *)0);
    }

    switch(NODE_CODE(e->thellnd))
    {
      case ERR_LABEL:
      {
	SgErrLabelExp *err = isSgErrLabelExp(e);
	lbExp = err->exp();
	break;
      }
      case END_LABEL:
      {
	SgEndLabelExp *end = isSgEndLabelExp(e);
	lbExp = end->exp();
	break;
      }
      case EMPTY_STORE:
      {
        SgEmptyStoreExp *empty = isSgEmptyStoreExp(e);
	SgExpression *emptyVar = empty->exp();
        char            *emptyStr;
	int         labelVal;
        SgLabel *label;

	fm_assert(emptyVar);

//	emptyStr = (empty->exp())->unparse();

	switch  (emptyVar->variant()) {
	  case INT_VAL:
	    labelVal = isSgValueExp(emptyVar)->intValue();
	    break;
	  case FLOAT_VAL:
	    labelVal = atoi(isSgValueExp(emptyVar)->floatValue());
	    break;
	  case EXPR_LIST:
	    return isSgExprListExp(emptyVar)->elem(0);
	    break;
	  default:
	    fprintf(stderr,"Invalid value for empty store: %s\n",
		    getVariantByName(emptyVar->variant()));
	}
	label = new SgLabel(labelVal);
        lbExp = new SgLabelRefExp(*label);
        break;
      }
      case LABEL_REF:
        return (e);
        break;
      default:
	printf("Huh, what is: %d\n", NODE_CODE(e->thellnd));
	lbExp = 0;
    }
    return(lbExp);
}

/*
 * replaceConstantInExpr()
 *
 * Replace all the occurances of the constant symbols (parameter names)
 * by their defined values
 *
 */

SgExpression &replaceConstantInExpr(SgExpression *e)
{
    SgExpression *constVal;

    switch (e->variant()) {
      case MINUS_OP:
	/* unary operations */
	return (replaceConstantInExpr(e->lhs()));
	break;
      case ADD_OP:
	return (replaceConstantInExpr(e->lhs()) 
		+ replaceConstantInExpr(e->rhs()));
	break;
      case SUBT_OP:
	return (replaceConstantInExpr(e->lhs()) 
		- replaceConstantInExpr(e->rhs()));
	break;
      case MULT_OP:
	return (replaceConstantInExpr(e->lhs()) 
		* replaceConstantInExpr(e->rhs()));
	break;
      case DIV_OP:
	return (replaceConstantInExpr(e->lhs()) 
		/ replaceConstantInExpr(e->rhs()));
	break;
      case CONST_REF:
	if (constVal = parameterMapGet(parameterMap, symbName(e))) {
	    return(replaceConstantInExpr(constVal));
	} 
	break;
//      case STAR_RANGE:
//	return (*c1);
//	break;
      default:
	break;
    }
    return(*e);
}

/*
 * IsInteger()
 * 
 * Return 1 when the indicated expression evaluates to an integer
 * 0 otherwise.
 *
 */

int IsInteger(SgExpression *e)
{
    SgExpression &e_1 = replaceConstantInExpr(e);
    
    if (e_1.isInteger()) {
	return(1);
    } else {
	return 0;
    }
}

/*
 * integerValueExp()
 * 
 * Returns an Integer Value Expression if E evaluates to an 
 * integer; a null value expression  otherwise.
 *
 */

SgExpression * integerValueExp(SgExpression *e)
{
    SgExpression &e_1 = replaceConstantInExpr(e);
    
    if (e_1.isInteger()) {
	SgValueExp *val = new SgValueExp(e_1.valueInteger());

	return val;
    } else {
	return (SgExpression *)0;
    }
}

/*
 * integerValue(SgExpression *)
 *
 * Return the integer Value of a Expression.
 * Be Sure to use the test IsInteger() before calling
 * this function.  Otherwise you would not get a valid 
 * value.
 *
 */

int integerValue(SgExpression *e)
{
    SgExpression &e_1 = replaceConstantInExpr(e);
    
    if (e_1.isInteger()) {
	return e_1.valueInteger();
    } else {
	fm_internal_error("\
Expression %s cannot evaluates to integer, should call IsInteger() first!",
			  e->unparse());
	return (0);
    }
}

/*
 * EPlusOne()
 *
 * Return an expression which is the increment of
 * the given expression.
 *
 */
    
SgExpression& EPlusOne(SgExpression &e)
{
    return(e + *c1);
}

/*
 * EAdd()
 *
 * Return an expression which is the sum of two 
 * given expressions
 *
 */
SgExpression& EAdd(SgExpression &e1, SgExpression &e2)
{
    SgExpression *e = &(e1 + e2);
    if (e->isInteger()) {
	SgValueExp *val = new SgValueExp(e->valueInteger());
	
	return(*val);
    } else {
	return( e1 + e2 );
    }
}

/*
 * EMul()
 * 
 * Given two expressions return the muliplication of the two.
 *
 */

SgExpression& EMul(SgExpression &e1, SgExpression &e2)
{
    return ( e1 * e2 );
}

/*
 * loopSize()
 *
 * Return a size of the iteration space for a do loop
 * Note: no recursion is involved so this only returns _ONE_
 *       loop size.
 */

SgExpression& loopSize(SgExpression *start, 
		       SgExpression *end, 
		       SgExpression *step)
{
    return ((step)?((*end)-(*start)+(*c1))/(*step) : (*end)-(*start)+(*c1));
}

/*
 * replacePortProtoArgInExpr(SgExpression *e)
 *
 * Return an expression which is the same expression as E except
 * that the port type proto arg is replaced by the real argument.
 *
 */

SgExpression &replacePortProtoArgInExpr(SgExpression *e)
{
    SgVarRefExp      *protoArg;
    SgExpression     *realArg;
    char             *varName;

    switch (e->variant()) {
      case MINUS_OP:
	/* unary operations */
	return (replacePortProtoArgInExpr(e->lhs()));
	break;
      case ADD_OP:
	return (replacePortProtoArgInExpr(e->lhs()) 
		+ replacePortProtoArgInExpr(e->rhs()));
	break;
      case SUBT_OP:
	return (replacePortProtoArgInExpr(e->lhs()) 
		- replacePortProtoArgInExpr(e->rhs()));
	break;
      case MULT_OP:
	return (replacePortProtoArgInExpr(e->lhs()) 
		* replacePortProtoArgInExpr(e->rhs()));
	break;
      case DIV_OP:
	return (replacePortProtoArgInExpr(e->lhs()) 
		/ replacePortProtoArgInExpr(e->rhs()));
	break;
      case VAR_REF:
	protoArg     = isSgVarRefExp(e);
	realArg      = portTypeMapGet(portTypeMap, symbName(protoArg));
	
	if (realArg) {
	    return(*realArg);
	} else {
	    return(*protoArg);
	}
	break;
      default:
	break;
    }
    return(*e);
}
    

SgExpression& replaceFormalArgConstInExpr(SgExpression *e)
{
    SgVarRefExp      *protoArg;
    SgExpression     *realArg;
    char             *varName;
    SgExpression     *constVal;

    switch (e->variant()) {
      case MINUS_OP:
	/* unary operations */
	return (replaceFormalArgConstInExpr(e->lhs()));
	break;
      case ADD_OP:
	return (replaceFormalArgConstInExpr(e->lhs()) 
		+ replaceFormalArgConstInExpr(e->rhs()));
	break;
      case SUBT_OP:
	return (replaceFormalArgConstInExpr(e->lhs()) 
		- replaceFormalArgConstInExpr(e->rhs()));
	break;
      case MULT_OP:
	return (replaceFormalArgConstInExpr(e->lhs()) 
		* replaceFormalArgConstInExpr(e->rhs()));
	break;
      case DIV_OP:
	return (replaceFormalArgConstInExpr(e->lhs()) 
		/ replaceFormalArgConstInExpr(e->rhs()));
	break;
      case VAR_REF:
	protoArg     = isSgVarRefExp(e);
	realArg      = exprMapGet(prosHedrMap, 
				  prosHedrMapKeyEncode(currentProcedureName,symbName(protoArg)));
	if (realArg) {
	    return(*realArg);
	} else {
	    return(*protoArg);
	}
	break;
      case CONST_REF:
	if (constVal = parameterMapGet(parameterMap,symbName(e))) {
	    return(*constVal);
	} 
	break;
      default:
	break;
    }
    return(*e);
}

void printExprByDeref(FILE *fp, SgExpression *e)
{
    SgVarRefExp      *protoArg;
    SgExpression     *realArg;
    char             *varName;
    SgExpression     *constVal;

    switch (e->variant()) {
      case MINUS_OP:
	/* unary operations */
	fprintf(fp, "-(");
	printExprByDeref(fp, e->lhs());
	fprintf(fp,")");
	break;
      case ADD_OP:
	fprintf(fp, "(");
	printExprByDeref(fp, e->lhs());
	fprintf(fp,")+(");
	printExprByDeref(fp, e->rhs());
	fprintf(fp, ")");
	break;
      case SUBT_OP:
	fprintf(fp, "(");
	printExprByDeref(fp, e->lhs());
	fprintf(fp,")-(");
	printExprByDeref(fp, e->rhs());
	fprintf(fp, ")");
	break;
      case MULT_OP:
	fprintf(fp, "(");
	printExprByDeref(fp, e->lhs());
	fprintf(fp,")*(");
	printExprByDeref(fp, e->rhs());
	fprintf(fp, ")");
	break;
      case DIV_OP:
	fprintf(fp, "(");
	printExprByDeref(fp, e->lhs());
	fprintf(fp,")/(");
	printExprByDeref(fp, e->rhs());
	fprintf(fp, ")");
	break;
      case VAR_REF:
	fprintf(fp, "*%s", e->symbol()->identifier());
	break;
      default:
	fprintf(fp, "%s", e->unparse());
	break;
    }
}
    
int is_f90_array_section_reference(SgExpression *e)
{
//    printf("port variant = %s\n", getVariantByName(e->variant()));
    if (e->variant() != ARRAY_REF) {
	return 0;
    } else {
	SgArrayRefExp *a;

	a             = isSgArrayRefExp(e);
	for (int i = 0; i < a->numberOfSubscripts(); i++) {
	    if (a->subscript(i)->variant() == DDOT) {
		return 1;
	    } 
	}
    }
    return 0;
}

/*
 * replace_array_section_with_index()
 *
 * Return a new array reference which is the same
 * as the array A except that the I_th index is
 * replaced by the given INDEX
 *
 */

SgExpression *replace_DDOT_with_index(SgArrayRefExp *array,
					       int I,
					       SgExpression &index)
{
    SgArrayRefExp   *new_array;
    SgExpression    *var;
    SgArrayRefExp   *a;
    SgArrayType     *array_t;
    SgSymbol        *array_symbol;
    SgExpression    *init, *limit, *increment;
    
    fm_assert(array);
    var             = findSymbInDeclStmt(array->symbol());
    if (var->variant() != ARRAY_REF) {
	return (SgExpression *)0;
    } 
    a               = isSgArrayRefExp(var);
//    array_t         = isSgArrayType(a->type());
//    fm_assert(array_t);
    array_t         = new SgArrayType(*integer_t);
    for (int sub_i = 0; sub_i < a->numberOfSubscripts(); sub_i++) {
	array_t->addRange(*a->subscript(sub_i));
    }

    array_symbol    = new SgVariableSymb(symbName(a), *array_t);
    new_array       = new SgArrayRefExp(*array_symbol);
    
    // 
    // find the i_th occurence of f90 array sections
    //
    //
    for (int dd_i = 0; dd_i < a->numberOfSubscripts(); dd_i++) {
	SgExpression *sub = array->subscript(dd_i);

	if (sub->variant() == DDOT && dd_i == I){
	    SgSubscriptExp *sub_exp;
	    sub_exp        = isSgSubscriptExp(sub);
	    init           = sub_exp->lbound();
	    limit          = sub_exp->ubound();
	    increment      = sub_exp->step();
	    new_array->addSubscript(index);
	} else {
	    new_array->addSubscript(*sub);
	}
    }
    return (new_array);
}

int get_DDOT_position(SgArrayRefExp *a, int i_th_occurence)
{
    int  occurence;

    fm_assert(a);
    occurence = 0;

    for (int dd_i = 0; dd_i < a->numberOfSubscripts(); dd_i++) {
	SgExpression *sub;

	sub          = a->subscript(dd_i);
	if (sub->variant() == DDOT && occurence == i_th_occurence-1) {
	    return(dd_i);
	} else if (sub->variant() == DDOT && occurence != i_th_occurence-1) {
	    occurence++;
	}
    }
    return -1;
}

SgSubscriptExp *get_DDOT_in_aref(SgArrayRefExp *a, int I)
{
    int  occurence;

    fm_assert(a);
    
    if (I < 0 || I >= a->numberOfSubscripts()) {
	fm_internal_error("\
get_DDOT_in_aref(), index out of range");
    }
    return(isSgSubscriptExp(a->subscript(I)));
}

SgExpression * get_array_lbound(SgArrayRefExp *a, int I)
{
    fm_assert(a);
    
    return (c1);
}

SgExpression * get_array_ubound(SgArrayRefExp *a, int I)
{
    SgArrayType  *array_t;
    SgSymbol     *symb = new SgVariableSymb(symbName(a));
    SgArrayRefExp *a_decl;

    fm_assert(a);
    
    a_decl = isSgArrayRefExp(findSymbInDeclStmt(symb));
    if (! a_decl) {
	fm_internal_error("\
Expected an array reference for variable %s",
			  symbName(a));
    }
    if (I > a_decl->numberOfSubscripts()) {
	fm_internal_error("\
Subscript out of range (%d) array %s only has %d subscripts",
			  I, symbName(a), a_decl->numberOfSubscripts());
    }
    return(a_decl->subscript(I));
}

SgArrayType *make_array_t(SgType *base_t,
			    SgExpression *length, 
			    SgArrayRefExp *a)
{
    SgArrayType *array_t = new SgArrayType(*base_t);

    array_t->addRange(*length);
    if (a) {
	for (int i = 0; i < a->numberOfSubscripts(); i++) {
	    array_t->addRange(*a->subscript(i));
	}
    }
    return (array_t);
}

SgArrayRefExp *make_array_ref(SgType *base_t,
			      SgExpression *length, 
			      SgExpression *var,
			      SgArrayRefExp *a)
{
    SgArrayType   *array_t;
    SgSymbol      *array_symb;
    SgArrayRefExp *array_ref;

    array_t       = make_array_t(base_t, length, a);
    array_symb    = new SgVariableSymb(symbName(var), *array_t);
    array_ref     = new SgArrayRefExp(*array_symb);

    array_ref->addSubscript(*length);
    if (a) {
	for (int i = 0; i < a->numberOfSubscripts(); i++) {
	    array_ref->addSubscript(*a->subscript(i));
	}
    }
    return (array_ref);
}

