// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#define GLOBAL_VARIABLES_ALLOCATED_HERE

#include "fmc_defs.h"
#include "global.h"

void initVariantNames(void);

StringSgStatementAVLMap *newStmtMap()
{
    SgStatement *mapInitialContents = new SgStatement(INPORT_DECL);
    StringSgStatementAVLMap *map;

    map = new StringSgStatementAVLMap(*mapInitialContents);
    return (map);
}

StringSgExpressionAVLMap *newExprMap()
{
    StringSgExpressionAVLMap *map;

    map = new StringSgExpressionAVLMap(*c0);
    return(map);
}


StringFMProgBlockAVLMap *newfmProgBlockMap()
{
    FMProgBlock *mapInitialContents = new FMProgBlock;
    StringFMProgBlockAVLMap *map;

    map = new StringFMProgBlockAVLMap(*mapInitialContents);
    return(map);
}

    
FMProgBlock *newFMProgBlock()
{
    FMProgBlock *fm_prog_block;

    portVarDict          = newStmtMap();
    portTypeMap          = newExprMap();
    prosMap              = newStmtMap();
    prosDefMap           = newStmtMap();
    prosHedrMap          = newExprMap();
    intentMap            = newExprMap();
    processorsMap        = newExprMap();
    parameterMap         = newExprMap();
    staticVarMap         = newExprMap();
    typeMap              = newExprMap();
    distributeMap        = newExprMap();
    mpcallMap            = newStmtMap();

    fm_prog_block        = new FMProgBlock;

    if (! portVarDict 
        || ! portTypeMap
        || ! prosMap
        || ! prosDefMap
        || ! prosHedrMap
        || ! intentMap
        || ! processorsMap
        || ! parameterMap
	|| ! staticVarMap
        || ! typeMap     
	|| ! distributeMap
	|| ! mpcallMap
        || ! fm_prog_block )
    {
        fm_internal_error("\
Unable to create a new FM program block Map at %s:%d",
                          __FILE__, __LINE__);
    }

    fm_prog_block->setFileName(currentFile);
    fm_prog_block->setFileBaseName(currentFileBase);
    fm_prog_block->setProc(currentProcedure);
    fm_prog_block->setProcName(currentProcedureName);
    fm_prog_block->setPortVarDict(portVarDict);
    fm_prog_block->setPortTypeMap(portTypeMap);
    fm_prog_block->setProsMap(prosMap);
    fm_prog_block->setProsDefMap(prosDefMap);
    fm_prog_block->setProsHedrMap(prosHedrMap);
    fm_prog_block->setIntentMap(intentMap);
    fm_prog_block->setProcessorsMap(processorsMap);
    fm_prog_block->setParameterMap(parameterMap);
    fm_prog_block->setStaticVarMap(staticVarMap);
    fm_prog_block->setTypeMap(typeMap);
    fm_prog_block->setDistributeMap(distributeMap);
    fm_prog_block->setMPcallMap(mpcallMap);

    return(fm_prog_block);
}

void switchToFMProgBlock(char *procName)
{
    FMProgBlock *fm_prog_block_p = fmProgBlockMapGet(fmProgBlockMap, procName);

    if (! fm_prog_block_p) {
        fm_internal_error("\
program/procedure/process %s does not exist in program block map at %s:%d",
                          procName, __FILE__, __LINE__);
    }
    
    currentFile          = fm_prog_block_p->fileName();
    currentFileBase      = fm_prog_block_p->fileBaseName();
    currentProcedure     = fm_prog_block_p->proc();
    currentProcedureName = fm_prog_block_p->procName();
    portVarDict          = fm_prog_block_p->portVarDict();
    portTypeMap          = fm_prog_block_p->portTypeMap();
    prosMap              = fm_prog_block_p->prosMap();
    prosHedrMap          = fm_prog_block_p->prosHedrMap();
    intentMap            = fm_prog_block_p->intentMap();
    processorsMap        = fm_prog_block_p->processorsMap();
    parameterMap         = fm_prog_block_p->parameterMap();
    staticVarMap         = fm_prog_block_p->staticVarMap();
    typeMap              = fm_prog_block_p->typeMap();
    distributeMap        = fm_prog_block_p->distributeMap();
    mpcallMap            = fm_prog_block_p->mpcallMap();
}


void fm_initialize()
{
    SgStatement *portMapInitialContents;
    FMProgBlock *fmProgBlkInitialContents = new FMProgBlock;

    fmProgBlockMap = new StringFMProgBlockAVLMap(*fmProgBlkInitialContents);
    // current file and procedure pointer

    currentFile                  = (char *)0;
    currentFileBase              = (char *)0;
    currentProcedure             = (SgStatement *)0;

    // constants
    c0                     = new SgValueExp(0);      
    c1                     = new SgValueExp(1);
    c_minus_1              = new SgValueExp(-1);

    // global maps
    portMapInitialContents = new SgStatement(INPORT_DECL);
    portSet                = new StringSgStatementAVLMap(*portMapInitialContents);

    mpBlock                = newStmtMap();
    portMangleMap          = newExprMap();

    // type constant

    integer_t              = new SgType(T_INT);
    bool_t                 = new SgType(T_BOOL);
    char_t                 = new SgType(T_CHAR);
    real_t                 = new SgType(T_FLOAT);
    double_t               = new SgType(T_DOUBLE);
    complex_t              = new SgType(T_COMPLEX);

    auto_t                 = new SgType(T_AUTO);

    inport_t               = new SgType(T_INPORT);
    outport_t              = new SgType(T_OUTPORT);

    // type expression

    integer_e              = new SgTypeExp(*integer_t);
    char_e                 = new SgTypeExp(*char_t);
    real_e                 = new SgTypeExp(*real_t);
    double_e               = new SgTypeExp(*double_t);
    complex_e              = new SgTypeExp(*complex_t);

    // attributes

    in_attribute           = new SgAttributeExp(IN_OP);
    out_attribute          = new SgAttributeExp(OUT_OP);
    inout_attribute        = new SgAttributeExp(INOUT_OP);
    
    // compiler generated symbols

    fmNProcessorsCallSymb  = new SgSymbol(FUNCTION_NAME, FM_N_PROCESSORS_PROC);
    fmNPIntrinsicSymb      = new SgSymbol(FUNCTION_NAME, "number_of_processors");
    fmInitCallSymb         = new SgSymbol(FUNCTION_NAME,FM_INIT_PROC);
    fmIsMasterSymb         = new SgVariableSymb(FM_IS_MASTER_VAR, 
						*integer_t);

    fmStartCallSymb        = new SgSymbol(PROCEDURE_NAME,FM_START_PROC);
    fmShutDownCallSymb     = new SgSymbol(PROCEDURE_NAME,FM_SHUTDOWN_PROC);

    fmVCShapeCallSymb      = new SgSymbol(PROCEDURE_NAME,FM_VC_SHAPE_PROC);
    fmChildVCOffsetCallSymb= new SgSymbol(PROCEDURE_NAME,FM_CHILD_VC_OFFSET_PROC);

    fmInitInportCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_INIT_INPORT_PROC);
    fmInitInportDebugCallSymb   
      = new SgSymbol(PROCEDURE_NAME,FM_INIT_INPORT_DEBUG_PROC);
    fmInportLeavingScopeCallSymb
      = new SgSymbol(PROCEDURE_NAME, FM_INPORT_LEAVING_SCOPE_PROC);

    fmInitOutportCallSymb  = new SgSymbol(PROCEDURE_NAME,FM_INIT_OUTPORT_PROC);
    fmInitOutportDebugCallSymb  
      = new SgSymbol(PROCEDURE_NAME,FM_INIT_OUTPORT_DEBUG_PROC);
    fmOutportLeavingScopeCallSymb
      = new SgSymbol(PROCEDURE_NAME, FM_OUTPORT_LEAVING_SCOPE_PROC);

    fmChannelCallSymb      = new SgSymbol(PROCEDURE_NAME,FM_CHANNEL_PROC);
    fmMoveInportCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_MOVE_INPORT_PROC);
    fmMoveOutportCallSymb  = new SgSymbol(PROCEDURE_NAME,FM_MOVE_OUTPORT_PROC);
    fmErrorMessageCallSymb = new SgSymbol(PROCEDURE_NAME,FM_ERROR_MESSAGE_PROC);
    fmMergerInitCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_MERGER_INIT_PROC);
    fmMergerAddCallSymb    = new SgSymbol(PROCEDURE_NAME,FM_MERGER_ADD_PROC);
    fmMergerDoneCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_MERGER_DONE_PROC);
    fmEndChannelCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_END_CHANNEL_PROC);
    fmProbeCallSymb        = new SgSymbol(PROCEDURE_NAME,FM_PROBE_PROC);

    fmPblockInitCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_PBLOCK_INIT_PROC);
    fmPblockSpawnCallSymb  = new SgSymbol(PROCEDURE_NAME,FM_PBLOCK_SPAWN_PROC);
    fmPblockResetCallSymb  = new SgSymbol(PROCEDURE_NAME,FM_PBLOCK_RESET_PROC);
    fmPblockInitInvokeCallSymb
                           = new SgSymbol(PROCEDURE_NAME,FM_PBLOCK_INIT_INVOKE_PROC);
    
    fmPblockInvokeCallSymb = new SgSymbol(PROCEDURE_NAME,FM_PBLOCK_INVOKE_PROC);
    fmPblockContextWaitCallSymb
                           = new SgSymbol(PROCEDURE_NAME,FM_PBLOCK_CONTEXT_WAIT_PROC);
    
    fmPortCreateCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_PORT_CREATE_PROC);
    fmNotifyPortsCallSymb  = new SgSymbol(PROCEDURE_NAME,FM_NOTIFY_PORTS_PROC);
    fmPblockWaitCallSymb   = new SgSymbol(PROCEDURE_NAME,FM_PBLOCK_WAIT_PROC);

    // Pablo Caller Info
    
    fmPabloCallerInfoSymb  = new SgSymbol(PROCEDURE_NAME, FM_PABLO_CALLER_INFO_PROC);

    // Control flow

    fmStopCallSymb         = new SgSymbol(PROCEDURE_NAME,FM_STOP_PROC);

    // Regular Expressions for Compiler Directives

    threadableDirective    = new Regex("^CFM\\$THREADABLEPROCESS");

    fmVecNoAnnValRef       = new SgValueExp(0);
    fmVecLctnValRef        = new SgValueExp(1);
    fmVecSubmValRef        = new SgValueExp(2);

    // sequence number variables for compiler generated names
    iovalSN                = 0;
    channelSN              = 0;
    portSN                 = 0;
    pblockSN               = 0;
    bufferSN               = 0;
    doVarSN                = 0;
    vecSN                  = 0;
    mpcSN                  = 0;
    emptySN                = 0;

    // initialize error and warning counters
    n_errors               = 0;
    n_warnings             = 0;
    max_mp_processors      = 0;

    // initialize the variant table
    initVariantNames();

}
