// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#ifdef USE_LIBG_GETOPT
#include <GetOpt.h>
#endif /* USE_LIBG_GETOPT */
#include <String.h>
#include <string.h>
#include <ctype.h>

/*
 * In unistd.h of GCC 2.5.8 the getopt() function proto
 * is missing.  I need to declare it so that I don't
 * stupid warnings
 *
 */
#if defined(__GNUC__) && defined(__GNUC_MINOR__)
#if __GNUC__ == 2 && __GNUC_MINOR__ == 5
extern "C" int getopt(int, char**, const char*);
#endif 
#endif /* __GNUC__ && __GNUC_MINOR __ */

static void           usage(char *);
char *                file_name_to_ident(const char *);
void                  do_compiler_directives(char *, SgStatement *);
int                   port_t_has_ports(SgStatement *);
int                   port_t_has_length_operator(SgStatement *);
SgInportStmt *        transform_inport_type_length_operator(SgInportStmt *);
SgOutportStmt *       transform_outport_type_length_operator(SgOutportStmt *);

int                   trace;
int                   do_not_generate__c_file = 0;

static char *         currentFileRealBase;

int main(int argc, char **argv)
{
    SgProject            *project;
    String               *s1;
    String            *s2;

    int                  opt_force_code_generation = 0;
    FILE                 *fm_outfile = 0;
    int                  optionChar;
    char                 *dot_proj = ".proj";
    char                 *c_postfix = "__.c";
    char                 *c_link_postfix = "__link.c";
    String               *fm_input_file_string;
    String               fm_c_outfile;
    String               fm_c_link_outfile;
    String               fm_proj_file_name;

#if 0
#ifdef USE_LIBG_GETOPT
    GetOpt getopt(argc, argv, "vagtfco:");
#else
    extern char          *optarg;
#endif
#endif
    int                  i;
    char                 *opt_file;
    
    // default options
    opt_debug                = 0;
    opt_verbose              = 0;

    opt_debug                = 0;
    opt_verbose              = 0;
    opt_use_auto_vars        = 0;
    opt_use_private_common   = 0;
    opt_allow_common         = 0;
    opt_must_thread          = 0;
    opt_i8                   = 0;
    trace                    = 0;

    // parse args
    i = 1;
    while (1)
    {
	if (i >= argc-1) break;
	
	if (! strcmp(argv[i], "-o")) 
	{
	    fm_assert(i+1 <= argc);
	    opt_file = strdup(argv[i+1]);
	    fm_outfile = fopen(opt_file, "w");
	    if (! fm_outfile) {
		fm_internal_error("Cannot open %s for writing", opt_file);
	    }
	    i += 2;
	} else if (! strcmp(argv[i], "-g")) {
	    opt_debug = 1;
	    i++;
	} else if (! strcmp(argv[i], "-a")) {
	    opt_use_auto_vars = 1;
	    i++;
	} else if (! strcmp(argv[i], "-trace")) {
	    trace = 1;
	    i++;
	} else if (! strcmp(argv[i], "-v")) {
	    opt_verbose = 1;
	    i++;
	} else if (! strcmp(argv[i], "-allow_common")) {
	    opt_allow_common = 1;
	    i++;
	} else if (! strcmp(argv[i], "-force_codegen")) {
	    opt_force_code_generation = 1;
	    i++;
	} else if (! strcmp(argv[i], "-must_thread")) {
	    opt_must_thread = 1;
	    i++;
	} else if (! strcmp(argv[i], "-i8")) {
	    opt_i8 = 1;
	    i++;
	} else {
	    fprintf(stderr, "Unknown switch %s\n", argv[i]);
	    i++;
	}
    } 
    
    if (i >= argc) {
	usage(argv[0]);
    } else {
	fm_input_file_string = new String(argv[i]);
    }
    if (! fm_input_file_string->contains(".proj")) {
	fm_proj_file_name = *fm_input_file_string + dot_proj;
    } else {
	String *base 
	  = new String(fm_input_file_string->chars(), 
		       fm_input_file_string->index(".proj"));
	
	fm_proj_file_name = *fm_input_file_string;
    }
	
    project = new SgProject((char *)fm_proj_file_name.chars());
    fm_initialize();
    for (i = 0; i < project->numberOfFiles(); i++) {
	Regex fm_postfix   = "\\.fm$";
	Regex f_postfix    = "\\.f$";
	
	currentFile        = project->fileName(i);
	currentFileNameExp = new SgValueExp(currentFile);
	
	s1                 = new String(currentFile);
	if (s1->contains(fm_postfix)) {
	    String s = s1->before(fm_postfix);
//	    s2          = new String((s1->before(fm_postfix)).chars());
	    s2          = &s;
	    currentFileRealBase = strdup(s2->chars());
	    currentFileBase    = file_name_to_ident(s2->chars());
	    s2          = new String(currentFileRealBase);
	} else if (s1->contains(f_postfix)) {
	    String s = s1->before(f_postfix);
//	    s2          = new String((s1->before(f_postfix)).chars());
	    s2          = &s;
	    currentFileRealBase = strdup(s2->chars());
	    currentFileBase    = file_name_to_ident(s2->chars());
	    s2          = new String(currentFileRealBase);
	    do_not_generate__c_file = 1;
	} else {
	    fprintf(stderr, "\
Illegal file extension %s, fmc only process .fm and .f files\n",
		    s1->chars());
	    exit(1);
	}

	processFMStatementsInFile(&project->file(i));
	restoreLBContStmts();
	rmFMStmts();

	//
	// if -must_thread switch is on
	// check whether all processes have
	// CFMThreadableProcess directive
	// if not flag an error.
	//
	if (opt_must_thread) {
	    all_processes_must_be_threadable();
	}

	if (opt_force_code_generation || n_errors == 0) {
	    // generate the __.c file
	    fm_c_outfile = *s2 + c_postfix;

	    if (! do_not_generate__c_file) {
		gen_c_aux_file((char *)fm_c_outfile.chars());
	    }

	    if (fm_outfile) {
		(project->file(i)).unparse(fm_outfile);
	    } else {
		(project->file(i)).unparsestdout();
	    }
	}
	if (n_errors) {
	    exit(1);
	}
	n_errors = 0;
	n_warnings = 0;
    }
    return 0;
}

static void usage(char *prog_name)
{
    fprintf(stderr, "\
Usage: %s [options] project-file\n", prog_name);
    exit (1);
}

void rmFMStmts(void)
{
    while (! fmStmtSLL.empty()) {
	SgStatement *s;
	s = fmStmtSLL.remove_front();
	if (0) {
	    fprintf(stderr,"removing statment: ");
	    s->unparsestdout();
	}
	s->deleteStmt();
    }
}

void restoreLBContStmts(void)
{
    while (! lbContStmtSLL.empty()) {
	SgStatement *s;
	s = lbContStmtSLL.remove_front();
	if (0) {
	    fprintf(stderr,"removing statment: ");
	    s->unparsestdout();
	}
	s->setVariant(CONT_STAT);
    }
}

    
/*
 * processFMStatementsInFile()
 *
 * Walk though the statements in the indicated FM file 
 * in the lexical order and process each FM statement to:
 *
 * 1.  generate a F77 program
 * 2.  generate a C program with auxiliary functions for
 *     concurrency and communication operations.
 *
 */

extern void print_statement(SgStatement *);
void processFMStatementsInFile(SgFile *file)
{
    SgStatement *line;
    // last statement processed
    SgStatement *lsp;
    SgStatement *lastExeStmt;
    int i;

    for (line = file->firstStatement(); line; line = lsp->lexNext()) {
	char *comments;

	currentStatement = line;
	comments         = line->comments();
	//
	// Do comments 
	//
	// check to see whether there is any compiler directives
	//
	if (comments) {
	    do_compiler_directives(comments, line);
	}

	if (line->variant() != GLOBAL)
	{
	    if (trace) {
		printf("processing statements:\n");
		print_statement(line);
	    }
	}
	switch (line->variant()) {
	  case PROG_HEDR:
	  case PROC_HEDR:
	  case FUNC_HEDR:
	  case PROS_HEDR:
	    // 
	    // insert a call to fm_shutdown() 
	    // at the end of the program
	    //

//	    if (line->variant() == PROG_HEDR) {
//		insertShutDownCall(line->lastExecutable());
//	    }
	    currentProcedure     = line;
	    currentProcedureName = (currentProcedure->symbol())->identifier();
	    lastDeclStmt         = currentProcedure->lastDeclaration();
	    firstUserStmt        = lastDeclStmt->lexNext();

	    // Get Leaving Scope statments
	    get_LS_stmts(currentProcedure);

	    if (line->variant() == PROG_HEDR) {
		firstExeStmt = new SgCallStmt(*fmStartCallSymb);
	    } else {
		firstExeStmt         = new SgContinueStmt;
		fmStmtSLL.append(firstExeStmt);
	    }
	    firstUserStmt->insertStmtBefore(*firstExeStmt);

	    if (! do_not_generate__c_file)
	      fmNProcessorsCallSymb->declareTheSymbol(*currentProcedure);
	    
	    if (! stmtMapGet(mpBlock, line->symbol()->identifier())) {
		line->replaceSymbBySymbSameName(*fmNPIntrinsicSymb, *fmNProcessorsCallSymb);
	    }

	    //
	    // Create a new FM program block table
	    //
	    currentProgBlock     = newFMProgBlock();
	    fmProgBlockMapSet(fmProgBlockMap,
			      line->symbol()->identifier(),
			      currentProgBlock);
	    //
	    // initialize default implicit type
	    //
	    initImplicitType(line);
	    //
	    // make a call to fm_init() in the PROGRAM block
	    //
	    if (line->variant() == PROG_HEDR 
		&& firstExeStmt->variant() != CONTROL_END) {
		SgFunctionCallExp *call;
		SgVarRefExp       *is_master_reference;
		SgAssignStmt      *fm_init_call_assign;

		fmIsMasterSymb->declareTheSymbol(*currentProcedure);
		fmInitCallSymb->declareTheSymbol(*currentProcedure);

		is_master_reference = new SgVarRefExp(*fmIsMasterSymb);
		call = new SgFunctionCallExp(*fmInitCallSymb);
		fm_init_call_assign = new SgAssignStmt(*is_master_reference,
						       *call);

		firstExeStmt->insertStmtBefore(*fm_init_call_assign);
#if 0
		lastExeStmt = line->lastExecutable();
		if (lastExeStmt->variant()!=PROCESSES_END) {
		    insertShutDownCall(lastExeStmt,0);
		} 		    
#endif
	    } 
	    //
	    // Process PROCESS header statement
	    //
	    if (line->variant() == PROS_HEDR) {
		doProsHedrStmt(line);
	    }
	    if (line->variant() == PROC_HEDR) {
		doProcHedrStmt(line);
	    }
	    if (line->variant() == FUNC_HEDR) {
		doFuncHedrStmt(line);
	    }
	    doCanonalizeDecl(line->lexNext(), line->lastNodeOfStmt());
	    doCommonStmtsInProg(line->lexNext(), line->lastDeclaration()->lexNext());
	    if (opt_use_auto_vars) {
		doParametersAndSaveStmtsInProg(line->lexNext(), line->lastDeclaration()->lexNext());
	    }

	    lsp = line;
	    break;
	  case DISTRIBUTE_STMT:
	    doDistributeStmt(line);
	    lsp = line;
	    break;
          case MP_STAT:
	  case MP_STAT_LCTN:
	    lsp = line;
	    break;
	  case MP_STAT_SUBM:
	    doMPCallStmt(line);
	    break;
	  case VAR_DECL:
	    doVarDeclStmt(line);
	    lsp = line;
	    break;
	  case DIM_STAT:
	    doDimensionStmt(line);
	    lsp = line;
	    break;
	  case INPORT_DECL:
	    SgInportStmt *inportStmt; // = isSgInportStmt(line);
	    
	    if (port_t_has_ports(line) ||
		port_t_has_length_operator(line))
	    {
		inportStmt = transform_inport_type_length_operator(isSgInportStmt(line));
		fm_assert(inportStmt);
		line->insertStmtBefore(*inportStmt);
		doInportDeclStmt(inportStmt);
		fmStmtSLL.append(line);
	    } else {
		inportStmt = isSgInportStmt(line);
		doInportDeclStmt(line);
	    }

	    for (i = 0; i < inportStmt->numberOfNames(); i++) {
		SgSymbol *symb = inportStmt->name(i)->symbol();
		
		portMapSet(prosMap,
			   prosMapKeyEncode(currentProcedureName, symb->identifier()),
			   inportStmt);
	    }
	    lsp = line;
	    break;
	  case OUTPORT_DECL:
	    SgOutportStmt *outportStmt; // = isSgOutportStmt(line);

	    if (port_t_has_ports(line) ||
		port_t_has_length_operator(line)) 
	    {
		outportStmt = transform_outport_type_length_operator(isSgOutportStmt(line));
		fm_assert(outportStmt);
		line->insertStmtBefore(*outportStmt);
		doOutportDeclStmt(outportStmt);
		fmStmtSLL.append(line);
	    } else {
		outportStmt = isSgOutportStmt(line);
		doOutportDeclStmt(line);
	    }
	    
	    for (i = 0; i < outportStmt->numberOfNames(); i++) {
		SgSymbol *symb = outportStmt->name(i)->symbol();
		
		portMapSet(prosMap,
			   prosMapKeyEncode(currentProcedureName, symb->identifier()),
			   outportStmt);
	    }
	    lsp = line;
	    break;
	  case CHANNEL_STAT:
	    doChannelStmt(line);
	    lsp = line;
	    break;
#ifdef _SAGEXX_1_3_
	  case MERGER_STAT:
#else
	  case MERGE_STAT:
#endif /* _SAGEXX_1_3 */
	    doMergerStmt(line);
	    lsp = line;
	    break;
	  case SEND_STAT:
	    doSendStmt(line);
	    lsp = line;
	    break;
	  case RECEIVE_STAT:
	    doReceiveStmt(line);
	    lsp = line;
	    break;
	  case PROBE_STAT:
	    doProbeStmt(line);
	    lsp = line;
	    break;
	  case ENDCHANNEL_STAT:
	    doEndchannelStmt(line);
	    lsp = line;
	    break;
	  case PROCESSORS_STAT:
	    doProcessorsStmt(line);
	    lsp = line;
	    break;
	  case PROCESSES_STAT:
	  case PROCESS_DO_STAT:
	    doParallelBlock(line,&lsp);
	    break;
	  case PROS_STAT:
	  case PROS_STAT_LCTN:
	  case PROS_STAT_SUBM:
	  {
	      SgProcessesStmt    *par_begin = new SgProcessesStmt;
	      SgEndProcessesStmt *par_end   = new SgEndProcessesStmt;
	      
	      line->insertStmtBefore(*par_begin);
	      line->insertStmtAfter(*par_end);
	      
	      doParallelBlock(par_begin,&lsp);
	      break;
	  }
	  case MOVE_PORT:
	    doMoveportStmt(line);
	    lsp = line;
	    break;
	  case PARAM_DECL:
	    if (!opt_use_auto_vars)
	      doParameterStmt(line);
	    lsp = line;
	    break;
#if NO_LONGER_NEEDED
	  case PROS_COMM:
	    if (! opt_use_auto_vars)
	      doProsCommStmt(line);
	    lsp = line;
	    break;
	  case COMM_STAT:
	    if (! opt_use_auto_vars) 
	      doCommonStmt(line);
	    lsp = line;
	    break;
#endif /* NO_LONGER_NEEDED */
	  case IMPL_DECL:
	    doImplicitStmt(line);
	    lsp = line;
	    break;
	  case INTENT_STMT:
	    doIntentStmt(line);
	    lsp = line;
	    break;
	  case STOP_STAT:
	    doStopStmt(line);
	    lsp = line;
	    break;
	  case CONTROL_END:
	    if (isSgProsHedrStmt(line->controlParent())) {
		checkProsArgumentTypesIntents(isSgProsHedrStmt(line->controlParent()));
	    } else if (isSgProgHedrStmt(line->controlParent())) {
		lastExeStmt = line->lastExecutable();
		if (lastExeStmt->variant()!=PROCESSES_END) {
		    insertShutDownCall(lastExeStmt,0);
		} 		    
	    }

	    if (line == currentProcedure->lastNodeOfStmt())
	        reset_LS_stmts();

	    lsp = line;
	    break;
	  case CONT_STAT:
	    line->setVariant(COMMENT_STAT);
	    lbContStmtSLL.append(line);
	    lsp = line;
	    break;
	  case IF_NODE:
	  {
	      SgStatement *dummyStmt = new SgContinueStmt;
	      
	      line->insertStmtBefore(*dummyStmt);
	      fmStmtSLL.append(dummyStmt);
	      lsp = line;
	      break;
	  }
	  case ELSEIF_NODE:
	  {
	      //
	      // Else if statement has caused
	      // problems when inserting before the
	      // next statement is required.
	      //
	      line->setVariant(IF_NODE);
	      SgStatement *dummyStmt = new SgContinueStmt;
	      
	      line->insertStmtAfter(*dummyStmt);
	      fmStmtSLL.append(dummyStmt);
	      line->setVariant(ELSEIF_NODE);
	      lsp = line;
	      break;
	  }
//	  case DATA_DECL:
//	    doDataDeclStmt(line);
//	    lsp = line;
//	    break;
	  default:
	    lsp = line;
	    break;
	}
    }
}
	    

void insertShutDownCall(SgStatement *where, int insert_before)
{
    SgCallStmt *callStmt;

    callStmt = new SgCallStmt(*fmShutDownCallSymb);
    if (insert_before) {
	where->insertStmtBefore(*callStmt);
    } else {
	where->insertStmtAfter(*callStmt);
    }
}

void doStopStmt(SgStatement *s)
{
    SgCallStmt *callStmt;
    SgStatement *control_parent;

    control_parent = s->controlParent();

    fm_assert(control_parent);

    if (control_parent->variant() == LOGIF_NODE) {
	control_parent->setVariant(IF_NODE);
    }
   
    callStmt = new SgCallStmt(*fmStopCallSymb);
    s->insertStmtBefore(*callStmt);
    copyLabelFromOriginalStmt(s, callStmt);
    fmStmtSLL.append(s);
}

void do_compiler_directives(char *comments, SgStatement *s)
{
    String comment_string(comments);
    char   *key;
    String *key_string;

    comment_string.upcase();
    if (comment_string.contains(*threadableDirective)) {
	if (currentProcedure->variant() == PROS_HEDR) {
	    key = directiveMapKeyEncode(currentProcedureName,
					"threadableProcess");
	    stmtMapSet(prosMap, key, s);
	}
    }
}

char * file_name_to_ident(const char * file_name)
{
    char *ident;

    if (! file_name) {
	return (char *)0;
    }
    
    ident = strdup((char *)file_name);
    for (int i = 0; i < strlen(file_name); i++)
    {
	if (! isalnum(ident[i]) && ident[i] != '_') {
	    ident[i] = '_';
	}
    }
    return ident;
}

void doDataDeclStmt(SgStatement *s)
{
    char *statbuf;
    char *name;
    char *p1, *p2, *nlbp;
    char nlistbuf[4098]; // quick hack, hope it's long enough

    statbuf = strdup(s->unparse());
    
    p1 = strchr(statbuf,'d');

    // get over data

    fm_assert(p1[0] == 'd' && p1[1] == 'a');
    p1 += 5;
    nlbp = &nlistbuf[0];

    while (1) {
	p2 = strchr(p1,'/');
	if (! p2) { break; }
	*p2 = '\000';
	strcpy(nlbp, p1);
	nlbp += strlen(nlbp);
	p2++;
	
	// jump over the vlist /v1,v2...,vn/
	p1 = strchr(p2,'/');
	if (! p1) { break; }
	p1++;
    }
//    printf("nlist=%s\n", nlistbuf);
    name = strtok(nlistbuf, " ,");
    while (1) {
	if (! name) { break; }
//	printf("var=%s\n", name);
	stmtMapSet(prosMap, commonMapKeyEncode(name), s);
	name = strtok(NULL, " ,");
    }
}
	
	
	
	    
