// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>
#include <string.h>


void doParameterStmt(SgStatement *s)
{
    SgParameterStmt *paramStmt = isSgParameterStmt(s);

    assert(paramStmt);
    for (int i = 0; i < paramStmt->numberOfConstants(); i++) {
	parameterMapSet(parameterMap, 
			paramStmt->constant(i)->identifier(),
			paramStmt->value(i));
	
    }
}

SgStatement * make_parameter_stmt(char *ident, 
				  SgExpression *val,
				  SgStatement *scope)
{
    SgSymbol       *const_symbol;
    SgExpression   *const_ref;
    SgExprListExp  *const_eL;
    SgStatement    *parameter_stmt;

    const_symbol   = new SgConstantSymb(ident, *scope, *val);
    const_ref      = new SgRefExp(CONST_REF, *const_symbol);
    const_eL       = new SgExprListExp(*const_ref);
    
    parameter_stmt = new SgParameterStmt(*const_eL);
    return (parameter_stmt);
}

void parameterDump()
{
    Pix ind_1, ind_2;

    for (ind_1 = fmProgBlockMap->first(); ind_1; fmProgBlockMap->next(ind_1)) {
	FMProgBlock *block = &(fmProgBlockMap->contents(ind_1));
	SgStatement *s = block->proc();
	char *procName = s->symbol()->identifier();
	printf("proc(%s) ::\n", procName);
	
	switchToFMProgBlock(procName);
	for (ind_2 = parameterMap->first(); ind_2; parameterMap->next(ind_2)) {
	    String &key = parameterMap->key(ind_2);
	    char *string = (char *)key.chars();
	    SgExpression *param = &(parameterMap->contents(ind_2));

	    printf("%s = ", string);
	    param->unparsestdout();
	    printf("\n");
	}
    }
}

void doParametersAndSaveStmtsInProg(SgStatement *start, SgStatement *end)
{
    SgStatement *line;

    for (line = start; line != end; line = line->lexNext()) {
	if (line->variant() == PARAM_DECL) 
	  doParameterStmt(line);
	else if (line->variant() == DATA_DECL)
	  doDataStmt(line);
	else if (line->variant() == SAVE_DECL)
	  doSaveStmt(line);
    }
}

	    



