// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>
#include <String.h>

static void transform_port_type(SgType *, SgExpression *, SgExpression *,
				SgType **, SgExpression **);
static void replacePOPinRef(SgExpression *);
static void declare_pops_as_port_stmts(SgStatement *);
     
void doInportDeclStmt(SgStatement *s)
{
    SgInportStmt                    *inportStmt;
    SgCallStmt                      *callStmt;
    char                            *ECstr;
    SgVarDeclStmt                   *varDeclStmt;

    inportStmt = isSgInportStmt(s);
    if (! inportStmt) {
	fm_internal_error("\
Expect an inport statement at %s:%d but get an %s",
			  __FILE__, __LINE__, getVariantByName(s->variant()));
    }

    //
    // Iterate through the inport variables declared
    // in the current inport statement and do the following:
    //
    // 1. declare the variable as INTEGER type
    // 2. generate a fm_init_port(name,arity) call
    //    where arity is the size of the port array. 
    //
    for (int i = 0; i < inportStmt->numberOfNames(); i++) {
	SgExpression                *e;
	SgSymbol *symb;

	e = inportStmt->name(i);
	if (! e) {
	    fm_internal_error("\
Expect a non-null port name at %s:%d but gets a null one instead",
			      __FILE__, __LINE__);
	}

	//
	// check whether the port is in a common block
	// or in the parameter list
	//

	if (! stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
	    ! exprMapGet(prosHedrMap,prosHedrMapKeyEncode(currentProcedureName,
							  symbName(e))) &&
	    opt_use_auto_vars)
	{
	    declare_auto_symb(symbName(e), currentProcedure);
	}

	if (opt_debug) {
	    callStmt = new SgCallStmt(*fmInitInportDebugCallSymb);
	} else {
	    callStmt = new SgCallStmt(*fmInitInportCallSymb);
	}

	s->insertStmtBefore(*declare_the_var(e, integer_t));
	
	if (e->variant() == VAR_REF) {
	    SgVarRefExp *v_ref = isSgVarRefExp(e);

	    symb = new SgVariableSymb(symbName(v_ref));
	    symb->setType(*integer_t);
//	    symb->declareTheSymbol(*currentProcedure);

	    callStmt->addArg(*e);
	    callStmt->addArg(*c1);
	    
#if DO_LEAVE_SCOPE
	    // 
	    // Generate {in|out}port_leaving_scope() calls
	    // for local ports
	    //
	    if (! stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
		! exprMapGet(prosHedrMap,prosHedrMapKeyEncode(currentProcedureName,  symbName(e))))
	    {
		SgCallStmt *callStmt = 
		  new SgCallStmt(*fmInportLeavingScopeCallSymb);
		callStmt->addArg(*e);
		callStmt->addArg(*c1);
		insert_port_LS_calls(callStmt);
	    }
#endif /* DO_LEAVE_SCOPE */

	    if (opt_debug) {
		SgValueExp    *port_name_exp;
		port_name_exp = new SgValueExp(symb->identifier());

		callStmt->addArg(*port_name_exp);
	    }
	} else if (e->variant() == ARRAY_REF) {
	    SgArrayRefExp *a_ref         = isSgArrayRefExp(e);
	    SgArrayType   *array_t       = new SgArrayType(*integer_t);
	    SgExpression  &e_array_size  = arraySizeExp(a_ref);
	    SgVarRefExp   *varRef;

	    for (int i = 0; i < a_ref->numberOfSubscripts(); i++) {
		array_t->addRange(*a_ref->subscript(i));
	    }
	    symb   = new SgVariableSymb(symbName(a_ref),*array_t);
//	    symb->declareTheSymbol(*currentProcedure);
//	    s->insertStmtBefore(*declare_the_symbol(symb,integer_t));
	    varRef = new SgVarRefExp(*symb);

	    callStmt->addArg(*varRef);
	    callStmt->addArg(e_array_size);
#if DO_LEAVE_SCOPE
	    // 
	    // Generate {in|out}port_leaving_scope() calls
	    // for local ports
	    //
	    if (! stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
		! exprMapGet(prosHedrMap,prosHedrMapKeyEncode(currentProcedureName, symbName(e))))
	    {
		SgCallStmt *callStmt 
		  = new SgCallStmt(*fmInportLeavingScopeCallSymb);
		callStmt->addArg(*varRef);
		callStmt->addArg(e_array_size);
		insert_port_LS_calls(callStmt);
	}
#endif /* DO_LEAVE_SCOPE */
	    if (opt_debug) {
		SgValueExp    *port_name_exp;
		port_name_exp = new SgValueExp(symb->identifier());

		callStmt->addArg(*port_name_exp);
	    }
	} else {
	    fm_error(s, "\
Invalid variable for a declaration statement: %s",
		     getVariantByName(e->variant()));
	}

	// 
	// Insert this port variable into the port dictionary
	// 

	ECstr = portDictKeyEncode(symb);
	portMapSet(portVarDict, ECstr, s);

	//
	// Insert this port type into the port set
	// Note if there is already one such port exisit in
	// the portSet then the portMapSet will not install 
	// the contents into portMap.
	//

	ECstr = portSetKeyEncode(s);
	portMapSet(portSet, ECstr,s);

	// 
	// if the port designated by E is a formal parameter
	// do not make fm_init_port().
	//

	if (portTypeMapGet(prosHedrMap, prosHedrMapKeyEncode(currentProcedureName, symbName(e))))
	{
	    delete callStmt;
	    continue;
	} else if (stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
		   (currentProcedure->variant() == PROC_HEDR ||
		    currentProcedure->variant() == FUNC_HEDR))
	{
	    delete callStmt;
	    continue;
	} else {
	    // Plant the calls to initialize the port
	    firstUserStmt->insertStmtBefore(*callStmt);
	}
    }

    //
    // add the statement to the FM statement list
    //
    fmStmtSLL.append(s);
}

void doOutportDeclStmt(SgStatement *s)
{
    SgOutportStmt                   *outportStmt;
    SgCallStmt                      *callStmt;
    char                            *ECstr;
    SgVarDeclStmt                   *varDeclStmt;

    outportStmt = isSgOutportStmt(s);

    if (! outportStmt) {
	fm_internal_error("\
Expect an outport statement at %s:%d but get an %s",
			  __FILE__, __LINE__, getVariantByName(s->variant()));
    }

    //
    // Iterate through the outport variables declared
    // in the current outport statement and do the following:
    //
    // 1. declare the variable as INTEGER type
    // 2. generate a fm_init_port(name,arity) call
    //    where arity is the size of the port array. 
    //
    for (int i = 0; i < outportStmt->numberOfNames(); i++) {
	SgExpression                *e;
	SgSymbol *symb;

	e = outportStmt->name(i);
	if (! e) {
	    fm_internal_error("\
Expect a non-null port name at %s:%d but gets a null one instead",
			      __FILE__, __LINE__);
	}

	//
	// check whether the port is in a common block
	// or in the parameter list
	//

	if (! stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
	    ! exprMapGet(prosHedrMap,prosHedrMapKeyEncode(currentProcedureName,
							  symbName(e))) &&
	    opt_use_auto_vars)
	    
	{
	    declare_auto_symb(symbName(e), currentProcedure);
	}

	if (opt_debug) {
	    callStmt = new SgCallStmt(*fmInitOutportDebugCallSymb);
	} else {
	    callStmt = new SgCallStmt(*fmInitOutportCallSymb);
	}

	s->insertStmtBefore(*declare_the_var(e, integer_t));

	if (e->variant() == VAR_REF) {
	    SgVarRefExp *v_ref = isSgVarRefExp(e);

	    symb = new SgVariableSymb(symbName(v_ref));
	    symb->setType(*integer_t);
//	    symb->declareTheSymbol(*currentProcedure);

	    callStmt->addArg(*e);
	    callStmt->addArg(*c1);

#if DO_LEAVE_SCOPE
	    // 
	    // Generate {in|out}port_leaving_scope() calls
	    // for local ports
	    //
	    if (! stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
		! exprMapGet(prosHedrMap,prosHedrMapKeyEncode(currentProcedureName,  symbName(e))))
	    {
		SgCallStmt *callStmt = 
		  new SgCallStmt(*fmOutportLeavingScopeCallSymb);
		callStmt->addArg(*e);
		callStmt->addArg(*c1);
		insert_port_LS_calls(callStmt);
	    }
#endif /* DO_LEAVE_SCOPE */

	    if (opt_debug) {
		SgValueExp    *port_name_exp;
		port_name_exp = new SgValueExp(symb->identifier());

		callStmt->addArg(*port_name_exp);
	    }
	} else if (e->variant() == ARRAY_REF) {
	    SgArrayRefExp *a_ref         = isSgArrayRefExp(e);
	    SgArrayType   *array_t       = new SgArrayType(*integer_t);
	    SgExpression  &e_array_size  = arraySizeExp(a_ref);
	    SgVarRefExp   *varRef;

	    for (int i = 0; i < a_ref->numberOfSubscripts(); i++) {
		array_t->addRange(*a_ref->subscript(i));
	    }
	    symb   = new SgVariableSymb(symbName(a_ref),*array_t);
//	    symb->declareTheSymbol(*currentProcedure);
	    varRef = new SgVarRefExp(*symb);

	    callStmt->addArg(*varRef);
	    callStmt->addArg(e_array_size);
#if DO_LEAVE_SCOPE
	    // 
	    // Generate {in|out}port_leaving_scope() calls
	    // for local ports
	    //
	    if (! stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
		! exprMapGet(prosHedrMap,prosHedrMapKeyEncode(currentProcedureName, symbName(e))))
	    {
		SgCallStmt *callStmt 
		  = new SgCallStmt(*fmOutportLeavingScopeCallSymb);
		callStmt->addArg(*varRef);
		callStmt->addArg(e_array_size);
		insert_port_LS_calls(callStmt);
	}
#endif /* DO_LEAVE_SCOPE */

	    if (opt_debug) {
		SgValueExp    *port_name_exp;
		port_name_exp = new SgValueExp(symb->identifier());

		callStmt->addArg(*port_name_exp);
	    }
	} else {
	    fm_error(s, "\
Invalid variable for a declaration statement: %s",
		     getVariantByName(e->variant()));
	}

	// 
	// Insert this port variable into the port dictionary
	// 

	ECstr = portDictKeyEncode(symb);
	portMapSet(portVarDict, ECstr, s);

	//
	// Insert this port type into the port set
	// Note if there is already one such port exisit in
	// the portSet then the portMapSet will not install 
	// the contents into portMap.
	//
	ECstr = portSetKeyEncode(s);
	portMapSet(portSet, ECstr,s);

	// 
	// if the port designated by E is a formal parameter
	// do not make fm_init_port().
	//
	if (portTypeMapGet(prosHedrMap, prosHedrMapKeyEncode(currentProcedureName, symbName(e))))
	{
	    delete callStmt;
	    continue;
	} else if (stmtMapGet(prosMap, commonMapKeyEncode(symbName(e))) &&
		   (currentProcedure->variant() == PROC_HEDR ||
		    currentProcedure->variant() == FUNC_HEDR ))
	{
	    delete callStmt;
	    continue;
	} else {
	    // Plant the calls to initialize the port
	    firstUserStmt->insertStmtBefore(*callStmt);
	}
    }

    //
    // add the statement to the FM statement list
    //
    fmStmtSLL.append(s);
}

SgStatement * declare_the_var(SgExpression *var, SgType *t)
{
    SgExprListExp   *var_list;
    SgVarDeclStmt   *decl_statement;

    fm_assert(var);
    fm_assert(t);

    var_list       = new SgExprListExp(*var);
    decl_statement = new SgVarDeclStmt(*var_list, *t);
    return (decl_statement);
}


SgInportStmt * transform_inport_type_length_operator(SgInportStmt *s)
{
    SgInportStmt   *newInportStmt;
    int            i;
    int            n_port_types;
    SgPortTypeExp  *new_port_t;
    fm_assert(s);

    declare_pops_as_port_stmts(s);

    n_port_types = s->numberOfPortTypes();

    for (i = 0; i < n_port_types; i++) {
	SgPortTypeExp *port_t;
	SgType        *t, *new_t;
	SgExpression  *ref, *new_ref;
	SgPortTypeExp *port_t_copy;

	port_t   = s->porttype(i);
//	port_t_copy = isSgPortTypeExp(&(port_t->copy()));

	t        = port_t->type();
	ref      = port_t->ref();

	if (port_t->variant() == INPORT_TYPE_OP) {
	    /* insert a port definition to
	       the portSetMap so that we can 
	       generate fwd_dest_t for this
	       type */
#if 0
	    SgInportStmt  *pop_inport;
	    SgPortTypeExp *pop_t;
	    SgSymbol      *pop_symb;
	    SgVarRefExp   *pop_var;
	    SgExprListExp *pop_names;

//	    replacePOPinRef(ref);

	    pop_t = isSgPortTypeExp(ref);
	    pop_symb = new SgVariableSymb("pop_port", 
					  *integer_t);

	    pop_var = new SgVarRefExp(*pop_symb);
	    pop_names = new SgExprListExp(*pop_var);

	    fm_assert(pop_t);
	    pop_inport = new SgInportStmt(*pop_names, *pop_t);

	    portMapSet(portSet, portSetKeyEncode(pop_inport), pop_inport);
#endif /* 0 */
	    /* transform the port for ease of the
	      code generation stage */
	    t = inport_t;
	    /* replace recursively all the nested inports/outports
	      with the type */
//	    ref = 0;
	} else if (port_t->variant() == OUTPORT_TYPE_OP) {
	    t = outport_t;
//	    replacePOPinRef(ref);
//	    ref = 0;
	}

	if (t->length()) {
	    SgExpression  *length;

	    length = isSgStringLengthExp(t->length())->length();
	    
	    if (! IsInteger(length)) {
		fm_error(s, "\
expect to get constant integer for a length operator but get %s",
			 length->unparse());
		return(s);
	    }
	
	    //
	    // build new_t and new_ref
	    //
	    transform_port_type(t, ref, length, &new_t, &new_ref);

#if 0
	    new_t = (t->variant()==T_STRING)? char_t : new SgType(t->variant());

	    if (ref) {
		new_ref = make_array_ref(new_t, length, ref, isSgArrayRefExp(ref));
	    } else {
		new_ref = make_array_ref(new_t, length, default_ref, (SgArrayRefExp *)0);
	    }
#endif /* 0 */

	    if (! new_ref) {
		new_port_t = new SgPortTypeExp(*new_t);
	    } else {
		new_port_t = new SgPortTypeExp(*new_t, *new_ref);
	    }

	    if (i == 0) {
		newInportStmt  = new SgInportStmt(*s->names(), *new_port_t);
	    } else {
		newInportStmt->addporttype(*new_port_t);
	    }
	} else {
	    if (ref) {
		new_port_t = new SgPortTypeExp(*t, *ref);
	    } else {
		new_port_t = new SgPortTypeExp(*t);
	    }

	    if (i == 0) {
		newInportStmt = new SgInportStmt(*s->names(), *new_port_t);
	    } else {
		newInportStmt->addporttype(*new_port_t);
	    }
	}
    }
    newInportStmt->setlineNumber(s->lineNumber());
    return newInportStmt;
}

SgOutportStmt * transform_outport_type_length_operator(SgOutportStmt *s)
{
    SgOutportStmt    *newOutportStmt;
    int              i;
    int n_port_types;
    SgPortTypeExp *new_port_t;

    fm_assert(s);

    declare_pops_as_port_stmts(s);

    n_port_types = s->numberOfPortTypes();

    for (i = 0; i < n_port_types; i++) {
	SgPortTypeExp *port_t, *port_t_copy;
	SgType        *t, *new_t;
	SgExpression  *ref, *new_ref;
	
	port_t   = s->porttype(i);
//	port_t_copy = isSgPortTypeExp(&(port_t->copy()));

	t        = port_t->type();
	ref      = port_t->ref();

	if (port_t->variant() == INPORT_TYPE_OP) {
	    t = inport_t;
//	    ref = 0;
	} else if (port_t->variant() == OUTPORT_TYPE_OP) {
	    t = outport_t;
//	    ref = 0;
	}

	if (t->length()) {
	    SgExpression  *length;

	    length = isSgStringLengthExp(t->length())->length();
	    
	    if (! IsInteger(length)) {
		fm_error(s, "\
expect to get constant integer for a length operator but get %s",
			 length->unparse());
		return(s);
	    }
	
	    //
	    // build new_t and new_ref
	    //
	    transform_port_type(t, ref, length, &new_t, &new_ref);

#if 0
	    new_t = (t->variant()==T_STRING)? char_t : new SgType(t->variant());
	    if (ref) {
		new_ref = make_array_ref(new_t, length, ref, isSgArrayRefExp(ref));
	    } else {
		new_ref = make_array_ref(new_t, length, default_ref, (SgArrayRefExp *)0);
	    }
#endif /* 0 */

	    if (! new_ref) {
		new_port_t = new SgPortTypeExp(*new_t);
	    } else {
		new_port_t = new SgPortTypeExp(*new_t, *new_ref);
	    }

	    if (i == 0) {
		newOutportStmt  = new SgOutportStmt(*s->names(),*new_port_t);
	    } else {
		newOutportStmt->addporttype(*new_port_t);
	    }
	} else {
	    if (ref) {
		new_port_t = new SgPortTypeExp(*t, *ref);
	    } else {
		new_port_t = new SgPortTypeExp(*t);
	    }
					       
	    if (i == 0) {
		newOutportStmt = new SgOutportStmt(*s->names(), *new_port_t);
	    } else {
		newOutportStmt->addporttype(*new_port_t);
	    }
	}
    }

    newOutportStmt->setlineNumber(s->lineNumber());
    return newOutportStmt;
}

int port_t_has_ports(SgStatement *s)
{
    int i;

    if (s->variant() == INPORT_DECL) {
	SgInportStmt *inport = isSgInportStmt(s);

	for (i = 0; i < inport->numberOfPortTypes(); i++) {
	    SgPortTypeExp *port_t = inport->porttype(i);
	    
	    if (port_t->variant() == INPORT_TYPE_OP ||
		port_t->variant() == OUTPORT_TYPE_OP)
	    {
#if 0
		fm_warning(s,"\
port migration is not yet supported, treat port type field %d as integer",
			   i+1);
#endif /* 0 */
		return 1;
	    }
	}
    } else if (s->variant() == OUTPORT_DECL) {
	SgOutportStmt *outport = isSgOutportStmt(s);

	for (i = 0; i < outport->numberOfPortTypes(); i++) {
	    SgPortTypeExp *port_t = outport->porttype(i);
	    
	    if (port_t->variant() == INPORT_TYPE_OP ||
		port_t->variant() == OUTPORT_TYPE_OP)
	    {
#if 0
		fm_warning(s,"\
port migration is not yet supported, treat port type field %d as integer",
			   i+1);
#endif /* 0 */
		return 1;
	    }
	}
    } else {
	fm_internal_error("\
Do not expect port statement to be %s at %s:%d",
			  getVariantByName(s->variant()),
			  __FILE__, __LINE__);
    }
    return 0;
}

int port_t_has_length_operator(SgStatement *s)
{
    int i;

    if (s->variant() == INPORT_DECL) {
	SgInportStmt *inport = isSgInportStmt(s);

	for (i = 0; i < inport->numberOfPortTypes(); i++) {
	    SgPortTypeExp *port_t = inport->porttype(i);
	    SgType        *t      = port_t->type();

	    //
	    // skip the recursive port type
	    //
	    if (port_t->variant() == INPORT_TYPE_OP ||
		port_t->variant() == OUTPORT_TYPE_OP)
	    {
		continue;
	    }

	    if (t->length()) {
		return 1;
	    } 
	}
    } else if (s->variant() == OUTPORT_DECL) {
	SgOutportStmt *outport = isSgOutportStmt(s);

	for (i = 0; i < outport->numberOfPortTypes(); i++) {
	    SgPortTypeExp *port_t = outport->porttype(i);
	    SgType        *t      = port_t->type();

	    //
	    // skip the recursive port type
	    //
	    if (port_t->variant() == INPORT_TYPE_OP ||
		port_t->variant() == OUTPORT_TYPE_OP)
	    {
		continue;
	    }

	    if (t->length()) {
		return 1;
	    } 
	}
    } else {
	fm_internal_error("\
Do not expect port statement to be %s at %s:%d",
			  getVariantByName(s->variant()),
			  __FILE__, __LINE__);
    }
    return 0;
}

static void 
transform_port_type(SgType *t, SgExpression *ref, SgExpression *length,
		    SgType **new_t, SgExpression **new_ref)
{
    SgSymbol       *default_symb;
    SgVarRefExp    *default_ref;

    if (t->variant() == T_INT && length) {
	if (integerValue(length) == 1) {
	    *new_t = char_t;
	} else if (integerValue(length) == 2) {
	    *new_t = new SgType(T_SHORT);
	} else if (integerValue(length) == 4) {
	    *new_t = integer_t;
	} else if (integerValue(length) == 8) {
	    *new_t = double_t;
	} else {
	    *new_t = t;
	}
        *new_ref = ref;
	return;
    }

    *new_t = (t->variant()==T_STRING)? char_t : new SgType(t->variant());
    
    default_symb = new SgVariableSymb("fm__ch_ref",*integer_t);
    default_ref  = new SgVarRefExp(*default_symb);

    if (ref) {
	*new_ref = make_array_ref(*new_t, length, ref, isSgArrayRefExp(ref));
    } else {
	*new_ref = make_array_ref(*new_t,length,default_ref, (SgArrayRefExp *)0);
    }
}    
	

static void replacePOPinRef(SgExpression *r)
{
    int i;
    SgPortTypeExp *p;
    SgType *t;

    if (! r) { return; }

    p = isSgPortTypeExp(r);
    if (! p) { return; }
    
    while (p) {
	t = p->type();

	if (t) {
	    p = p->next();
	    continue;
	}

	if (p->variant() == INPORT_TYPE_OP) {
	    p->setType(*inport_t);
	    replacePOPinRef(p->ref());
	} else if (p->variant() == OUTPORT_TYPE_OP) {
	    p->setType(*outport_t);
	    replacePOPinRef(p->ref());
	} 
	    
	p = p->next();
    }
}


/*
 * void walk_through_pop_and_stash_ports()
 *
 * Add port over port decls into portSet.
 * e.g. for:  inport(outport(inport(integer, real x(10))))
 * we would add the following port decls:
 *     outport(inport(integer, real x(10)))
 *     inport(integer, real x(10))
 *
 */

static void declare_pops_as_port_stmts(SgStatement *s)
{
    int i;
    int ntypes = 0;
    SgInportStmt *inportStmt;
    SgOutportStmt *outportStmt;

    if (s->variant() == INPORT_DECL) {
	inportStmt = isSgInportStmt(s);
	ntypes = inportStmt->numberOfPortTypes();
    } else if (s->variant() == OUTPORT_DECL) {
	outportStmt = isSgOutportStmt(s);
	ntypes = outportStmt->numberOfPortTypes();
    }

    for (i = 0; i < ntypes; i++) {
	SgPortTypeExp *port_t = (s->variant() == INPORT_DECL) ?
	  inportStmt->porttype(i) : outportStmt->porttype(i);
	
	if (port_t->variant() == INPORT_TYPE_OP) {
	    SgInportStmt  *pop_inport;
	    SgPortTypeExp *pop_t;
	    SgSymbol      *pop_symb;
	    SgVarRefExp   *pop_var;
	    SgExprListExp *pop_names;
	    SgExpression  *ref = port_t->ref();

	    fm_assert(ref);
	    
	    pop_t = isSgPortTypeExp(ref);
	    pop_symb = new SgVariableSymb("pop_port", 
					  *integer_t);
	    
	    pop_var = new SgVarRefExp(*pop_symb);
	    pop_names = new SgExprListExp(*pop_var);
	    
	    fm_assert(pop_t);
	    pop_inport = new SgInportStmt(*pop_names, *pop_t);

//	    pop_inport->unparsestdout();
	    portMapSet(portSet, portSetKeyEncode(pop_inport), pop_inport);
	    /*
	     * Do the nested pop
	     */
	    declare_pops_as_port_stmts(pop_inport);
	} else if (port_t->variant() == OUTPORT_TYPE_OP) {
	    SgOutportStmt  *pop_outport;
	    SgPortTypeExp *pop_t;
	    SgSymbol      *pop_symb;
	    SgVarRefExp   *pop_var;
	    SgExprListExp *pop_names;
	    SgExpression  *ref = port_t->ref();
	    
	    fm_assert(ref);
	    
	    pop_t = isSgPortTypeExp(ref);
	    pop_symb = new SgVariableSymb("pop_port", 
					  *integer_t);
	
	    pop_var = new SgVarRefExp(*pop_symb);
	    pop_names = new SgExprListExp(*pop_var);
		
	    fm_assert(pop_t);
	    pop_outport = new SgOutportStmt(*pop_names, *pop_t);

//	    pop_outport->unparsestdout();

	    portMapSet(portSet, portSetKeyEncode(pop_outport),pop_outport);
	    /*
             * Do the nested pop
	     */
	    declare_pops_as_port_stmts(pop_outport);
	} 
    }
}
