// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>

void doProsHedrStmt(SgStatement *s)
{
    char *prosName = s->symbol()->identifier();

    SgProsHedrStmt   *prosHedrStmt = isSgProsHedrStmt(s);
    SgStatement *line;
    SgStatement *lsp = prosHedrStmt->lastDeclaration();
    int i,j,k;
    
    // 
    // stuff process parameters into prosHedrMap
    // the associated value are A_0, A_1, ...
    //
    for (i = 0; i < prosHedrStmt->numberOfParameters(); i++) {
	char argStr[32];
	SgVariableSymb *symb;
	SgVarRefExp *argRef;
	SgSymbol *realParam = prosHedrStmt->parameter(i);

	sprintf(argStr, "A_%d", i);
	symb = new SgVariableSymb(argStr, *integer_t);
	argRef = new SgVarRefExp(*symb);

	portTypeMapSet(prosHedrMap, 
		       prosHedrMapKeyEncode(prosName, realParam->identifier()),
		       argRef);
    }
    portMapSet(prosDefMap, 
	       prosDefMapKeyEncode(prosName),
	       s);
}


void checkProsArgumentTypesIntents(SgProsHedrStmt *prosHedrStmt)
{
    int i;
    char *prosName;

    fm_assert(prosHedrStmt);
    prosName = prosHedrStmt->symbol()->identifier();

    for (i = 0; i < prosHedrStmt->numberOfParameters(); i++) {
	SgSymbol *symb = prosHedrStmt->parameter(i);
	SgExpression *attrib;

	fm_assert(symb);
        //
	// check type
	//
	if (! portMapGet(prosMap, 
			 prosMapKeyEncode(prosName,symb->identifier()))) 
	{
	    SgType *t = getImplicitType(prosName, symb);
            SgVarDeclStmt *decl_s;
	    
	    if (! t) {
		fm_error(prosHedrStmt, "\
argument %d (%s) type unknown", i+1, symb->identifier());
	    } else {
		SgExprListExp *varList;
                SgVariableSymb *varSymb =new SgVariableSymb(symb->identifier(),
                                                            *t);
		SgVarRefExp *var; 
		
		varSymb->setType(*t);
		var = new SgVarRefExp(*varSymb);
		varList = new SgExprListExp(*var);

                decl_s = new SgVarDeclStmt(*varList, *t);
//                decl_s->unparsestdout();
//                printf("\n");
		portMapSet(prosMap, 
			   prosMapKeyEncode(prosName,symb->identifier()),
			   decl_s);
		
		varSymb->declareTheSymbol(*currentProcedure);
	    }
	} else {
	    SgExpression *var;
	    
	    var = findSymbInDeclStmt(symb);
	    if (var->variant() == ARRAY_REF) {
		SgArrayRefExp *a;

		a = isSgArrayRefExp(var);
		for (int var_i = 0; var_i < a->numberOfSubscripts(); var_i++)
		{
		    SgExpression *sub = a->subscript(var_i);
		    
		    if (sub->variant()  == STAR_RANGE)
		    {
			fm_error(currentProcedure,"\
unknown array length * in %s is not allowed for the declaration of process argument", 
				 a->unparse());
		    }
		}
	    }
	}

	// 
	// Check intents
	//
	attrib = intentMapGet(intentMap, symb->identifier());
	if (! attrib) {
	    SgVarRefExp       *portRef;

	    portRef           = new SgVarRefExp(*symb);
	    if (getPortVariant(portRef) >= 0) {
		intentMapSet(intentMap, symb->identifier(), in_attribute);
	    } else {
		intentMapSet(intentMap, symb->identifier(), inout_attribute);
	    }
	} else {
	    SgVarRefExp       *portRef;

	    portRef           = new SgVarRefExp(*symb);
	    if (getPortVariant(portRef) >= 0) {
		// 
		// Give warnings for ports whose attribute is OUT/INOUT
		//
		if (attrib->variant() == OUT_OP||attrib->variant() == INOUT_OP)
		{
		    fm_warning(prosHedrStmt, "\
Ports can only have intent(in) attribute, but %s is declared as intent(%s)",
			       symb->identifier(),
			       (attrib->variant()==OUT_OP)? "out" : "inout");
		    exprMapOverwrite(intentMap,symb->identifier(),in_attribute);
		}
	    }
	}
    }
}

/*
 * doProcHedrStmt()
 *
 * Put the parameter names into the prosHedrMap so
 * that no port initialization will be performed for
 * ports that are in the argument list.
 *
 */

void doProcHedrStmt(SgStatement *s)
{
    SgProcHedrStmt  *hedr;
    char            *procName;

    hedr            = isSgProcHedrStmt(s);
    
    fm_assert(hedr);
    procName        = hedr->symbol()->identifier();

    for (int i = 0; i < hedr->numberOfParameters(); i++) {
	char *argument_ident;
	
	fm_assert(hedr->parameter(i));

	argument_ident = hedr->parameter(i)->identifier();
	exprMapSet(prosHedrMap,prosHedrMapKeyEncode(procName,argument_ident),
		   c0);
    }
}

void doFuncHedrStmt(SgStatement *s)
{
    char *funcName = s->symbol()->identifier();
    int   i;
    
    SgFuncHedrStmt  *funcHedrStmt = isSgFuncHedrStmt(s);
    for (i = 0; i < funcHedrStmt->numberOfParameters(); i++) {
	char argStr[32];
	SgVariableSymb *symb;
	SgVarRefExp *argRef;
	SgSymbol *realParam = funcHedrStmt->parameter(i);

	sprintf(argStr, "A_%d", i);
	symb = new SgVariableSymb(argStr, *integer_t);
	argRef = new SgVarRefExp(*symb);

	exprMapSet(prosHedrMap, 
		   prosHedrMapKeyEncode(funcName,realParam->identifier()),
		   argRef);
    }
}

	
