// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>
#include "string.h"

void doSendStmt(SgStatement *s)
{
    SgSendStmt        *sendStmt = isSgSendStmt(s);
    SgExpression      *outport  = portExp(sendStmt->outport(0));
//    SgExpression      *ioStore  = ioStoreExp(sendStmt->ioStore());
    SgExpression      *ioStore  = sendStmt->ioStore();
    SgExpression      *errLabel = labelExp(sendStmt->errLabel());
    char              *pname;
    char              *fmIOValStr;              
    SgSymbol          *fmIOValSymb;
    SgVarRefExp       *fmIOValRef;
    char              *sendFuncStr;
    char              *sendFuncFullStr;
    SgValueExp        *sendMgFuncExp;
    SgSymbol          *sendFuncSymb;
    SgCallStmt        *callStmt;
    int               sendFuncStrLen;
    SgOutportStmt     *portStmt;
    char              *ECstr;
    SgVarRefExp       *varRef;
    SgArrayRefExp     *arrayRef;
    SgExprListExp     *eList;
    SgStatement       *stmt;
    SgExpression      *ioTempVar;
    char              *outportName;
    int               i;


    //
    // get port declration from the map: portVarDict 
    // 
    stmt = getPortStatement(outport);
    portStmt = isSgOutportStmt(stmt);

    if (! portStmt) {
	outportName = symbName(portExp(outport));
	fm_assert(outportName);
	if (outportName) {
	    fm_error(s, "\
%s is not an outport", outportName);
	} 
    } else {
	//
	// Encode send function name
	//
	pname = portSetKeyEncode(stmt);
#if 0	
	sendFuncStrLen = 3 + strlen(pname) + 5;
	sendFuncStr = (char *)malloc(sendFuncStrLen + 1);
	assert(sendFuncStr);
	sprintf(sendFuncStr, "fm_%s_send", pname);
#endif /* 0 */

	sendFuncFullStr = ppString("fm_", pname, 
			       (opt_debug) ? "_snd_d" : "_snd");

	sendMgFuncExp = isSgValueExp(exprMapGet(portMangleMap,sendFuncFullStr));

	if (! sendMgFuncExp) {
	    //
	    // generate a four digit send/receive name
	    // 
	    //   send:  fm_<f_name_base>_sndXXXX
	    //   recv:  fm_<f_name_base>_rcvXXXX
	    //
	    sendFuncStr = ppString(ppString("fm_",currentFileBase,""),
				   (opt_debug) ? "_snd_d" : "_snd",
				   encodeStr(EC_MANGLE));

	    exprMapSet(portMangleMap, sendFuncFullStr,
		       new SgValueExp(sendFuncStr));
	} else {
	    sendFuncStr = sendMgFuncExp->stringValue();
	}

	sendFuncSymb = new SgSymbol(PROCEDURE_NAME, sendFuncStr);
	callStmt = new SgCallStmt(*sendFuncSymb);
	callStmt->addArg(*outport);
	
	//
	// Consistency Check
	//
	//    Number of Arguments should match the number
	//    of types in the corresponding port declaration
	//    statement.
	//
	if (portStmt->numberOfPortTypes() != sendStmt->numberOfArguments()) {
	    fm_error(s, "number of argument do not match number port types");
	}

	//
	// put type field which has a ref into
	// the portTypeMap
	//
	
	portTypeMapClear();

	for (i = 0; i < sendStmt->numberOfArguments(); i++) {
	    SgPortTypeExp *pte = portStmt->porttype(i);
	    SgExpression *arg = sendStmt->argument(i);

	    if (pte->ref() && pte->type() &&
		pte->ref()->variant() != PORT_TYPE_OP &&
		pte->ref()->variant() != INPORT_TYPE_OP &&
		pte->ref()->variant() != OUTPORT_TYPE_OP &&
		pte->ref()->variant() != T_INPORT &&
		pte->ref()->variant() != T_OUTPORT)
	    {
		char *name = symbName(pte->ref());
		
		portTypeMapSet(portTypeMap, name, arg);
	    }
	}
	//
	// do type checking by matching the argument list type
	// againt the port type
	//
	for (i = 0; i < sendStmt->numberOfArguments(); i++) {
	    SgPortTypeExp *pte = portStmt->porttype(i);
	    SgExpression  *arg = sendStmt->argument(i);
	    int t1, t2;
	    
	    matchPortAndArgTypes(i, pte, arg);
	}
        //
        // generate arguments to the fm_ptype_send() call
        //
	for (i = 0; i < sendStmt->numberOfArguments(); i++) {
	    SgPortTypeExp *pte = portStmt->porttype(i);
	    SgExpression  *arg = sendStmt->argument(i);
	    SgExpression  *e   = pte->ref();
	    
	    if (arrayRef = isSgArrayRefExp(e)) {
		SgExpression  *eSub  = &(arraySizeExp(arrayRef));
		SgExpression *realSize;

		realSize = &(replacePortProtoArgInExpr(eSub));

		callStmt->addArg(*arg);
		//	    eSub.unparsestdout();
		callStmt->addArg(*realSize);
	    } else {
		callStmt->addArg(*arg);
	    }
	}	
	ioTempVar = fmIOVarExp(ioStore);
	callStmt->addArg(*ioTempVar);

	if (opt_debug) {
	    SgValueExp *fileName = new SgValueExp(currentFile);
	    SgValueExp *lineNumber = new SgValueExp(s->lineNumber());

	    callStmt->addArg(*fileName);
	    callStmt->addArg(*lineNumber);
	}
	
	s->insertStmtBefore(*callStmt);
	s->insertStmtBefore(*checkIOStat(1,ioStore,ioTempVar,errLabel));
    }
    fmStmtSLL.append(s);
}
    
void doReceiveStmt(SgStatement *s)
{
    SgReceiveStmt     *receiveStmt = isSgReceiveStmt(s);
    SgExpression      *inport  = portExp(receiveStmt->inport(0));
    SgExpression      *ioStore  = receiveStmt->ioStore();
    SgExpression      *errLabel = labelExp(receiveStmt->errLabel());
    char              *pname;
    char              *fmIOValStr;              
    SgSymbol          *fmIOValSymb;
    SgVarRefExp       *fmIOValRef;
    char              *receiveFuncStr;
    char              *receiveFuncFullStr;
    SgValueExp        *receiveMgFuncExp;
    SgSymbol          *receiveFuncSymb;
    SgCallStmt        *callStmt;
    int               receiveFuncStrLen;
    SgInportStmt      *portStmt;
    char              *ECstr;
    SgVarRefExp       *varRef;
    SgArrayRefExp     *arrayRef;
    SgExprListExp     *eList;
    SgStatement       *stmt;
    SgExpression      *endLabel = labelExp(receiveStmt->endLabel());
    SgExpression      *ioTempVar;
    char              *inportName;
    int i;

    stmt = getPortStatement(inport);
    portStmt = isSgInportStmt(stmt);

    if (! portStmt) {
	inportName = symbName(portExp(inport));
	fm_assert(inportName);
	if (inportName) {
	    fm_error(s, "\
%s is not an inport", inportName);
	}
    } else {
	//
	// Encode receive function name
	//
	pname = portSetKeyEncode(stmt);

#if 0	
	receiveFuncStrLen = 3 + strlen(pname) + 8;
	receiveFuncStr = (char *)malloc(receiveFuncStrLen + 1);
	assert(receiveFuncStr);
	sprintf(receiveFuncStr, "fm_%s_receive", pname);
	//    printf("receive Func = %s\n", receiveFuncStr);
#endif /* 0 */
	receiveFuncFullStr = ppString("fm_", pname, 
				      (opt_debug)? "_rcv_d" : "_rcv");


	receiveMgFuncExp = isSgValueExp(exprMapGet(portMangleMap,receiveFuncFullStr));

	if (! receiveMgFuncExp) {
	    //
	    // generate a four digit send/receive name
	    // 
	    //   send:  fm_<f_name_base>_sndXXXX
	    //   recv:  fm_<f_name_base>_rcvXXXX
	    //
	    receiveFuncStr = ppString(ppString("fm_",currentFileBase,""),
				   (opt_debug) ? "_rcv_d" : "_rcv",
				   encodeStr(EC_MANGLE));

	    exprMapSet(portMangleMap, receiveFuncFullStr,
		       new SgValueExp(receiveFuncStr));
	} else {
	    receiveFuncStr = receiveMgFuncExp->stringValue();
	}

	receiveFuncSymb = new SgSymbol(PROCEDURE_NAME, receiveFuncStr);

	callStmt = new SgCallStmt(*receiveFuncSymb);
	callStmt->addArg(*inport);
	//    inport->unparsestdout();
	
	//
	// Consistency Check
	//
	//    Number of Arguments should match the number
	//    of types in the corresponding port declaration
	//    statement.
	//
	if (portStmt->numberOfPortTypes() != receiveStmt->numberOfArguments()) {
	    fm_error(s, "number of argument do not match number port types");
	}
	
	//
	// put type field which has a ref into
	// the portTypeMap and do some sanity checking too
	//
	portTypeMapClear();
	for (i = 0; i < receiveStmt->numberOfArguments(); i++) {
	    SgPortTypeExp *pte = portStmt->porttype(i);
	    SgExpression  *arg = receiveStmt->argument(i);
	    
	    if (pte->ref() && pte->type() && 
		pte->ref()->variant() != PORT_TYPE_OP &&
		pte->ref()->variant() != INPORT_TYPE_OP &&
		pte->ref()->variant() != OUTPORT_TYPE_OP &&
		pte->ref()->variant() != T_INPORT &&
		pte->ref()->variant() != T_OUTPORT) 
	    {
		SgSymbol *symb = (pte->ref())->symbol();
		
		portTypeMapSet(portTypeMap, symb->identifier(), arg);
	    }
	}
        //
        // type checking
        //
	for (i = 0; i < receiveStmt->numberOfArguments(); i++) {
	    SgPortTypeExp *pte = portStmt->porttype(i);
	    SgExpression  *arg = receiveStmt->argument(i);
	    int t1, t2;
	    
            matchPortAndArgTypes(i, pte, arg);
	}

        //
        // Generate arguments to the fm_ptype_receive() call
        //
	for (i = 0; i < receiveStmt->numberOfArguments(); i++) {
	    SgPortTypeExp *pte = portStmt->porttype(i);
	    SgExpression  *arg = receiveStmt->argument(i);
	    SgType        *t   = pte->type();
	    SgExpression  *e   = pte->ref();
	    
	    callStmt->addArg(*arg);
	    if (arrayRef = isSgArrayRefExp(e)) {
		SgExpression  *var_declared;
		SgExpression  *eSub; // = &(arraySizeExp(arrayRef));
		SgExpression  *realSize;
		
		if (arg->variant() != VAR_REF &&
		    arg->variant() != ARRAY_REF) 
		{
		    fm_warning(s,"\
Expects an array reference for argument %d but get an %s",
			       i+1, getVariantByName(arg->variant()));
		} else {
		    SgType *arg_t;
		    SgStatement *stmt;
		    SgVarDeclStmt *var_decl_stmt;
		    
		    stmt = stmtMapGet(prosMap, prosMapKeyEncode(currentProcedureName,symbName(arg)));
		    //
		    // if ARG is not defined then
		    // use the implicit type
		    //
		    if ( ! stmt ) {
			SgType *impl_t = getImplicitType(currentProcedureName,
						    arg->symbol());
			SgType *arg_t = arg->type();
			
			if (! impl_t) {
			    fm_error(s, "\
unknown  type for argument %s", 
				     symbName(arg));
			    return;
			}

			if (arg_t->variant() == T_ARRAY) {
			    SgArrayType *array_t = isSgArrayType(arg_t);
			    SgArrayRefExp *a;
			    int i;
			    
			    a = new SgArrayRefExp(*arg->symbol());
//			    printf("arg(%s) = %d\n", symbName(arg),
//				   array_t->dimension());
			    for (i = 0; i < array_t->dimension(); i++) {
				SgExpression *range;

				range = array_t->sizeInDim(i);
				a->addSubscript(*range);
			    }
			    stmt = new_decl_stmt(impl_t, a);
			} else {
			    stmt = new_decl_stmt(impl_t, arg);
			}
			stmtMapSet(prosMap, prosMapKeyEncode(currentProcedureName,symbName(arg)), stmt);
		    }
		    var_decl_stmt = isSgVarDeclStmt(stmt);
		    fm_assert(var_decl_stmt);
		    arg_t = var_decl_stmt->type();

		    var_declared = findSymbInDeclStmt(arg->symbol());
		    if (var_declared->variant() != ARRAY_REF) {
			//
			// check whether there is 
			// enough space to hold the data
			//
			SgExpression *ratio;
			int div;

			ratio = type_size_ratio(t,arg_t,&div);
			if (div > 0) {
			    fm_warning(s,"\
Argument %s has smaller size than the corresponding port type, this may cause potential receive overrun.",
				       symbName(arg));
			}
			callStmt->addArg(*c1);
		    } else {
			if (arg->variant() == ARRAY_REF) {
			    SgArrayRefExp *a;
			    SgExpression  *offset;
			    //
			    // size ratio of the argument
			    // against the protocol type
			    //
			    int div;
			    SgExpression  *p_a_ratio;
			    
			    a = isSgArrayRefExp(arg);
			    if (a->numberOfSubscripts() == 0) {
				offset = c1;
			    } else {
				offset = &(arraySizeExp(a));
			    }
			    eSub = &(arraySizeExp(isSgArrayRefExp(var_declared)));
			    p_a_ratio 
			      = type_size_ratio(t,arg_t, &div);
			    if (div >= 0) {
				eSub = &(eSub->copy()/(*p_a_ratio)-(*offset)+(*c1));
			    } else {
				eSub=&(eSub->copy()*(*p_a_ratio)-(*offset)+(*c1));
			    }
			} else {
			    eSub = &(arraySizeExp(isSgArrayRefExp(var_declared)));
			}
			if (! IsInteger(eSub)) {
			    callStmt->addArg(*c_minus_1);
			} else {
			    callStmt->addArg(*eSub);
			}
		    }
		}
	    } 
	}	

	ioTempVar = fmIOVarExp(ioStore);
	callStmt->addArg(*ioTempVar);

	if (opt_debug) {
	    SgValueExp *fileName = new SgValueExp(currentFile);
	    SgValueExp *lineNumber = new SgValueExp(s->lineNumber());

	    callStmt->addArg(*fileName);
	    callStmt->addArg(*lineNumber);
	}

	s->insertStmtBefore(*callStmt);
	if (endLabel) {
	    s->insertStmtBefore(*checkIOStat(0,ioStore,ioTempVar,endLabel));
	}
	s->insertStmtBefore(*checkIOStat(1,ioStore,ioTempVar,errLabel));
    }
    
    fmStmtSLL.append(s);
}

void doProbeStmt(SgStatement *s)
{
    SgProbeStmt *probe = isSgProbeStmt(s);
    SgCallStmt  *callStmt = new SgCallStmt(*fmProbeCallSymb);
    SgExpression *ioStore;
    SgExpression *ioTempVar;
    SgExpression *emptyTempVar;
    SgExpression *emptyVar;
    SgExpression *errLabel;
    char         *ident;

    callStmt->addArg(*portExp(probe->inport(0)));
    ioStore = probe->ioStore();
    ioTempVar = fmIOVarExp(ioStore);
    callStmt->addArg(*ioTempVar);

    errLabel = labelExp(probe->errLabel());

    if (! probe->emptyStore()) {
	fm_error(s, "\
must have EMPTY= specifier");
	return;
    }

    emptyVar = labelExp(probe->emptyStore());

    if (emptyVar->variant() == VAR_REF) {
	callStmt->addArg(*emptyVar);
    } else {
	ident = encodeStr(EC_EMPTY);
	SgSymbol *emptyTempSymb = new SgVariableSymb(ident, *integer_t);
	emptyTempSymb->declareTheSymbol(*currentProcedure);

	if (opt_use_auto_vars)
	  declare_auto_symb(ident, currentProcedure);

	emptyTempVar = new SgVarRefExp(*emptyTempSymb);
	callStmt->addArg(*emptyTempVar);
    }

    s->insertStmtBefore(*callStmt);

    s->insertStmtBefore(*checkIOStat(1,ioStore,ioTempVar,errLabel));

    if (emptyVar->variant() != VAR_REF) {
	s->insertStmtBefore(*checkIOStat(2,probe->ioStore(),
					 emptyTempVar, emptyVar));
    } 

    fmStmtSLL.append(s);
}



SgStatement *checkIOStat(int greaterThan,
			 SgExpression *ioStore, 
			 SgExpression *ioTempVar,
			 SgExpression *errORempty)
{
    SgLogIfStmt *logIfStmt;
    SgCallStmt  *callStmt;

    fm_assert(ioTempVar);

    if (! ioStore && ! errORempty) {
	callStmt = new SgCallStmt(*fmErrorMessageCallSymb);
	callStmt->addArg(*ioTempVar);
	return(callStmt);
    }
    
    if (errORempty) {
	SgExpression *e = labelExp(errORempty);
	SgLabel *label = (isSgLabelRefExp(e))->label();
	SgGotoStmt *gotoStmt  = new SgGotoStmt(*label);
	
	fm_assert(label);

	if (greaterThan == 1) {
	    logIfStmt = new SgLogIfStmt(*ioTempVar > *c0,
					*gotoStmt);
	} else if (greaterThan == 2) {
	    SgUnaryExp *cond = new SgUnaryExp(NOT_OP, *ioTempVar);

	    logIfStmt = new SgLogIfStmt(*cond, *gotoStmt);
	} else {
	    logIfStmt = new SgLogIfStmt(*ioTempVar < *c0,
					*gotoStmt);
	}
	return(logIfStmt);
    } else {
	callStmt = new SgCallStmt(*fmErrorMessageCallSymb);
	callStmt->addArg(*ioStoreExp(ioStore));
	return (callStmt);
    }
}
	
SgExpression *fmIOVarExp(SgExpression *ioStore)
{
    SgSymbol    *ioStoreSymb;
    SgVarRefExp *ioStoreRef;
    char        *ident;

    if (! ioStore) {
	ident = encodeStr(EC_IOVAL);
	ioStoreSymb = new SgVariableSymb(ident, *integer_t);
	ioStoreSymb->declareTheSymbol(*currentProcedure);
	
	if (opt_use_auto_vars) 
	  declare_auto_symb(ident, currentProcedure);

	ioStoreRef = new SgVarRefExp(*ioStoreSymb);
	return ( ioStoreRef );
    } else {
	return (ioStoreExp(ioStore));
    }
}

void matchPortAndArgTypes(int i, SgPortTypeExp *pte, SgExpression *arg)
{
    int t1, t2;
    SgType *type1, *type2;

    //
    // base type checking
    //
    
    type1 = pte->type();
    type2 = 0;

    //
    // change port variable to integer type
    //
    if (getPortVariant(arg) >= 0) {
	type2 = integer_t;
    } else {
	SgType *impl_t = 0;
	if (arg->variant() == VAR_REF ||
	    arg->variant() == ARRAY_REF) 
	{
	    SgVarDeclStmt *arg_decl;
	    arg_decl = isSgVarDeclStmt(stmtMapGet(prosMap,prosMapKeyEncode(currentProcedureName, symbName(arg))));
	    if (! arg_decl) 
	    {
		// not defined
		// use the implicit type
		impl_t = getImplicitType(currentProcedureName, arg->symbol());
	    } else {
		type2 = arg_decl->type();
	    }
	} 

	if (! type2)
	  type2 = arg->type();

	// Use the implicit type if there is a
	// disagreement between the sage type and
	// the implicit one
	if (impl_t) {
	    if (impl_t->variant() != type2->variant()) {
		type2 = impl_t;
	    }
	}
    }
    fm_assert(type1);
    fm_assert(type2);

    t1 = pte->type()->variant();
    t2 = arg->type()->variant();
    
    if (t1 != t2) {
//    if (! type1->equivalentToType(*type2)) {
	if (t2 == T_ARRAY) {
	    SgArrayType *array_t = isSgArrayType(arg->type());
	    SgType *base_t = array_t->baseType();
	    int t3 = base_t->variant();
	    
	    if (t3 != t1) {
		if (t3 == T_CHAR && t1 == T_STRING)
		{
		    ; 
		} else {
		    fm_warning(currentStatement, "\
argument %d base type is %s port type is %s", 
			     i+1, 
			     getTypeString(t3),
			     getTypeString(t1));
		}
	    } 
	} else if (t1 == ARRAY_REF && t2 == T_ARRAY ) {
            ;
        } else if (t1 == ARRAY_REF && t2 != T_ARRAY && isSgArrayRefExp(arg)) {
            ;
        } else if (t1 == T_DOUBLE && t2 == T_FLOAT && isSgValueExp(arg)) {
	    ;
	} else if (t1 == T_FLOAT && t2 == T_DOUBLE && isSgValueExp(arg)) {
	    ;
	} else if (t1 == T_INPORT) {
	    if (getPortVariant(arg) != INPORT_DECL) {
		fm_error(currentStatement, "\
expect an inport for argument %d but get a %s",
			 i+1,
			 getTypeCString(t2));
		return;
	    } 
	} else if (t1 == T_OUTPORT) {
	    if (getPortVariant(arg) != OUTPORT_DECL) {
		fm_error(currentStatement, "\
expect an outport for argument %d but get a %s",
			 i+1,
			 getTypeCString(t2));
		return;
	    }		
	} else  {
	    fm_warning(currentStatement, "\
argument %d type is %s, port type is %s", 
		       i+1,
		       getTypeString(t2),
		       getTypeString(t1));
//		       getVariantByName(t2),
//		       getVariantByName(t1));
	}
    }
}
	
	

