// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"

void initImplicitType(SgStatement *s)
{
    if (s->variant() != PROG_HEDR &&
	s->variant() != PROC_HEDR &&
	s->variant() != PROS_HEDR &&
	s->variant() != FUNC_HEDR   )
    {
	fm_internal_error("\
invoking initImplicitType from a invalid statement: %s", 
			  getVariantByName(s->variant()));
    }
    newImplTypeTemplate(s->symbol()->identifier());
}

void doImplicitStmt(SgStatement *s)
{
    SgImplicitStmt *implStmt = isSgImplicitStmt(s);
    SgStatement *proc = whichProcAmI(s);
    char *procName = proc->symbol()->identifier();
    int i, j;
    int ch;

    if (! implStmt->numberOfImplicitTypes()) {
	// implicit none
	deleteImplTypeTemplate(procName);
	return;
    }

    if (! typeMapGet(typeMap, typeMapKeyEncode(procName,'a'))) {
	newImplTypeTemplate(procName);
    }
    
    for (i = 0; i < implStmt->numberOfImplicitTypes(); i++) {
	SgType *impl_t = implStmt->implicitType(i);
	SgTypeExp *impl_e = new SgTypeExp(*impl_t);
	SgExprListExp *rangeList 
	  = isSgExprListExp(implStmt->implicitRangeList(i));

#if 0
	printf("type(%d) = ", i);
	printVariantName(impl_t->variant());
	printf("\n");
	printf("range list = ");
	printVariantName(rangeList->variant());
	printf("\n");
#endif /* 0 */
	if (! rangeList) {
	    fm_internal_error("\
Huh? implicit range list is not SgExprListExp but a %s",
			      getVariantByName(implStmt->implicitRangeList(i)->variant()));
	}
	
	for (j = 0; j < rangeList->length(); j++) {
	    SgExpression *e = rangeList->elem(j);

//	    printf("range specifier(%d) = ", j);
//	    printVariantName(e->variant());
//	    printf("\n");
	    
	    switch (e->variant()) {
	      case CHAR_VAL:
	    {
		SgValueExp *c_val = isSgValueExp(e);

		typeMapOverwrite(typeMap, 
				 typeMapKeyEncode(procName,
						  (int)c_val->charValue()),
				 impl_e);
		break;
	    }
	      case DDOT:
	      {
		SgSubscriptExp *sub = isSgSubscriptExp(e);
		SgValueExp *l_char = isSgValueExp(sub->lbound());
		SgValueExp *u_char = isSgValueExp(sub->ubound());
		
		if (! l_char || ! u_char) {
		    fm_internal_error("\
unknown subscript type for implicit statement at %s:%d", __FILE__, __LINE__);
		}
		for (ch =  (int)l_char->charValue();
		     ch <= (int)u_char->charValue();
		     ch++) 
		{
		    typeMapOverwrite(typeMap,
				     typeMapKeyEncode(procName,	(int)ch),
				     impl_e);
		}
		break;
	      }
	      default:
		fm_internal_error("\
unknown type for implicit specifier: %s", getVariantByName(e->variant()));
		break;
	    }
	}
    }
}

void exprMapOverwrite(StringSgExpressionAVLMap *map,char *s,SgExpression *e)
{
    String *key = new String(s);

    if (map->contains(*key)) {
	map->del(*key);
    }
    (*map)[*key] = *e;
}

void stmtMapOverwrite(StringSgStatementAVLMap *map,char *s,SgStatement *e)
{
    String *key = new String(s);

    if (map->contains(*key)) {
	map->del(*key);
    }
    (*map)[*key] = *e;
}


void deleteImplTypeTemplate(char *procName)
{
    for (int i = 'a'; i <= 'z'; i++) {
	String key(typeMapKeyEncode(procName, i));
	typeMap->del(key);
    }
}

void newImplTypeTemplate(char *procName)
{
    int i;
    for (i = 'a'; i <= 'h'; i++) {
	typeMapSet(typeMap, typeMapKeyEncode(procName, i), real_e);
    }
    for (i = 'i'; i <= 'n'; i++) {
	typeMapSet(typeMap, typeMapKeyEncode(procName, i), integer_e);
    }
    for (i = 'o'; i <= 'z'; i++) {
	typeMapSet(typeMap, typeMapKeyEncode(procName, i), real_e);
    }
}

char *typeMapKeyEncode(char *procName, int ch)
{
    int ECStrLen = strlen(procName) + 1;
    char *ECStr  = (char *)malloc(ECStrLen + 1);

    if (! ECStr) {
	fm_internal_error("\
Cannot allocate memory in %s:%d", __FILE__, __LINE__);
    }
    
    sprintf(ECStr, "%s%c", procName, ch);

    return(ECStr);
}

SgType * getImplicitType(char *procName, SgSymbol *symb)
{
    char *ident = symb->identifier();
    SgExpression *e 
      = typeMapGet(typeMap, typeMapKeyEncode(procName, (int)*ident));
    SgTypeExp *type_e;

    if (! e) { return ((SgType *)0); }
    
    type_e = isSgTypeExp(e);
    if (! type_e) {
	fm_internal_error("\
Huh? the contents of typeMap is not a SgTypeExp!");
    }

    return(type_e->type());
}

char *getTypeString(int type_variant)
{
    char *t_string;

    switch (type_variant) {
      case T_INT:
	t_string = strdup("integer");
	break;
      case T_SHORT:
	t_string = strdup("short");
	break;
      case T_FLOAT:
	t_string = strdup("real");
	break;
      case T_DOUBLE:
	t_string = strdup("double");
	break;
      case T_CHAR:
	t_string = strdup("char");
	break;
      case T_STRING:
	t_string = strdup("char");
	break;
      case T_ARRAY:
	t_string = strdup("array");
	break;
      case T_COMPLEX:
	t_string = strdup("complex");
	break;
      case T_DCOMPLEX:
	t_string = strdup("doublecomplex");
	break;
      case T_INPORT:
	t_string = strdup("integer");
	break;
      case T_OUTPORT:
	t_string = strdup("integer");
	break;
      case T_BOOL:
	t_string = strdup("logical");
	break;
      default:
	// fm_internal_error()
	fprintf(stderr, "dont know what to do for type (%s)\n",
		getVariantByName(type_variant));
	t_string = strdup("unknown_type");
    }
    return t_string;
}

char *getTypeCString(int type_variant)
{
    char *t_string;

    switch (type_variant) {
      case T_INT:
      case T_BOOL:
	t_string = strdup("int");
	break;
      case T_SHORT:
	t_string = strdup("short");
	break;
      case T_FLOAT:
	t_string = strdup("float");
	break;
      case T_DOUBLE:
	t_string = strdup("double");
	break;
      case T_CHAR:
	t_string = strdup("char");
	break;
      case T_STRING:
	t_string = strdup("char");
	break;
      case T_COMPLEX:
	t_string = strdup("complex");
	break;
      case T_DCOMPLEX:
	t_string = strdup("doublecomplex");
	break;
      case T_INPORT:
	t_string = strdup("inport");
	break;
      case T_OUTPORT:
	t_string = strdup("outport");
	break;
      default:
	// fm_internal_error()
	fprintf(stderr, "dont know what to do for type (%s)\n",
		getVariantByName(type_variant));
	t_string = strdup("unknown_type");
    }
    return t_string;
}

char *getTypePrintfFormat(int type_variant)
{
    char *t_string;

    switch (type_variant) {
      case T_INT:
      case T_SHORT:
      case T_INPORT:
      case T_OUTPORT:
      case T_BOOL:
	t_string = strdup("%d");
	break;
      case T_FLOAT:
	t_string = strdup("%f");
	break;
      case T_DOUBLE:
	t_string = strdup("%lf");
	break;
      case T_CHAR:
	t_string = strdup("%c");
	break;
      case T_STRING:
	t_string = strdup("%s");
	break;
      case T_COMPLEX:
	t_string = strdup("%f");
	break;
      case T_DCOMPLEX:
	t_string = strdup("%lf");
	break;
      default:
	// fm_internal_error()
	fprintf(stderr, "dont know what to do for type (%s)\n",
		getVariantByName(type_variant));
	t_string = strdup("unknown_type");
    }
    return t_string;
}



int type_size(SgType *t)
{
    int size;
    switch (t->variant()) {
      case T_INT:
      case T_INPORT:
      case T_OUTPORT:
      case T_BOOL:
	size = sizeof(int);
	break;
      case T_SHORT:
	size = sizeof(short);
	break;
      case T_FLOAT:
	size = sizeof(float);
	break;
      case T_DOUBLE:
	size = sizeof(double);
	break;
      case T_CHAR:
	size = sizeof(char);
	break;
      case T_STRING:
	size = sizeof(char);
	break;
      case T_COMPLEX:
	size = 2 * sizeof(float);
	break;
      case T_DCOMPLEX:
	size = 2 * sizeof(double);
	break;
      default:
	fm_internal_error("\
dont know what to do for type (%s)\n",
		getVariantByName(t->variant()));
	size = 0;
    }
    return size;
}

SgExpression *type_size_ratio(SgType *t1, SgType *t2, int *div)
{
    int size1, size2;
    int size_ratio;
    SgExpression *the_size_ratio;

    if (!t1 || ! t2) {
	fm_internal_error("\
Expect an non-null type at %s:%d",
			  __FILE__, __LINE__);
    }

    size1 = type_size(t1);
    size2 = type_size(t2);

    if (size1 == size2) {
	*div = 0;
    } else if (size1 > size2) {
	*div = 1;
    } else {
	*div = -1;
    }
    size_ratio = (size1 >= size2) ? (size1 / size2) : (size2 / size1);

    the_size_ratio = new SgValueExp(size_ratio);

    return the_size_ratio;
}


char *num_to_string(int n)
{
    char str[1024];

    sprintf(str, "%d", n);
    return(strdup(str));
}
