/**********************************************************************
$Id: global.h,v 1.12 1995/06/14 04:41:02 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/global.h,v $

$Log: global.h,v $
Revision 1.12  1995/06/14  04:41:02  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.11  1995/05/12  20:26:18  tuecke
Updated to new Nexus.
Added AIX shared library support:
	- initwrap.c is necessary on AIX to avoid undefined external symbols
	- Moved sentinel code into here from ../sentinel

Revision 1.10  1995/04/17  06:30:11  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.9  1995/04/06  06:32:15  kohr
Added documentation, cleaned up a bit.


FM runtime global variable declarations.
**********************************************************************/


#ifndef _fm_global_h
#define _fm_global_h

/*
The macro ALLOC_GLOBALS determines whether the declarations in this file
are "extern" or not (hence whether or not they imply allocating storage).
*/

#ifdef ALLOC_GLOBALS
#define FM_GLOBAL
#else
#define FM_GLOBAL extern
#endif


/*
_fm_no_threads is set to true if the -nothreads option is given. This
forces all FM processes to go into separate contexts.
*/

FM_GLOBAL fm_bool_t _fm_no_threads;


/*
_fm_is_node_context is set to true if this (unix) process is the original
FM node context. It will be false for (unix) processes that were created
with nexus_create_context() for the purposes of nonshared (FM) processes.
*/

FM_GLOBAL fm_bool_t _fm_is_node_context;


/*
_fm_linktime_init_func is a function pointer to _fm_linktime_init(),
which is created during linking by the fm compiler.  It is set
in initwrap.c and used in init.c.  It is necessary on the IBM
because shared libraries cannot have unresolved external symbols.
Resolving this in initwrap.c allows that file to be the only statically
linked file.
*/
FM_GLOBAL void (*_fm_linktime_init_func)();


/*
Debugging stuff...
*/

#ifdef FM_DEBUG
#ifndef ALLOC_GLOBALS

extern fm_subsystem_debug_entry_t _fm_subsystem_debug_table[];

#else /* ALLOC_GLOBALS */

#undef FM_SUBSYSTEM

#define FM_SUBSYSTEM(option_key, desc, id)				\
	{ FM_SUBSYSTEM_ ## id, 0, #option_key, desc },

fm_subsystem_debug_entry_t _fm_subsystem_debug_table[] =
{
#include "subsystems.def"
	{-1, -1, (char *) NULL, (char *) NULL}
};

#endif /* ALLOC_GLOBALS */
#endif /* FM_DEBUG */


/*
Instrumentation stuff...
*/

#ifdef FM_SIMPLE_INST
FM_GLOBAL fm_bool_t _fm_simple_inst;
#ifdef ALLOC_GLOBALS
char _fm_simple_inst_log_dir[1024];
#else
extern char _fm_simple_inst_log_dir[];
#endif
#endif


#endif /* _fm_global_h */
