/**********************************************************************
$Id: init.c,v 1.26 1995/06/14 04:41:03 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/init.c,v $

$Log: init.c,v $
Revision 1.26  1995/06/14  04:41:03  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.25  1995/05/12  20:26:21  tuecke
Updated to new Nexus.
Added AIX shared library support:
	- initwrap.c is necessary on AIX to avoid undefined external symbols
	- Moved sentinel code into here from ../sentinel

Revision 1.24  1995/04/17  06:30:13  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.23  1995/04/06  06:32:17  kohr
Added documentation, cleaned up a bit.


FM runtime startup code.
**********************************************************************/


		/* Causes globals to be allocated in this .o file. */
#define ALLOC_GLOBALS
#include "fm.h"

#ifdef FM_SIMPLE_INST
#include "simple_inst.h"
#endif

#ifdef FM_SENTINEL_SUPPORT
#include "sentinel.h"
#endif

#include <errno.h>
#include <sys/file.h>

#ifdef NEXUS_ARCH_AIX
#include <fcntl.h>
#endif

#ifdef NeXT
#include <libc.h>
#endif

static char *logfile_fmt;


/*
 * _fm_package_args_init()
 */
int _fm_package_args_init(int *argc, char ***argv)
{
    int arg_num;
    int val;
    int i;
    
#ifdef FM_DEBUG
    _fm_pause_init();
#endif

    logfile_fmt = (char *) NULL;
    _fm_no_threads = NEXUS_FALSE;

#ifdef FM_SIMPLE_INST
    _fm_simple_inst = NEXUS_FALSE;
    strcpy(_fm_simple_inst_log_dir, ".");
#endif

#ifdef FM_DEBUG
    if ((arg_num = nexus_find_argument(argc, argv, "Dfm", 2)) >= 0)
    {
	val = atoi((*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
	for (i = 0; i < (int) FM_SUBSYSTEM_LAST; i++)
	{
	    _fm_subsystem_debug_table[i].debug_level = val;
	}
    }

    for (i = 0; i < (int) FM_SUBSYSTEM_LAST; i++)
    {
	if ((arg_num = nexus_find_argument(argc, argv,
					   _fm_subsystem_debug_table[i].option,
					   2)) >= 0)
	{
	    _fm_subsystem_debug_table[i].debug_level
		= atoi((*argv)[arg_num + 1]);
	    nexus_remove_arguments(argc, argv, arg_num, 2);
	}
    }
#endif /* FM_DEBUG */
    
#ifdef O_WRONLY
    if ((arg_num = nexus_find_argument(argc, argv, "log", 2)) >= 0)
    {
	char path[1024];
	int fd;
	FILE *fp;
	extern FILE *_nx_stdout;
	extern int errno;
	extern char *sys_errlist[];

	logfile_fmt = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);
	
	for (i = 0; ; i++)
	{
	    sprintf(path, logfile_fmt, i);
	    if ((fd = open(path, O_WRONLY | O_CREAT | O_EXCL, 0664)) >= 0)
	    {
		fp = fdopen(fd, "w");
		nexus_printf("Logging stdout to %s\n", path);
		fflush(_nx_stdout);
		_nx_stdout = fp;
		setbuf(fp, NULL);
		break;
	    }
	    else if (errno != EEXIST)
	    {
		nexus_fatal("Cannot write logfile %s: %s\n",
			    path, sys_errlist[errno]);
	    }
	}
    }
#endif /* O_WRONLY */

#ifdef FM_DEBUG
    if ((arg_num = nexus_find_argument(argc, argv, "pause", 2)) >= 0)
    {
	_fm_add_pausepoint(atoi((*argv)[arg_num + 1]));
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "sleep", 3)) >= 0)
    {
	_fm_add_sleeppoint(atoi((*argv)[arg_num + 1]),
			   atoi((*argv)[arg_num + 2]));
	nexus_remove_arguments(argc, argv, arg_num, 3);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "show_pausepoints",1)) >= 0)
    {
	_fm_set_show_pausepoints(NEXUS_TRUE);
	nexus_remove_arguments(argc, argv, arg_num, 1);
    }
#endif /* FM_DEBUG */
#ifdef FM_SIMPLE_INST
    if ((arg_num = nexus_find_argument(argc, argv, "simple_inst", 1)) >= 0)
    {
	_fm_simple_inst = NEXUS_TRUE;
	nexus_remove_arguments(argc, argv, arg_num, 1);
    }
    if ((arg_num = nexus_find_argument(argc,argv,"simple_inst_log_dir",2)) >=0)
    {
	strcpy(_fm_simple_inst_log_dir, (*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
#endif /* FM_SIMPLE_INST */
	
    if ((arg_num = nexus_find_argument(argc, argv, "no_threads", 1)) >= 0)
    {
	_fm_no_threads = NEXUS_TRUE;
	nexus_remove_arguments(argc, argv, arg_num, 1);
    }
    
#ifdef FM_SENTINEL_SUPPORT
    sentinel_args_init(argc, argv);
#endif

    return (0);
} /* _fm_package_args_init() */


/*
 * _fm_usage_message()
 */
void _fm_usage_message()
{
    int i;
    char temp[100];
    printf("Usage message for FM...\n");

    printf("    -no_threads               : "
           "Locate all FM processes in their own context\n");
#ifdef FM_SIMPLE_INST
    printf("    -simple_inst              : "
           "Enable simple instrumentation library\n");
#endif
#ifdef FM_DEBUG
    for (i = 0; i < (int) FM_SUBSYSTEM_LAST; i++)
    {
	sprintf(temp, "%s <integer>", _fm_subsystem_debug_table[i].option);
	    
	printf("    -D%-24s: Debug level for %s\n",
	       temp,
	       _fm_subsystem_debug_table[i].description);
    }
    printf("    -Dfm <integer>            : "
           "Set all FM debugging flags to <integer>\n");
    printf("    -pause <id>               : "
           "Pause at pausepoint <id>\n");
    printf("    -sleep <id> <n>           : "
           "Sleep for <n> seconds at pausepoint <id>\n");
    printf("    -show_pausepoints         : "
           "Print a message when each pausepoint is \n");
    printf("                                executed\n");
#endif /* FM_DEBUG */
#ifdef FM_SENTINEL_SUPPORT
    sentinel_usage_message();
#endif
	       
} /* _fm_usage_message() */


/*
 * _fm_new_process_params()
 */
int _fm_new_process_params(char *buf, int size)
{
    int i;
    char temp[200];
    int n;

#ifdef FM_DEBUG
    n = 0;
    
    for (i = 0; i < (int) FM_SUBSYSTEM_LAST; i++)
    {
	if (_fm_subsystem_debug_table[i].debug_level > 0)
	{
	    sprintf(temp, "-D%s %d ", 
		    _fm_subsystem_debug_table[i].option,
		    _fm_subsystem_debug_table[i].debug_level);
	    if (n + strlen(temp) <= size)
	    {
		strcat(buf, temp);
		n += strlen(temp);
	    }
	}
    }
#endif

    if (logfile_fmt != (char *) NULL)
    {
	sprintf(temp, " -log %s", logfile_fmt);
	if (n + strlen(temp) <= size)
	{
	    strcat(buf, temp);
	    n += strlen(temp);
	}
    }
#ifdef FM_DEBUG
    n = _fm_pause_params(n, buf, size);
#endif
#ifdef FM_SENTINEL_SUPPORT
    n = sentinel_params(n, buf, size);
#endif

    if (_fm_no_threads)
    {
	strcpy(temp, " -no_threads");
	if (n + strlen(temp) <= size)
	{
	    strcat(buf, temp);
	    n += strlen(temp);
	}
    }
    
#ifdef FM_SIMPLE_INST
    if (_fm_simple_inst)
    {
	strcpy(temp, " -simple_inst");
	if (n + strlen(temp) <= size)
	{
	    strcat(buf, temp);
	    n += strlen(temp);
	}
	sprintf(temp, " -simple_inst_log_dir %s", _fm_simple_inst_log_dir);
	if (n + strlen(temp) <= size)
	{
	    strcat(buf, temp);
	    n += strlen(temp);
	}
    }
#endif /* FM_SIMPLE_INST */

    if (n + 1 <= size)
    {
       strcat(buf, " ");
       n++;
    }
 
    return (n);
} /* _fm_new_process_params() */


/*
 * NexusExit()
 */
void NexusExit()
{
#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("NexusExit()\n");
#endif
} /* NexusExit() */


/*
 * NexusUnknownHandler()
 */
void NexusUnknownHandler(void *address, nexus_buffer_t *buffer,
			 char *handler_name, int handler_id)
{
    nexus_fatal("NexusUnknownHandler(): handler_name=%s, handler_id=%d\n",
		handler_name, handler_id);
} /* NexusUnknownHandler() */


static nexus_handler_t system_handlers[] =
{ \
  {"NexusExit",
       NEXUS_EXIT_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NexusExit},
  {"NexusUnknownHandler",
       NEXUS_UNKNOWN_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NexusUnknownHandler},
  {FM_VC_INIT_NAME, FM_VC_INIT_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_vc_init_handler},
  {FM_VC_SHUTDOWN_CONTEXTS_NAME, FM_VC_SHUTDOWN_CONTEXTS_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_vc_shutdown_contexts_handler},
  {FM_VC_SHUTDOWN_CONTEXTS_ACK_NAME, FM_VC_SHUTDOWN_CONTEXTS_ACK_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_vc_shutdown_contexts_ack_handler},
  {FM_SHOULD_SHARE_RETURN_NAME, FM_SHOULD_SHARE_RETURN_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_should_share_return_handler},
  {FM_CREATE_PROCESS_NAME, FM_CREATE_PROCESS_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_create_process_handler},
  {FM_CREATE_PROCESS_RETURN_NAME, FM_CREATE_PROCESS_RETURN_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_create_process_return_handler},
  {FM_PORT_SEND_NAME, FM_PORT_SEND_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_port_send_handler},
  {FM_PORT_SEND_SETUP_NAME, FM_PORT_SEND_SETUP_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_port_send_setup_handler},
  {FM_PORT_SEND_SETUP_RETURN_NAME, FM_PORT_SEND_SETUP_RETURN_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_port_send_setup_return_handler}, 
  {FM_PROCESS_RETURN_NAME, FM_PROCESS_RETURN_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_process_return_handler},
  {FM_CHANNEL_MIGRATING_NAME, FM_CHANNEL_MIGRATING_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_channel_migrating_handler},
  {FM_CHANNEL_FORWARD_BUFFER_NAME, FM_CHANNEL_FORWARD_BUFFER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_channel_forward_buffer_handler},
  {FM_CHANNEL_MIGRATING_ACK_NAME, FM_CHANNEL_MIGRATING_ACK_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_channel_migrating_ack_handler},
  {FM_OUTPORT_MIGRATING_NAME, FM_OUTPORT_MIGRATING_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_outport_migrating_handler},
  {FM_OUTPORT_MIGRATING_ACK_NAME, FM_OUTPORT_MIGRATING_ACK_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_outport_migrating_ack_handler},
  {FM_FWD_CHANNEL_MIGRATING_NAME, FM_FWD_CHANNEL_MIGRATING_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_fwd_channel_migrating_handler},
  {FM_FWD_CHANNEL_MIGRATING_ACK_NAME, FM_FWD_CHANNEL_MIGRATING_ACK_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_fwd_channel_migrating_ack_handler},
  {FM_FWD_OUTPORT_MIGRATING_NAME, FM_FWD_OUTPORT_MIGRATING_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_fwd_outport_migrating_handler},
  {FM_FWD_OUTPORT_MIGRATING_ACK_NAME, FM_FWD_OUTPORT_MIGRATING_ACK_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_fwd_outport_migrating_ack_handler},
  {FM_END_CHANNEL_NAME, FM_END_CHANNEL_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_end_channel_handler},
  {FM_END_CHANNEL_ACK_NAME, FM_END_CHANNEL_ACK_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _fm_end_channel_ack_handler},
  {(char *) NULL,
       0,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NULL},
};

/*
 * NexusBoot()
 */
int NexusBoot()
{
    int rc;
    
    nexus_register_handlers(system_handlers);
    _fm_ports_module_init();
    (*_fm_linktime_init_func)();
    _fm_queue_module_init();
    _fm_util_init();

#ifdef FM_SENTINEL_SUPPORT
    sentinel_init();
#endif

#ifdef FM_SIMPLE_INST
    si_init();
#endif
    
    rc = 0;
    
    return (rc);
} /* NexusBoot() */


int _fm_real_init(nexus_node_t *nodes, int n_nodes)
{
    int n_node_gps;
    nexus_global_pointer_t *node_gps;
    int i, j;
    
    /*
     * For now, give the VC stuff a list of the unique
     * global pointers. (It will likely get confused if it
     * is given multiple nodes on the same context).
     */

    n_node_gps = 0;
    for (i = 0; i < n_nodes; i++)
    {
	if (nodes[i].return_code == NEXUS_NODE_NEW)
	    n_node_gps++;
    }
    NexusMalloc(FM_INIT(), node_gps, nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t) * n_node_gps);

    for (i = 0, j = 0; i < n_nodes; i++)
    {
	if (nodes[i].return_code == NEXUS_NODE_NEW)
	{
	    node_gps[j] = nodes[i].gp;
	    j++;
	}
	    
    }

    _fm_is_node_context = FALSE;
    _fm_vc_alloc();
    _fm_vc_init(node_gps, n_node_gps);

    NexusFree(node_gps);
    
    return n_node_gps > 0;
}

void FM_START()
{
    nexus_start();
#ifdef FM_SIMPLE_INST
    if (_fm_simple_inst)
	si_program_init();
#endif
}


/*
 * FM_SHUTDOWN()
 *
 * Close down the FM computation.
 */
void FM_SHUTDOWN()
{
    _fm_vc_shutdown_contexts();
    nexus_destroy_current_context(NEXUS_FALSE);
} /* FM_SHUTDOWN */



/*
 * FM_STOP()
 * 
 * Handle a fortran STOP statement.
 * 
 */
void FM_STOP()
{
    nexus_printf("Fortran STOP\n");
    nexus_exit(0, 1);
} /* FM_STOP() */


/*
 * FM_POLL()
 * 
 * Just invoke nexus_poll() to handle any messages that may be ready.
 * 
 */
void FM_POLL()
{
#ifdef FM_SIMPLE_INST
    if (_fm_simple_inst)
	si_sys_counter_increment(SI_SYS_COUNTER_FM_POLL);
#endif
    nexus_poll();
} /* FM_POLL() */



/*
 * FM_YIELD()
 * 
 * Yield the processor.
 * 
 */
void FM_YIELD()
{
#ifdef FM_SIMPLE_INST
    if (_fm_simple_inst)
	si_sys_counter_increment(SI_SYS_COUNTER_FM_YIELD);
#endif
    nexus_thread_yield();
} /* FM_YIELD() */
