/**********************************************************************
$Id: pause.c,v 1.7 1995/06/14 04:41:07 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/pause.c,v $

$Log: pause.c,v $
Revision 1.7  1995/06/14  04:41:07  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.6  1995/04/17  06:30:18  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.5  1995/04/06  06:32:22  kohr
Added documentation, cleaned up a bit.


FM runtime support for user-defined pause- and sleep-points.
**********************************************************************/


#include "fm.h"

#ifdef FM_DEBUG

#ifdef NeXT
#include <libc.h>
#endif

static	fm_pausepoint_t _fm_pausepoints;
static	fm_bool_t	_fm_show_pausepoints;


void _fm_pause_init()
{
    _fm_pausepoints = (fm_pausepoint_t) NULL;
    _fm_show_pausepoints = NEXUS_FALSE;
}


void _fm_set_show_pausepoints(fm_bool_t val)
{
    _fm_show_pausepoints = val;
}


static fm_pausepoint_t lookup_pausepoint(int id)
{
    fm_pausepoint_t ppoint;

    for (ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next)
    {
	if (ppoint->id == id)
	    return ppoint;
    }
    return (fm_pausepoint_t) NULL;
}


void _fm_sizeof_pausepoints(nexus_buffer_t *buf, int *size, int *count)
{
    fm_pausepoint_t ppoint;
    int n;

    for (n = 0, ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next, n++)
	;

    /* For the number of pausepoints */
    *size = nexus_sizeof_int(buf, n * 3 + 1);
    *count = n * 3 + 1;
}


int _fm_pause_params(int n, char *buf, int size)
{
    fm_pausepoint_t ppoint;
    char temp[100];

    for (ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next)
    {
        if (ppoint->pause_me)
        {
            sprintf(temp, " -pause %d", ppoint->id);
            if (n + strlen(temp) <= size)
            {
                strcat(buf, temp);
                n += strlen(temp);
            }
        }
        if (ppoint->sleep_me > 0)
        {
            sprintf(temp, " -sleep %d %d", ppoint->id, ppoint->sleep_me);
            if (n + strlen(temp) <= size)
            {
                strcat(buf, temp);
                n += strlen(temp);
            }
        }
    }
    if (_fm_show_pausepoints)
    {
        strcpy(temp, " -show_pausepoints");
        if (n + strlen(temp) <= size)
        {
            strcat(buf, temp);
            n += strlen(temp);
        }
    }
    
    return n;
}


void _fm_put_pausepoints(nexus_buffer_t *buf)
{
    fm_pausepoint_t ppoint;
    int n;

    for (n = 0, ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next, n++)
	;

    nexus_put_int(buf, &n, 1);

    for (ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next)
    {
	nexus_put_int(buf, &ppoint->id, 1);
	nexus_put_int(buf, &ppoint->pause_me, 1);
	nexus_put_int(buf, &ppoint->sleep_me, 1);
    }
}


void _fm_get_pausepoints(nexus_buffer_t *buf)
{
    int n;
    int i;
    fm_pausepoint_t ppoint;

    nexus_get_int(buf, &n, 1);

    for (i = 0; i < n; i++)
    {
	NexusMalloc(_fm_get_pausepoints(), ppoint,
		    fm_pausepoint_t, sizeof( struct _fm_pausepoint_t));

	nexus_get_int(buf, &ppoint->id, 1);
	nexus_get_int(buf, &ppoint->pause_me, 1);
	nexus_get_int(buf, &ppoint->sleep_me, 1);

	ppoint->next = _fm_pausepoints;
	_fm_pausepoints = ppoint;
    }
}


void _fm_add_pausepoint(int id)
{
    fm_pausepoint_t ppoint;

    ppoint = lookup_pausepoint(id);

    if (ppoint == (fm_pausepoint_t) NULL)
    {
	NexusMalloc(_fm_add_pausepoint(), ppoint,
		    fm_pausepoint_t, sizeof( struct _fm_pausepoint_t));

	ppoint->id = id;
	ppoint->pause_me = NEXUS_FALSE;
	ppoint->sleep_me = 0;

	ppoint->next = _fm_pausepoints;
	_fm_pausepoints = ppoint;
    }
    
    ppoint->pause_me = NEXUS_TRUE;
    
}


void _fm_add_sleeppoint(int id, int secs)
{
    fm_pausepoint_t ppoint;

    ppoint = lookup_pausepoint(id);

    if (ppoint == (fm_pausepoint_t) NULL)
    {
	NexusMalloc(_fm_add_sleeppoint(), ppoint,
		    fm_pausepoint_t, sizeof( struct _fm_pausepoint_t));

	ppoint->id = id;
	ppoint->pause_me = NEXUS_FALSE;
	ppoint->sleep_me = 0;

	ppoint->next = _fm_pausepoints;
	_fm_pausepoints = ppoint;
    }
    
    ppoint->sleep_me = secs;
}


void FM_PAUSE(f_integer_t *f_id, f_character_t f_msg CHAR_LEN_DECL(f_msg_len))
{
    char msg[2048];
    int id;
    LOCAL_CHAR_LEN_DECL(f_msg_len)

    CONVERT_CHAR_FTN_TO_C_INPLACE(f_msg, f_msg_len, msg);

    id = (int) *f_id;

    _fm_pause(id, msg);
}


void _fm_pause(int id, char *msg)
{
    fm_pausepoint_t ppoint;
    int done_pausing = 0;

    ppoint = lookup_pausepoint(id);

    if (_fm_show_pausepoints)
    {
	nexus_printf("Pausepoint %d: %s\n", id, msg);
    }
    
    if (ppoint != (fm_pausepoint_t) NULL)
    {
	if (ppoint->pause_me)
	{
	    nexus_printf("Pausing at pausepoint %d: %s\n", id, msg);
	    while (!done_pausing)
	    {
                
		pause();
	    }
	    nexus_printf("Continuing from pausepoint %d\n", id);
	}
	else if (ppoint->sleep_me > 0)
	{
	    nexus_printf("Sleeping %d at pausepoint %d: %s\n", ppoint->sleep_me, id, msg);
	    sleep(ppoint->sleep_me);
	    nexus_printf("Continuing from pausepoint %d\n", id);
	}
    }
}


#else /* FM_DEBUG */

void
FM_PAUSE(f_integer_t *f_id, f_character_t f_msg CHAR_LEN_DECL(f_msg_len))
{
}


#endif /* FM_DEBUG */
