/**********************************************************************
$Id: ports.c,v 1.16 1995/04/17 21:48:05 mqxu Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/ports.c,v $

$Log: ports.c,v $
Revision 1.16  1995/04/17  21:48:05  mqxu
64-bit incompatible bugs fixed

Revision 1.15  1995/04/17  06:30:26  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.14  1995/04/06  06:32:30  kohr
Added documentation, cleaned up a bit.


FM runtime support for assorted port/merger/channel operations.
**********************************************************************/


#include "fm.h"

static void disconnect_inport_channel(fm_inport_t *inport);
static void disconnect_outport_channel(fm_outport_t *outport);

static void init_inport(fm_inport_t *ip);
static void init_outport(fm_outport_t *op);
static void init_channel(fm_channel_t *c);
static void reinit_inport(fm_inport_t *ip);
static void reinit_outport(fm_outport_t *op);
static void reinit_channel(fm_channel_t *c);

/*
FM_INIT_INPORT() : Initialize an inport array.
*/

void FM_INIT_INPORT(f_integer_t *f_inports,
		    f_integer_t *f_n_inports)
{
    int n_inports;
    int i;
    fm_inport_t *ip;

    n_inports = (int) *f_n_inports;

    NexusAssert(n_inports > 0);

    for (i = 0; i < n_inports; i++)
    {
	ip = _fm_alloc_inport();

	f_inports[i] = (f_integer_t) ip;
    }
    
} /* FM_INIT_INPORT */


#ifdef FM_DEBUG

/*
FM_INIT_INPORT_DEBUG() : Initialize an inport array.
*/

void FM_INIT_INPORT_DEBUG(f_integer_t *f_inports,
			  f_integer_t *f_n_inports,
			  f_character_t port_name
			  CHAR_LEN_DECL(port_name_len))
{
    int n_inports;
    int i;
    fm_inport_t *ip;
    LOCAL_CHAR_LEN_DECL(port_name_len)
    char buf[120];

    CONVERT_CHAR_FTN_TO_C_INPLACE(port_name, port_name_len, buf);

    FM_INIT_INPORT(f_inports, f_n_inports);

    n_inports = (int) *f_n_inports;

    NexusAssert(n_inports > 0);

    for (i = 0; i < n_inports; i++)
    {
	ip = (fm_inport_t *) f_inports[i];

	sprintf(ip->name, "%s(%d)", buf, i);
    }
    
} /* FM_INIT_INPORT */

#endif /* FM_DEBUG */


/*
FM_INIT_OUTPORT() : Initialize an outport array.
*/

void FM_INIT_OUTPORT(f_integer_t *f_outports,
		     f_integer_t *f_n_outports)
{
    int n_outports;
    int i;
    fm_outport_t *op;

    n_outports = (int) *f_n_outports;

    NexusAssert(n_outports > 0);

    for (i = 0; i < n_outports; i++)
    {
	op = _fm_alloc_outport();

	f_outports[i] = (f_integer_t) op;
    }
    
} /* FM_INIT_OUTPORT */


#ifdef FM_DEBUG

/*
FM_INIT_OUTPORT_DEBUG() : Initialize an outport array.
*/

void FM_INIT_OUTPORT_DEBUG(f_integer_t *f_outports,
			   f_integer_t *f_n_outports,
			   f_character_t port_name
			   CHAR_LEN_DECL(port_name_len))
{
    int n_outports;
    int i;
    fm_outport_t *op;
    LOCAL_CHAR_LEN_DECL(port_name_len)
    char buf[120];

    CONVERT_CHAR_FTN_TO_C_INPLACE(port_name, port_name_len, buf);

    FM_INIT_OUTPORT(f_outports, f_n_outports);

    n_outports = (int) *f_n_outports;

    NexusAssert(n_outports > 0);

    for (i = 0; i < n_outports; i++)
    {
	op = (fm_outport_t *) f_outports[i];

	sprintf(op->name, "%s(%d)", buf, i);
    }
    
} /* FM_INIT_OUTPORT */

#endif /* FM_DEBUG */


/*
FM_CHANNEL() : Create a channel between two ports.
*/

void FM_CHANNEL(f_integer_t *f_inport,
		f_integer_t *f_outport,
		void (*init_func)(fm_channel_t *),
		f_integer_t *f_ioval)
{
    fm_inport_t *inport;
    fm_outport_t *outport;
    fm_channel_t *channel;

    inport = (fm_inport_t *) *f_inport;
    outport = (fm_outport_t *) *f_outport;

    InportMagicCheck(inport);
    OutportMagicCheck(outport);

    channel = _fm_alloc_channel();
    _fm_set_channel_size(channel, 1);

    /* 
     * We must check to see if the inport and outport are already
     * connected to something. If they are, disconnect them (sending
     * an out-of-scope message to the port's channel.
     */

    if (InportState(inport) != FM_STATE_UNDEFINED)
/*	disconnect_inport_channel(inport); */
        nexus_fatal("\
fm_channel(): inport %s is already connected to an outport\n",
#ifdef FM_DEBUG
		    _fm_inport_string(inport));
#else
                    "<noname>");
#endif /* FM_DEBUG */


    if (OutportState(outport) != FM_STATE_UNDEFINED)
/*	disconnect_outport_channel(outport); */
        nexus_fatal("\
fm_channel(): outport %s is already connected to an inport\n",
#ifdef FM_DEBUG
		    _fm_outport_string(outport));
#else
                    "<noname>");
#endif /* FM_DEBUG */

    NexusAssert(InportState(inport) == FM_STATE_UNDEFINED);
    NexusAssert(nexus_is_null_global_pointer(&inport->channel_gp));
		
    NexusAssert(OutportState(outport) == FM_STATE_UNDEFINED);
    NexusAssert(nexus_is_null_global_pointer(&outport->channel_gp));

    nexus_global_pointer(&channel->inport, (void *) inport);
    nexus_global_pointer(&(channel->outports[0]), (void *) outport);
    channel->n_outports = 1;

    nexus_global_pointer(&inport->channel_gp, (void *) channel);
    nexus_global_pointer(&outport->channel_gp, (void *) channel);

    ChannelSetState(channel, FM_STATE_NORMAL);
    InportSetState(inport, FM_STATE_NORMAL);
    OutportSetState(outport, FM_STATE_NORMAL);

    NexusAssert(init_func != (void (*)(fm_channel_t *)) NULL);

    channel->send_func = _fm_channel_queueing_send;

    (*init_func)(channel);
    
#ifdef FM_DEBUG
    sprintf(channel->name, "%s->%s", inport->name, outport->name);
#endif

    *f_ioval = (f_integer_t) 0;
} /* FM_CHANNEL */


/*
FM_MERGER_INIT() : Initialize a merger.
*/

void FM_MERGER_INIT(f_integer_t *f_channel,
		    f_integer_t *f_inport,
		    f_integer_t *f_size,
		    void (*init_func)(fm_channel_t *),
		    f_integer_t *f_ioval)
{
    fm_inport_t *inport;
    fm_channel_t *channel;
    int size;

    inport = (fm_inport_t *) *f_inport;
    size = (int) *f_size;

    InportMagicCheck(inport);

    NexusAssert(size >= 0);

    channel = _fm_alloc_channel();
    _fm_set_channel_size(channel, size);

    /* 
     * We must check to see if the inport is already
     * connected to something. If it is, disconnect it (sending
     * an out-of-scope message to the port's channel.
     */

    if (InportState(inport) != FM_STATE_UNDEFINED)
	disconnect_inport_channel(inport);

    NexusAssert(InportState(inport) == FM_STATE_UNDEFINED);
    NexusAssert(nexus_is_null_global_pointer(&inport->channel_gp));
		
    nexus_global_pointer(&channel->inport, (void *) inport);

    nexus_global_pointer(&inport->channel_gp, (void *) channel);

    NexusAssert(init_func != (void (*)(fm_channel_t *)) NULL);

    channel->send_func = _fm_channel_queueing_send;

    (*init_func)(channel);

#ifdef FM_DEBUG
    strcpy(channel->name, inport->name);
#endif

    InportSetState(inport, FM_STATE_NORMAL);

    *f_channel = (f_integer_t) channel;
    *f_ioval = (f_integer_t) 0;
} /* FM_MERGER_INIT() */


/*
FM_MERGER_ADD() : Add an outport to a merger.
*/

void FM_MERGER_ADD(f_integer_t *f_channel,
		   f_integer_t *f_outport,
		   f_integer_t *f_ioval)
{
    fm_outport_t *outport;
    fm_channel_t *channel;
    int idx;

    outport = (fm_outport_t *) *f_outport;
    channel = (fm_channel_t *) *f_channel;

    OutportMagicCheck(outport);
    ChannelMagicCheck(channel);

    /* 
     * We must check to see if the outport is already
     * connected to something. If it is disconnect  (sending
     * an out-of-scope message to the port's channel.
     */

    if (OutportState(outport) != FM_STATE_UNDEFINED)
	disconnect_outport_channel(outport);

    NexusAssert(OutportState(outport) == FM_STATE_UNDEFINED);
    NexusAssert(nexus_is_null_global_pointer(&outport->channel_gp));

    NexusAssert(channel->n_outports < channel->n_outports_alloc);
    idx = channel->n_outports++;

    nexus_global_pointer(&(channel->outports[idx]), (void *) outport);

    nexus_global_pointer(&outport->channel_gp, (void *) channel);

    OutportSetState(outport, FM_STATE_NORMAL);

#ifdef FM_DEBUG
    if (sizeof(channel->name) - strlen(channel->name) >
        strlen(outport->name) + 4)
    {
	strcat(channel->name, ":");
	strcat(channel->name, outport->name);
    }
#endif

    *f_ioval = (f_integer_t) 0;
} /* FM_MERGER_ADD() */


/*
FM_MERGER_DONE() : Finish up the creation of a merger.
*/

void FM_MERGER_DONE(f_integer_t *f_channel,
		   f_integer_t *f_ioval)
{
    fm_channel_t *channel;

    channel = (fm_channel_t *) *f_channel;

    ChannelMagicCheck(channel);

    NexusAssert(channel->n_outports == channel->n_outports_alloc);

    ChannelSetState(channel, FM_STATE_NORMAL);

    *f_ioval = (f_integer_t) 0;
    
} /* FM_MERGER_DONE() */


/*
FM_MOVE_INPORT() :  Move the inport from the source to the destination. To
avoid communication with the channel, we change the destination port
pointer to point at the port data structure that the source pointer
previously pointed at; deallocating the destination port, and pointing the
source port at a new port. 
*/

void
FM_MOVE_INPORT(f_integer_t *f_src, f_integer_t *f_dst,f_integer_t * f_ioval)
{
    int one;

#ifdef FM_DEBUG
    if (FMDebug(Ports, 1))
    {
	fm_inport_t *src, *dst;
	
	src = (fm_inport_t *) *f_src;
	InportMagicCheck(src);

	dst = (fm_inport_t *) *f_dst;
	InportMagicCheck(dst);

	nexus_printf("FM_MOVE_INPORT(): before: f_src=%x f_dst=%x src=%s "
		     "dst=%s\n",
		     f_src, f_dst,
		     _fm_inport_string(src),
		     _fm_inport_string(dst));
    }
#endif

    one = 1;
    FM_INVALIDATE_INPORT(f_dst, &one);

    *f_dst = *f_src;

    FM_INIT_INPORT(f_src, &one);

#ifdef FM_DEBUG
    if (FMDebug(Ports, 2))
    {
	fm_inport_t *src, *dst;

	src = (fm_inport_t *) *f_src;
	InportMagicCheck(src);

	dst = (fm_inport_t *) *f_dst;
	InportMagicCheck(dst);

	nexus_printf("FM_MOVE_INPORT(): after: f_src=%x f_dst=%x src=%s "
		     "dst=%s\n",
		     f_src, f_dst,
		     _fm_inport_string(src),
		     _fm_inport_string(dst));
    }
#endif

    *f_ioval = (f_integer_t) 0;
    
} /* FM_MOVE_INPORT() */


/*
FM_MOVE_OUTPORT() :  Move the outport from the source to the destination.
To avoid communication with the channel, we change the destination port
pointer to point at the port data structure that the source pointer
previously pointed at; deallocating the destination port, and pointing the
source port at a new port. 
*/

void
FM_MOVE_OUTPORT(f_integer_t *f_src, f_integer_t *f_dst,
		f_integer_t *f_ioval)
{
    int one;

#ifdef FM_DEBUG
    if (FMDebug(Ports, 1))
    {
	fm_outport_t *src, *dst;

	src = (fm_outport_t *) *f_src;
	OutportMagicCheck(src);

	dst = (fm_outport_t *) *f_dst;
	OutportMagicCheck(dst);

	nexus_printf("FM_MOVE_OUTPORT(): before: f_src=%x f_dst=%x src=%s "
		     "dst=%s\n",
		     f_src, f_dst,
		     _fm_outport_string(src),
		     _fm_outport_string(dst));
    }
#endif

    one = 1;
    FM_INVALIDATE_OUTPORT(f_dst, &one);

    *f_dst = *f_src;

    FM_INIT_OUTPORT(f_src, &one);

#ifdef FM_DEBUG
    if (FMDebug(Ports, 2))
    {
	fm_outport_t *src, *dst;

	src = (fm_outport_t *) *f_src;
	OutportMagicCheck(src);

	dst = (fm_outport_t *) *f_dst;
	OutportMagicCheck(dst);

	nexus_printf("FM_MOVE_OUTPORT(): after: f_src=%x f_dst=%x src=%s "
		     "dst=%s\n",
		     f_src, f_dst,
		     _fm_outport_string(src),
		     _fm_outport_string(dst));
    }
#endif
    *f_ioval = (f_integer_t) 0;
    
} /* FM_MOVE_OUTPORT() */


/*
FM_INVALIDATE_INPORT() : This routine is invoked whenever a "live" inport
goes out of scope or is otherwise destroyed (e.g. when it is a target of a
moveport).
*/

void FM_INVALIDATE_INPORT(f_integer_t *f_inports, f_integer_t *f_n_inports)
{
    int n_inports = (int) *f_n_inports;
    int i;
    fm_inport_t *inport;

    for (i = 0; i < n_inports; i++)
    {
	inport = (fm_inport_t *) f_inports[i];
	_fm_invalidate_inport(inport);
    }
} /* FM_INVALIDATE_INPORT() */


/*
FM_INVALIDATE_OUTPORT() : This routine is invoked whenever a "live" outport
goes out of scope or is otherwise destroyed (e.g. when it is a target of a
moveport).
*/

void
FM_INVALIDATE_OUTPORT(f_integer_t *f_outports, f_integer_t *f_n_outports)
{
    int n_outports = (int) *f_n_outports;
    int i;
    fm_outport_t *outport;

    for (i = 0; i < n_outports; i++)
    {
	outport = (fm_outport_t *) f_outports[i];
	_fm_invalidate_outport(outport);
    }
} /* FM_INVALIDATE_OUTPORT() */


/*
_fm_invalidate_ports() : Invalidate all the ports in the lists of lists.
*/

void
_fm_invalidate_ports(int n_ipls, fm_ipl_t *ipls, int n_total_inports,
		     int n_opls, fm_opl_t *opls, int n_total_outports)
{
    int i, j;
    fm_inport_t *inport;
    fm_outport_t *outport;
    fm_ipl_t *ipl;
    fm_opl_t *opl;

    for (i = 0; i < n_ipls; i++)
    {
	ipl = &(ipls[i]);

	for (j = 0; j < ipl->n_entries; j++)
	{
	    inport = ipl->entries[j];
	    _fm_invalidate_inport(inport);
	}
    }
    for (i = 0; i < n_opls; i++)
    {
	opl = &(opls[i]);

	for (j = 0; j < opl->n_entries; j++)
	{
	    outport = opl->entries[j];
	    _fm_invalidate_outport(outport);
	}
    }
} /* _fm_invalidate_ports() */


/*
_fm_invalidate_inport() : Invalidate an inport (that is, deallocate its
storage and mark it as undefined).
*/

void _fm_invalidate_inport(fm_inport_t *inport)
{
    InportMagicCheck(inport);
    InportLock(inport);
    
    if (InportState(inport) != FM_STATE_UNDEFINED)
    {
	reinit_inport(inport);
    }
    InportUnlock(inport);
    
} /* _fm_invalidate_inport() */


/*
_fm_invalidate_outport() : Invalidate an outport (that is, deallocate its
storage and mark it as undefined).
*/

void _fm_invalidate_outport(fm_outport_t *outport)
{
    OutportMagicCheck(outport);
    OutportLock(outport);
    
    if (OutportState(outport) != FM_STATE_UNDEFINED)
    {
	reinit_outport(outport);
    }
    OutportUnlock(outport);
    
} /* _fm_invalidate_outport() */


/*
disconnect_inport_channel() : Not yet implemented--DRK
*/

static void disconnect_inport_channel(fm_inport_t *inport)
{
#if 0
    nexus_buffer_t mbuf;
    nexus_global_pointer_t inport_gp;

    InportMagicCheck(inport):

    nexus_init_remote_service_request(&mbuf,
				      &inport->channel_gp,
				      FM_INPORT_OUT_OF_SCOPE_NAME,
				      FM_INPORT_OUT_OF_SCOPE_HASH,
				      NEXUS_FALSE);

    
#endif
    
    nexus_fatal("disconnect_inport_channel() is not yet implemented\n");
} /* disconnect_inport_channel() */


/*
disconnect_outport_channel() : Not yet implemented--DRK
*/

static void disconnect_outport_channel(fm_outport_t *outport)
{
    nexus_fatal("disconnect_outport_channel() is not yet implemented\n");
} /* disconnect_outport_channel() */


static void init_inport(fm_inport_t *ip)
{
    InportMagicInit(ip);

    nexus_null_global_pointer(&ip->channel_gp);
    InportSetState(ip, FM_STATE_UNDEFINED);
    _fm_synch_init(&ip->busy_synch);
    nexus_null_global_pointer(&ip->future_home);
    nexus_mutex_init(&ip->lock, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&ip->cond, (nexus_condattr_t *) NULL);
#ifdef FM_DEBUG
    strcpy(ip->name, "<unnamed inport>");
#endif
}


static void init_outport(fm_outport_t *op)
{
    OutportMagicInit(op);

    nexus_null_global_pointer(&op->channel_gp);
    OutportSetState(op, FM_STATE_UNDEFINED);
    _fm_synch_init(&op->busy_synch);
    nexus_null_global_pointer(&op->future_home);
    nexus_mutex_init(&op->lock, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&op->cond, (nexus_condattr_t *) NULL);
    _fm_synch_init(&op->synch);
#ifdef FM_DEBUG
    strcpy(op->name, "<unnamed outport>");
#endif
}


static void init_channel(fm_channel_t *c)
{
    ChannelMagicInit(c);

    c->outports = (nexus_global_pointer_t *) NULL;

    nexus_null_global_pointer(&c->inport);
    nexus_null_global_pointer(&c->future_home);

    nexus_mutex_init(&c->lock, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&c->cond, (nexus_condattr_t *) NULL);

    ChannelSetState(c, FM_STATE_UNDEFINED);
    _fm_synch_init(&c->busy_synch);
    _fm_init_queue(&c->queue);
    _fm_synch_init(&c->synch);

    c->direct_copy_destination = (void *) NULL;
    c->waiting_for_direct_send = NEXUS_FALSE;
    c->send_func = (fm_channel_send_func_t) NULL;
    c->close_func = (fm_channel_close_func_t) NULL;
    c->copy_func = (fm_channel_copy_func_t) NULL;
    c->stashed_copy_func = (fm_channel_stashed_copy_func_t) NULL;
    c->forward_func = (fm_channel_forward_func_t) NULL;

    c->index = -1;
#ifdef FM_DEBUG
    strcpy(c->name, "<unnamed channel>");
#endif
}


void _fm_set_channel_size(fm_channel_t *c, int n_outports)
{
    int i;
    
    ChannelMagicCheck(c);

/*
    NexusAssert(n_outports > 0);
*/

    c->n_outports_alloc = n_outports;
    c->n_outports = 0;

    NexusMalloc(_fm_set_channel_size(), c->outports,
		nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t) * n_outports);

    for (i = 0; i < n_outports; i++)
    {
        nexus_null_global_pointer(&c->outports[i]);
    }
}


static void reinit_inport(fm_inport_t *ip)
{
    InportMagicInit(ip);

    nexus_null_global_pointer(&ip->channel_gp);
    nexus_null_global_pointer(&ip->future_home);
    InportSetState(ip, FM_STATE_UNDEFINED);
    _fm_synch_clear(&ip->busy_synch);
#ifdef FM_DEBUG
    strcpy(ip->name, "<unnamed inport>");
#endif
}


static void reinit_outport(fm_outport_t *op)
{
    OutportMagicInit(op);

    nexus_null_global_pointer(&op->channel_gp);
    nexus_null_global_pointer(&op->future_home);
    OutportSetState(op, FM_STATE_UNDEFINED);
    _fm_synch_clear(&op->busy_synch);
    _fm_synch_clear(&op->synch);
#ifdef FM_DEBUG
    strcpy(op->name, "<unnamed outport>");
#endif
}


static void reinit_channel(fm_channel_t *c)
{
    ChannelMagicInit(c);

    if (c->outports)
	NexusFree(c->outports);
    
    c->outports = (nexus_global_pointer_t *) NULL;

    nexus_null_global_pointer(&c->inport);
    nexus_null_global_pointer(&c->future_home);

    ChannelSetState(c, FM_STATE_UNDEFINED);
    _fm_synch_clear(&c->busy_synch);
    _fm_reinit_queue(&c->queue);
    _fm_synch_clear(&c->synch);

    c->direct_copy_destination = (void *) NULL;
    c->waiting_for_direct_send = NEXUS_FALSE;
    c->send_func = (fm_channel_send_func_t) NULL;
    c->close_func = (fm_channel_close_func_t) NULL;
    c->copy_func = (fm_channel_copy_func_t) NULL;
    c->stashed_copy_func = (fm_channel_stashed_copy_func_t) NULL;
    c->forward_func = (fm_channel_forward_func_t) NULL;
    c->index = -1;
#ifdef FM_DEBUG
    strcpy(c->name, "<unnamed channel>");
#endif
}


#include "freelist.h"

FM_FREE_LIST_CODE(inport)
FM_FREE_LIST_CODE(outport)
FM_FREE_LIST_CODE(channel)


/*
 * _fm_ports_module_init()
 * 
 * Initialize the FM port allocation code...
 * 
 */

void _fm_ports_module_init()
{
    FM_FREE_LIST_INIT(inport);
    FM_FREE_LIST_INIT(outport);
    FM_FREE_LIST_INIT(channel);
} /* _fm_ports_module_init() */
