/**********************************************************************
$Id: simple_inst.h,v 1.4 1995/04/17 06:30:33 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/simple_inst.h,v $

$Log: simple_inst.h,v $
Revision 1.4  1995/04/17  06:30:33  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.3  1995/04/06  06:32:37  kohr
Added documentation, cleaned up a bit.


Definitions and declarations for simple instrumentation in FM.
**********************************************************************/


#ifdef FM_SIMPLE_INST

#ifndef _fm_simple_inst_h
#define _fm_simple_inst_h


#define SI_DEBUG

typedef struct _si_counter_t
{
    int count;
} si_counter_t;

typedef struct _si_timer_t
{
    double start_time;
    double accumulated_time;
    double last_elap_time;
    int running;
} si_timer_t;

typedef struct _si_timerhist_t
{
    int *bins;
    int n_bins;
    double bottom, top;
} si_timerhist_t;

typedef struct _si_inthist_t
{
    int *bins;
    int n_bins;
    int bottom, top;
} si_inthist_t;

typedef struct _si_instrument_set_t
{
    int n_counters;
    si_counter_t *counters;
    int n_timers;
    si_timer_t *timers;
    int n_timerhists;
    si_timerhist_t *timerhists;
    int n_inthists;
    si_inthist_t *inthists;
} si_instrument_set_t;

typedef struct _si_instrument_block_t
{
    si_instrument_set_t user;
    si_instrument_set_t sys;
} si_instrument_block_t;

extern void si_init();
extern void si_instrument_init(si_instrument_block_t *block,
			       int n_timers,
			       int n_timerhists,
			       int n_counters,
			       int n_inthists);
extern void si_instrument_set_init(si_instrument_set_t *set,
				   int n_timers,
				   int n_timerhists,
				   int n_counters,
				   int n_inthists);

extern void si_inthist_init(int id, int n_bins, int bottom, int top);
extern void si_timerhist_init(int id, int n_bins, double bottom, double top);
extern void si_instrument_set_dump(FILE *fp, si_instrument_set_t *set,
				   char *prefix);
extern void si_instrument_dump(char *process_name);
extern void si_instrument_free();

extern void si_counter_increment(int id);
extern void si_timer_start(int id);
extern void si_timer_stop(int id);

extern void si_timerhist_add(int id, int timer_id);

extern void si_inthist_add(int id, int value);

extern void si_sys_counter_increment(int id);
extern void si_sys_timer_start(int id);
extern void si_sys_timer_stop(int id);
extern void si_sys_timerhist_add(int id, int timer_id);
extern void si_sys_inthist_add(int id, int value);

#ifdef SI_DEBUG

#define SII_RANGECHECK_INTHIST(FUNC, SET, ID)				      \
    if (ID < 0 || ID >= (SET)->n_inthists)				      \
	nexus_fatal(#FUNC "(%d): id out of range (0-%d)\n",		      \
		    ID, (SET)->n_inthists - 1);

#define SII_RANGECHECK_TIMERHIST(FUNC, SET, ID)				      \
    if (ID < 0 || ID >= (SET)->n_timerhists)				      \
	nexus_fatal(#FUNC "(%d): id out of range (0-%d)\n",		      \
		    ID, (SET)->n_timerhists - 1);

#define SII_RANGECHECK_TIMER(FUNC, SET, ID)				      \
    if (ID < 0 || ID >= (SET)->n_timers)				      \
	nexus_fatal(#FUNC "(%d): id out of range (0-%d)\n",		      \
		    ID, (SET)->n_timers - 1);

#define SII_RANGECHECK_COUNTER(FUNC, SET, ID)				      \
    if (ID < 0 || ID >= (SET)->n_counters)				      \
	nexus_fatal(#FUNC "(%d): id out of range (0-%d)\n",		      \
		    ID, (SET)->n_counters - 1);

#else /* SI_DEBUG */

#define SII_RANGECHECK_INTHIST(FUNC, SET, ID)
#define SII_RANGECHECK_TIMERHIST(FUNC, SET, ID)
#define SII_RANGECHECK_TIMER(FUNC, SET, ID)
#define SII_RANGECHECK_COUNTER(FUNC, SET, ID)

#endif /* SI_DEBUG */

#define SII_INTHIST_INIT(FUNC, SET, ID, N_BINS, BOTTOM, TOP)		      \
{									      \
    si_inthist_t *hist;							      \
    int i;								      \
    									      \
    hist = &((SET)->inthists[ID]);					      \
									      \
    SII_RANGECHECK_INTHIST(FUNC, SET, ID)				      \
    NexusMalloc(FUNC,							      \
		hist->bins,						      \
		int *,							      \
		sizeof(int) * N_BINS);					      \
									      \
    for (i = 0; i < N_BINS; i++)					      \
	hist->bins[i] = 0;						      \
									      \
    hist->n_bins = N_BINS;						      \
    hist->bottom = BOTTOM;						      \
    hist->top = TOP;							      \
}

#define SII_TIMERHIST_INIT(FUNC, SET, ID, N_BINS, BOTTOM, TOP)		      \
{									      \
    si_timerhist_t *hist;						      \
    int i;								      \
    									      \
    hist = &((SET)->timerhists[ID]);					      \
									      \
    SII_RANGECHECK_TIMERHIST(FUNC, SET, ID)				      \
    NexusMalloc(FUNC,							      \
		hist->bins,						      \
		int *,							      \
		sizeof(int) * N_BINS);					      \
									      \
    for (i = 0; i < N_BINS; i++)					      \
	hist->bins[i] = 0;						      \
									      \
    hist->n_bins = N_BINS;						      \
    hist->bottom = BOTTOM;						      \
    hist->top = TOP;							      \
}

#define SII_COUNTER_INCREMENT(FUNC, SET, ID)				      \
    SII_RANGECHECK_COUNTER(FUNC, SET, ID)				      \
    (SET)->counters[ID].count++;

#ifdef SI_DEBUG
#define SII_TIMER_RUNNING_WARNING(FUNC, ID) \
	nexus_warning(#FUNC "(%d): timer already running\n", ID);
#else
#define SII_TIMER_RUNNING_WARNING(FUNC, ID)
#endif

#define SII_TIMER_START(FUNC, SET, ID)					      \
{									      \
    si_timer_t *timer;							      \
									      \
									      \
    SII_RANGECHECK_TIMER(FUNC, SET, ID)					      \
									      \
    timer = &((SET)->timers[id]);					      \
    if (timer->running)							      \
    {									      \
	SII_TIMER_RUNNING_WARNING(FUNC,ID)				      \
	return;								      \
    }									      \
									      \
    timer->start_time = nexus_wallclock();				      \
    timer->running = 1;							      \
}

#ifdef SI_DEBUG
#define SII_TIMER_STOPPED_WARNING(FUNC, ID) 				      \
	nexus_warning(#FUNC "(%d): timer not running\n", ID);
#else
#define SII_TIMER_STOPPED_WARNING(FUNC, ID)
#endif

#define SII_TIMER_STOP(FUNC, SET, ID)					      \
{									      \
    si_timer_t *timer;							      \
    double stop_time;							      \
									      \
    SII_RANGECHECK_TIMER(FUNC, SET, ID);				      \
    									      \
    timer = &((SET)->timers[id]);					      \
    if (!timer->running)						      \
    {									      \
	SII_TIMER_STOPPED_WARNING(FUNC, ID);				      \
	return;								      \
    }									      \
									      \
    stop_time = nexus_wallclock();					      \
    timer->last_elap_time = stop_time - timer->start_time;		      \
    timer->accumulated_time += timer->last_elap_time;			      \
    timer->running = 0;							      \
}

#define SII_TIMERHIST_ADD(FUNC, SET, HIST_ID, TIMER_ID)			      \
{									      \
    si_timer_t *timer;							      \
    si_timerhist_t *timerhist;						      \
    int bin;								      \
    double range;							      \
									      \
    SII_RANGECHECK_TIMER(FUNC, SET, TIMER_ID);				      \
    SII_RANGECHECK_TIMERHIST(FUNC, SET, HIST_ID);			      \
									      \
    timer = &((SET)->timers[TIMER_ID]);					      \
    timerhist = &((SET)->timerhists[HIST_ID]);				      \
									      \
    range = timerhist->top - timerhist->bottom;				      \
    									      \
    bin = (int) ((double) (timerhist->n_bins - 2) *			      \
                          (timer->last_elap_time - timerhist->bottom)/range); \
									      \
    if (bin < 0)							      \
	timerhist->bins[0]++;						      \
    else if (bin > timerhist->n_bins - 2)				      \
	timerhist->bins[timerhist->n_bins - 1]++;			      \
    else								      \
	timerhist->bins[bin]++;						      \
}

#define SII_INTHIST_ADD(FUNC, SET, ID, VALUE)				      \
{									      \
    si_inthist_t *inthist;						      \
    int bin;								      \
    int range;								      \
									      \
    SII_RANGECHECK_INTHIST(FUNC, SET, ID);				      \
									      \
    inthist = &((SET)->inthists[ID]);					      \
									      \
    range = inthist->top - inthist->bottom;				      \
    									      \
    bin = (int) ((double) (inthist->n_bins - 2) *			      \
    		          (double) (VALUE - inthist->bottom)/(double) range); \
									      \
    if (bin < 0)							      \
	bin = 0;							      \
    else if (bin > inthist->n_bins - 2)					      \
	bin = inthist->n_bins - 1;					      \
									      \
    inthist->bins[bin]++;						      \
}

#endif /* _fm_simple_inst_h */


#endif /* FM_SIMPLE_INST */
