/**********************************************************************
$Id: util.c,v 1.9 1995/04/17 06:30:45 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/util.c,v $

$Log: util.c,v $
Revision 1.9  1995/04/17  06:30:45  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.8  1995/04/06  06:32:49  kohr
Added documentation, cleaned up a bit.


FM runtime miscellaneous routines: string freelist, dump ports.
**********************************************************************/


#include "fm.h"


static nexus_mutex_t string_pool_lock;

#define STRING_POOL_SIZE 64

static char string_pool[STRING_POOL_SIZE][1024];
static int next_string = 0;


void _fm_util_init()
{
    nexus_mutex_init(&string_pool_lock, (nexus_mutexattr_t *) NULL);
}


char *_fm_util_get_string()
{
    char *buf;

    nexus_mutex_lock(&string_pool_lock);
    buf = string_pool[next_string];
    next_string = (next_string + 1) % STRING_POOL_SIZE;
    nexus_mutex_unlock(&string_pool_lock);

    return buf;
}


#ifdef FM_DEBUG

/*
_fm_{inport|outport|channel}_string() don't lock their inputs!--DRK
*/

/*
_fm_inport_string() : Return a string representing the given inport.
*/

char *_fm_inport_string(fm_inport_t *inport)
{
    char *buf;

    buf = _fm_util_get_string();

    InportMagicCheck(inport);
    sprintf(buf, "INPORT(\"%s\":%x:%d:chan=%s:fh=%s)",
#ifdef FM_DEBUG
	    inport->name,
#else
	    "<noname>",
#endif
	    (unsigned long) inport,
	    InportState(inport),
	    nexus_global_pointer_string(&inport->channel_gp),
	    nexus_global_pointer_string(&inport->future_home));

    return buf;
} /* _fm_inport_string() */


/*
_fm_outport_string() : Return a string representing the given outport.
*/

char *_fm_outport_string(fm_outport_t *outport)
{
    char *buf;

    buf = _fm_util_get_string();

    OutportMagicCheck(outport);
    sprintf(buf, "OUTPORT(\"%s\":%x:%d:chan=%s:fh=%s)",
#ifdef FM_DEBUG
	    outport->name,
#else
	    "<noname>",
#endif
	    (unsigned long) outport,
	    OutportState(outport),
	    nexus_global_pointer_string(&outport->channel_gp),
	    nexus_global_pointer_string(&outport->future_home));

    return buf;
} /* _fm_outport_string() */


/*
_fm_channel_string() : Return a string representing the given channel.
*/

char *_fm_channel_string(fm_channel_t *channel)
{
    char *buf;

    buf = _fm_util_get_string();

    ChannelMagicCheck(channel);
    sprintf(buf, "CHANNEL(\"%s\":%x:%d:n_ops=%d:op0=%s:ip=%s:fh=%s:wfds=%d)",
#ifdef FM_DEBUG
	    channel->name,
#else
	    "<noname>",
#endif
	    (unsigned long) channel,
	    ChannelState(channel),
	    channel->n_outports,
	    (channel->n_outports > 0) ? nexus_global_pointer_string(channel->outports) : "",
	    nexus_global_pointer_string(&channel->inport),
	    nexus_global_pointer_string(&channel->future_home),
	    channel->waiting_for_direct_send);

    return buf;
} /* _fm_channel_string() */


#endif /* FM_DEBUG */


#ifdef FM_DEBUG

void FMD_DUMP_FORTRAN_INPORT_LIST(f_integer_t *f_inports, f_integer_t *n)
{
    fm_inport_t *inports, *inport;
    int i;

    for (i = 0; i < *n; i++)
    {
	inport = (fm_inport_t *) f_inports[i];
	nexus_printf("inport %x: start_magic=%x end_magic=%x\n",
		     (unsigned long) inport,
		     inport->start_magic,
		     inport->end_magic);
	nexus_printf("inport %x: name=\"%s\"\n",
		     (unsigned long) inport,
		     inport->name);
	if (getenv("LONG_DUMP"))
	    nexus_printf("%s\n", _fm_inport_string(inport));
    }
}


void FMD_DUMP_FORTRAN_OUTPORT_LIST(f_integer_t *f_outports, f_integer_t *n)
{
    fm_outport_t *outports, *outport;
    int i;

    outports = (fm_outport_t *) *f_outports;

    for (i = 0; i < *n; i++)
    {
	outport = (fm_outport_t *) f_outports[i];
	nexus_printf("outport %x: start_magic=%x end_magic=%x\n",
		     (unsigned long) outport,
		     outport->start_magic,
		     outport->end_magic);
	nexus_printf("outport %x: name=\"%s\"\n",
		     (unsigned long) outport,
		     outport->name);
	if (getenv("LONG_DUMP"))
	    nexus_printf("%s\n", _fm_outport_string(outport));
    }
}

#endif /* FM_DEBUG */
