/*
 * nexus.h
 *
 * rcsid = "$Header: /home/compose/Master/FM/src/nexus/nexus.h,v 1.31 1995/08/22 20:31:46 tuecke Exp $"
 * nexus_rcsid = "Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/nexus.h,v 1.65 1995/08/21 23:39:24 patton Exp "
 *
 * This header contains the exported interface of Nexus.
 */

#ifndef _NEXUS_INCLUDE_NEXUS_H
#define _NEXUS_INCLUDE_NEXUS_H

/* EXTERN_C_BEGIN and EXTERN_C_END should surround any Nexus prototypes in
   nexus.h or Nexus .h files that are included by nexus.h.  This will
   allow C++ codes to include nexus.h and properly link against the
   Nexus library.
*/
#ifdef __cplusplus
#define EXTERN_C_BEGIN extern "C" {
#define EXTERN_C_END }
#else
#define EXTERN_C_BEGIN
#define EXTERN_C_END
#endif

EXTERN_C_BEGIN

/*
 * By defining this, Nexus will always try to use macros unless
 * this is overridden by NEXUS_DO_NOT_USE_MACROS.
 */
#define NEXUS_USE_MACROS


/*
 * nexus_path_type_t
 */
typedef enum _nexus_path_type_t
{
    NEXUS_PATH_TYPE_DATABASE = 1,
    NEXUS_PATH_TYPE_LISTENER = 2,
    NEXUS_PATH_MAXVAL
} nexus_path_type_t;

    
/*
 * nexus_arch_type_t
 */
typedef enum _nexus_arch_type_t
{
    NEXUS_ARCH_TYPE_UNKNOWN,
    NEXUS_ARCH_TYPE_ANY,
    NEXUS_ARCH_TYPE_SUN3,
    NEXUS_ARCH_TYPE_SUN4,
    NEXUS_ARCH_TYPE_NEXT040,
    NEXUS_ARCH_TYPE_NEXTX86,
    NEXUS_ARCH_TYPE_NEXTPA,
    NEXUS_ARCH_TYPE_NEXTSPARC,
    NEXUS_ARCH_TYPE_RS6000,
    NEXUS_ARCH_TYPE_PARAGON,
    NEXUS_ARCH_TYPE_NTX86,
    NEXUS_ARCH_TYPE_CRAYC90
} nexus_arch_type_t;


/*
 * nexus_class_type_t
 */
typedef enum _nexus_class_type_t
{
    NEXUS_CLASS_TYPE_UNKNOWN,
    NEXUS_CLASS_TYPE_32BIT_LE,	/* 32  bit integer, little endian,
				 * 32  bit IEEE float
				 * 64  bit IEEE double
				 */
    NEXUS_CLASS_TYPE_32BIT_BE,	/* 32  bit integer, big endian,
				 * 32  bit IEEE float
				 * 64  bit IEEE double
				 */
    NEXUS_CLASS_TYPE_64BIT_LE,	/* 64  bit integer, little endian,
				 * 64  bit IEEE float
				 * 128 bit IEEE double
				 */
    NEXUS_CLASS_TYPE_64BIT_BE,	/* 64  bit integer, big endian,
				 * 64  bit IEEE float
				 * 128 bit IEEE double
				 */
    NEXUS_CLASS_TYPE_CRAYC90,
    NEXUS_CLASS_TYPE_MAXVAL
} nexus_class_type_t;


/*
 * NEXUS_PROTO_TYPE_*
 *
 * Each different Nexus protocol module should have a unique
 * type identifier for all times...
 */
typedef enum _nexus_proto_type_t
{
    NEXUS_PROTO_TYPE_LOCAL	= 0,
    NEXUS_PROTO_TYPE_TCP	= 1,
    NEXUS_PROTO_TYPE_SHMEM	= 2,
    NEXUS_PROTO_TYPE_INX	= 3,
    NEXUS_PROTO_TYPE_PVM	= 4,
    NEXUS_PROTO_TYPE_EUI	= 5,
    NEXUS_PROTO_TYPE_ATM	= 6,
    NEXUS_PROTO_TYPE_MPL	= 7,
    NEXUS_PROTO_TYPE_MPINX	= 8,
    NEXUS_PROTO_TYPE_UDP	= 9,
    NEXUS_PROTO_TYPE_MN_UDP	= 10,
    NEXUS_PROTO_TYPE_MAXVAL, /* do not assign a value.. should float to one plus maximum valid number */
    NEXUS_PROTO_TYPE_INTERNAL1	= 1000001,
    NEXUS_PROTO_TYPE_INTERNAL2	= 1000002
} nexus_proto_type_t;

#ifdef NEXUS_DEBUG
#include "dbgstate.h"
#endif /* NEXUS_DEBUG */

EXTERN_C_END


#ifdef NEXUS_LITE
#include <nexuslite_config.h>
#else
#include <nexus_config.h>
#endif

#if defined(NEXUS_LITE) && !defined(PORTS0_LITE)
#define PORTS0_LITE
#endif

/*
 * Nexus should use macros internally when calling other
 * Nexus routines.  And other code using Nexus should probably
 * use macros wherever possible to avoid function call overhead.
 * But if you do not want Nexus to use macros,
 * and instead to always use real function calls, then
 * define NEXUS_DO_NOT_USE_MACROS.
 * For example, this could be defined in nexus_config.h
 * or in user code before including nexus.h.
 */
#ifdef NEXUS_DO_NOT_USE_MACROS
#undef NEXUS_USE_MACROS
#endif

#ifndef NEXUS_ARCH_TYPE
#define NEXUS_ARCH_TYPE NEXUS_ARCH_TYPE_UNKNOWN
#endif

#ifndef NEXUS_CLASS_TYPE
#define NEXUS_CLASS_TYPE NEXUS_CLASS_TYPE_UNKNOWN
#endif

EXTERN_C_BEGIN

#define	NEXUS_TRUE		1
#define	NEXUS_FALSE		0

#define NEXUS_MAX(V1,V2) (((V1) > (V2)) ? (V1) : (V2))
#define NEXUS_MIN(V1,V2) (((V1) < (V2)) ? (V1) : (V2))

#ifdef NEXUS_DEFINE_GLOBALS
#define NEXUS_GLOBAL
#else
#define NEXUS_GLOBAL extern
#endif

#define NEXUS_MAX_HANDLER_NAME_LENGTH 255

#define NEXUS_MAX_EXECUTABLE_PATH_LENGTH 1023

#define NEXUS_MAX_DEBUG_LEVELS 10

#ifdef NEXUS_DEBUG
#define NexusAssert(assertion) \
    if (!(assertion)) \
    { \
        nexus_fatal("Assertion " #assertion " failed in file %s at line %d\n",\
		   __FILE__, __LINE__); \
    }
#define NexusAssert2(assertion, print_args) \
    if (!(assertion)) \
    { \
        nexus_fatal("Assertion " #assertion " failed in file %s at line %d: %s", \
		   __FILE__, __LINE__, nexus_assert_sprintf print_args); \
    }
#else /* NEXUS_DEBUG */
#define NexusAssert(assertion)
#define NexusAssert2(assertion, print_args)
#endif /* NEXUS_DEBUG */


#define NexusMalloc(Func, Var, Type, Size) \
{ \
    if ((Size) > 0) \
    { \
	if (((Var) = (Type) ports0_malloc (Size)) == (Type) NULL) \
	{ \
	    nexus_fatal("%s: malloc of size %d failed for %s %s in file %s line %d\n", \
                        #Func, (Size), #Type, #Var, __FILE__, __LINE__); \
	} \
    } \
    else \
    { \
	(Var) = (Type) NULL; \
    } \
}

#define NexusFree(Ptr) \
{ \
    if ((Ptr) != NULL) \
    { \
	ports0_free(Ptr); \
    } \
}


/*
 * Some forward typedef declarations...
 */
typedef struct _nexus_proto_funcs_t	nexus_proto_funcs_t;
typedef struct _nexus_buffer_funcs_t	nexus_buffer_funcs_t;
typedef struct _nexus_buffer_t *	nexus_buffer_t;
typedef struct _nexus_stashed_buffer_t *nexus_stashed_buffer_t;
typedef struct _nexus_global_pointer_t	nexus_global_pointer_t;
typedef struct _nexus_proto_t		nexus_proto_t;
typedef struct _nexus_mi_proto_t	nexus_mi_proto_t;


/*
 * Handler stuff
 */
typedef void (*nexus_non_threaded_handler_func_t)(void *address,
						  nexus_buffer_t *buffer);
typedef void (*nexus_threaded_handler_func_t)(void *address,
					      nexus_stashed_buffer_t *buffer);
typedef void (*nexus_unknown_handler_func_t)(void *address,
					     nexus_buffer_t *buffer,
					     char *handler_name,
					     int handler_id);
typedef void (*nexus_exit_handler_func_t)();
typedef int  (*nexus_acquired_as_node_handler_func_t)();
typedef void *nexus_handler_func_t;

typedef enum _nexus_handler_type_t
{
    NEXUS_HANDLER_TYPE_THREADED,
    NEXUS_HANDLER_TYPE_NON_THREADED
} nexus_handler_type_t;

typedef struct _nexus_handler_t {
    char *			name;
    int				id;  
    nexus_handler_type_t	type;
    nexus_handler_func_t	func;	/* nexus_non_threaded_handler_func_t */
					/* or nexus_threaded_handler_func_t  */
} nexus_handler_t;


/*
 * Other typedefs...
 */
typedef void *(*nexus_thread_func_t)(void *user_arg);
typedef int nexus_bool_t;


/*
 * The user must provide a variable containing an initial set of
 * modules that are staticly linked into the executable.
 * This variables contains an array of the triples:
 *	{family_name, module_name, info_function_ptr}
 * Where:
 *   family_name is:
 *	"startups"	startup module
 *	"protocols"	protocol module
 *	"database"	database module
 *   module_name is a unique (within the family) string name for the module.
 *   info_function_ptr is a function pointer to the module's info function.
 *
 * The general form of a module information function is
 *	_nx_<family_name>_<module_name>_info().
 * So the user might supply a table that looks like:
 *
 * extern void *_nx_st_rsh_info();
 * extern void *_nx_pr_tcp_info();
 * nexus_module_list_t nexus_module_list[] =
 * {
 *     {"startups", "rsh", _nx_st_rsh_info},
 *     {"protocols", "tcp", _nx_pr_tcp_info},
 *     {NULL, NULL, NULL},
 * };
 *
 * This would specify that the rsh startup module, and the tcp protocol
 * module are staticly linked into the program.
 *
 * Note: On machines that provide dynamic linking, this list can be
 * empty.  All necessary modules will be dynamically loaded.  However,
 * linking some in using this list may provide for quicker startup.
 * Also, some startup modules need to be staticly linked to bootstrap
 * properly.
 */
typedef struct _nexus_module_list_t
{
    char *family_name;
    char *module_name;
    void *(*info_func)();
} nexus_module_list_t;
/*
extern nexus_module_list_t nexus_module_list[];
*/

EXTERN_C_END


/*********************************************************************
 * Include ports0
 */
#if defined(NEXUS_USE_MACROS) && !defined(PORTS0_USE_MACROS)
#define PORTS0_USE_MACROS
#endif
#include "ports0.h"


/*********************************************************************
 * Include sanity checking header stuff
 */
#include "nx_sanity.h"


#ifndef NEXUS_LITE

/*********************************************************************
 * Include the thread module's external header, which should contain:
 *   1) #include's for system thread headers
 *   2) typedefs for external nexus types
 *   3) macros for external nexus routines
 */
#ifndef NEXUS_TH_HEADER
#define NEXUS_TH_HEADER "th.h"
#endif

#include NEXUS_TH_HEADER

#endif /* NEXUS_LITE */


/*********************************************************************
 * Some typdefs
 */

EXTERN_C_BEGIN

#ifdef NEXUS_LITE
/*
 * For NexusLite we need to dummy up some simple versions
 * of the mutex and cond stuff.
 */
typedef int	nexus_mutex_t;
typedef int	nexus_cond_t;
typedef int	nexus_mutexattr_t;
typedef int	nexus_condattr_t;
#endif /* NEXUS_LITE */


/*
 * nexus_create_context_handle_t
 */
typedef struct _nexus_create_context_handle_t
{
    int				type;
    nexus_mutex_t		mutex;
    nexus_cond_t		cond;
    int				total;
    int				checked_in;
    int				next_gp;
    nexus_global_pointer_t **	gp;
    int **			rc;
} nexus_create_context_handle_t;


/*
 * Table containing sizes of the various data types, in bytes.
 */
typedef struct _nexus_sizeof_table_t
{
    int float_size;
    int double_size;
    int short_size;
    int u_short_size;
    int int_size;
    int u_int_size;
    int long_size;
    int u_long_size;
    int char_size;
    int u_char_size;
} nexus_sizeof_table_t;


/*
 * nexus_buffer_t
 *
 * This is the only part of a nexus_buffer_t that
 * needs to be exported to the user.  (i.e., There is no need
 * to export specific protocol buffer information to the user.)
 * It is the part of nexus_buffer_t that is common to the buffer_t's
 * used by each different protocol.
 *
 * This part of the nexus_buffer_t needs to live here so that
 * macros below can work.
 *
 * All protocol buffer_t's (i.e., tcp_buffer_t) should start with
 * the same fields as nexus_buffer_t.  They should use
 * typecasts to convert to their own local type.
 *
 * Technically we are using this as a union of all of the different
 * buffer_t's of the protocols.  However, since the set of protocols
 * to use in a particular executable should be selectable at
 * link time, we cannot use an actual union, but must rely on type
 * casting instead.
 */
struct _nexus_buffer_t
{
#ifdef NEXUS_DEBUG
    int				magic;
#endif
    nexus_buffer_funcs_t *	funcs;
};

struct _nexus_stashed_buffer_t
{
#ifdef NEXUS_DEBUG
    int				magic;
#endif
    nexus_buffer_funcs_t *	funcs;
};


/*
 * nexus_proto_t
 *
 * nexus_proto_t is similar to nexus_buffer_t.
 * Each protocol will have to add data to its version of a proto_t.
 * This structure is the common root to all of them -- all proto_t's
 * should start with these same fields.
 *
 * This part of the nexus_proto_t needs to live here so that
 * macros below can work.
 */
struct _nexus_proto_t
{
    nexus_proto_type_t	 type;
    nexus_proto_funcs_t *funcs;
};


/*
 * nexus_global_pointer_t
 *
 * A global pointer points to an address within a particular context.
 * That context obviously lives within some process (or Mach task)
 * on some machine.
 * There may or may not be other contexts on within that same process.
 *
 * The 'proto' and 'all_protos' fields provide a description of how
 * to get to the correct process for a given context.  'all_protos'
 * is a machine independent representation of all of the methods
 * that can be used to get to that process.  'proto' is one of
 * these protocols (the one that should be used when communicating
 * with this gp) that has been translated into a useful form.
 *
 * Once you to the correct process, 'context' discriminates between
 * the contexts that are in that process.
 *
 * And 'address' is the address within that context that this
 * global pointer points to.
 */
struct _nexus_global_pointer_t
{
    nexus_proto_t *	proto;
    nexus_mi_proto_t *	all_protos;
    unsigned long	context;
    unsigned long	address;
    nexus_class_type_t  class_type;
#ifdef NEXUS_PROFILE
    int			node_id;
    int			context_id;
#endif /* NEXUS_PROFILE */    
};


/*
 * nexus_node_t
 *
 * An array of these is returned by nexus_init() on the master node,
 * and by nexus_acquire_nodes().
 */
typedef struct _nexus_node_t
{
    char *			name;
    int				number;
    nexus_global_pointer_t	gp;
    int				return_code;
} nexus_node_t;

/* Values for nexus_node_t.return_code */
#define NEXUS_NODE_NEW		0
#define NEXUS_NODE_OLD_SAME_EXE	1
#define NEXUS_NODE_OLD_DIFF_EXE	2


#if defined(NEXUS_THREAD_SAFE_STDIO) || defined(NEXUS_LITE)
#define nexus_stdio_lock()
#define nexus_stdio_unlock()
#else
NEXUS_GLOBAL nexus_mutex_t nexus_stdio_mutex;
#define nexus_stdio_lock() nexus_mutex_lock(&nexus_stdio_mutex);
#define nexus_stdio_unlock() nexus_mutex_unlock(&nexus_stdio_mutex);
#endif /* NEXUS_THREAD_SAFE_STDIO */

/*********************************************************************
 * 		Function indirection table declarations
 *********************************************************************/

/*
 * This is a function indirection table used in conn_t's
 */
struct _nexus_proto_funcs_t
{
    nexus_proto_type_t	(*proto_type)();
    void		(*init)(int *argc, char ***argv);
    void		(*shutdown)(nexus_bool_t shutdown_others);
    void		(*abort)();
    void		(*poll)();
    void		(*blocking_poll)();
    void      (*init_remote_service_request)(nexus_buffer_t *buffer,
					     nexus_global_pointer_t *gp,
					     char *handler_name,
					     int handler_id);
    void		(*destroy_proto)(nexus_proto_t *proto);
    void                (*copy_proto)(nexus_proto_t *proto_in,
				      nexus_proto_t **proto_out);
    nexus_mi_proto_t *	(*get_my_mi_proto)();
    nexus_bool_t		(*construct_from_mi_proto)(nexus_proto_t **proto,
						   nexus_mi_proto_t *mi_proto);
    void		(*get_creator_proto_params)(char *buf,
						    int buf_size,
						    char *node_name,
						    int node_number);
    void		(*construct_creator_proto)(nexus_proto_t **proto);
    int			(*compare_protos)(nexus_proto_t *proto1,
					  nexus_proto_t *proto2);
};


/*
 * This is a function indirection table used in nexus_buffer_t's
 */
struct _nexus_buffer_funcs_t
{
    void		(*set_buffer_size)(nexus_buffer_t *buffer,
					   int size, int n_elements);
    int			(*check_buffer_size)(nexus_buffer_t *buffer,
					     int slack, int increment);
    void		(*send_remote_service_request)
						(nexus_buffer_t *buffer);
    void		(*send_urgent_remote_service_request)
						(nexus_buffer_t *buffer);
    void		(*free_buffer)(nexus_buffer_t *buffer);
    void		(*stash_buffer)(nexus_buffer_t *buffer,
				       nexus_stashed_buffer_t *stashed_buffer);
    void		(*free_stashed_buffer)
				      (nexus_stashed_buffer_t *stashed_buffer);

    int (*sizeof_float)(nexus_buffer_t *buffer, int count);
    int (*sizeof_double)(nexus_buffer_t *buffer, int count);
    int (*sizeof_short)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_short)(nexus_buffer_t *buffer, int count);
    int (*sizeof_int)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_int)(nexus_buffer_t *buffer, int count);
    int (*sizeof_long)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_long)(nexus_buffer_t *buffer, int count);
    int (*sizeof_char)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_char)(nexus_buffer_t *buffer, int count);

    void (*put_float)(nexus_buffer_t *buffer,
		      float *data, int count);
    void (*put_double)(nexus_buffer_t *buffer,
		       double *data, int count);
    void (*put_short)(nexus_buffer_t *buffer,
		      short *data, int count);
    void (*put_u_short)(nexus_buffer_t *buffer,
			unsigned short *data, int count);
    void (*put_int)(nexus_buffer_t *buffer,
		    int *data, int count);
    void (*put_u_int)(nexus_buffer_t *buffer,
		      unsigned int *data, int count);
    void (*put_long)(nexus_buffer_t *buffer,
		     long *data, int count);
    void (*put_u_long)(nexus_buffer_t *buffer,
		       unsigned long *data, int count);
    void (*put_char)(nexus_buffer_t *buffer,
		     char *data, int count);
    void (*put_u_char)(nexus_buffer_t *buffer,
		       unsigned char *data, int count);

    void (*get_float)(nexus_buffer_t *buffer,
		      float *dest, int count);
    void (*get_double)(nexus_buffer_t *buffer,
		       double *dest, int count);
    void (*get_short)(nexus_buffer_t *buffer,
		      short *dest, int count);
    void (*get_u_short)(nexus_buffer_t *buffer,
			unsigned short *dest, int count);
    void (*get_int)(nexus_buffer_t *buffer,
		    int *dest, int count);
    void (*get_u_int)(nexus_buffer_t *buffer,
		      unsigned int *dest, int count);
    void (*get_long)(nexus_buffer_t *buffer,
		     long *dest, int count);
    void (*get_u_long)(nexus_buffer_t *buffer,
		       unsigned long *dest, int count);
    void (*get_char)(nexus_buffer_t *buffer,
		     char *dest, int count);
    void (*get_u_char)(nexus_buffer_t *buffer,
		       unsigned char *dest, int count);

    void (*get_stashed_float)(nexus_stashed_buffer_t *buffer,
			      float *dest, int count);
    void (*get_stashed_double)(nexus_stashed_buffer_t *buffer,
			       double *dest, int count);
    void (*get_stashed_short)(nexus_stashed_buffer_t *buffer,
			      short *dest, int count);
    void (*get_stashed_u_short)(nexus_stashed_buffer_t *buffer,
				unsigned short *dest, int count);
    void (*get_stashed_int)(nexus_stashed_buffer_t *buffer,
			    int *dest, int count);
    void (*get_stashed_u_int)(nexus_stashed_buffer_t *buffer,
			      unsigned int *dest, int count);
    void (*get_stashed_long)(nexus_stashed_buffer_t *buffer,
			     long *dest, int count);
    void (*get_stashed_u_long)(nexus_stashed_buffer_t *buffer,
			       unsigned long *dest, int count);
    void (*get_stashed_char)(nexus_stashed_buffer_t *buffer,
			     char *dest, int count);
    void (*get_stashed_u_char)(nexus_stashed_buffer_t *buffer,
			       unsigned char *dest, int count);
};



/*********************************************************************
 * Declarations for functions that must be supplied by the package
 *********************************************************************/

extern int	NexusBoot();


/*********************************************************************
 * Declarations for functions that may optionally be supplied by
 * the package by registering them as handlers.
 *********************************************************************/

extern void	NexusExit();
extern void	NexusUnknownHandler(void *address,
				    nexus_buffer_t *buffer,
				    char *handler_name,
				    int handler_id);
extern int	NexusAcquiredAsNode();

#define NEXUS_EXIT_HASH 941
#define NEXUS_UNKNOWN_HANDLER_HASH 964
#define NEXUS_ACQUIRED_AS_NODE_HASH 894
#define NEXUS_LISTENER_PATH_HASH 745


/*********************************************************************
 * 		Exported interface declarations
 *********************************************************************/
extern int	nexus_user_iargc();
extern char *	nexus_user_getarg(int i);
extern int	nexus_package_iargc();
extern char *	nexus_package_getarg(int i);
extern void	nexus_get_argc_and_argv(int *argc, char ***argv);
extern void     nexus_set_path(nexus_path_type_t, char *);

extern void	nexus_init(int *argc,
			   char ***argv,
			   char *args_env_variable,
			   char *package_designator,
			   int (*package_args_init_func)(int *argc,
							 char ***argv),
			   void (*usage_message_func)(),
			   int (*new_process_params_func)(char *buf, int size),
			   nexus_module_list_t module_list[],
			   nexus_node_t **nodes,
			   int *n_nodes);
extern void	nexus_start();
extern void	nexus_start_nonblocking(void);
extern void	nexus_shutdown_nonexiting(void);
extern void	nexus_exit(int rc, int shutdown);
extern void	nexus_shutdown();
extern void	nexus_abort();

extern void	nexus_destroy_current_context(
				int called_from_non_threaded_handler);
extern void	nexus_acquire_nodes(char *node_name,
				    int node_number,
				    int count,
				    char *directory_path,
				    char *executable_path,
				    nexus_node_t **nodes,
				    int *n_nodes);

extern void	nexus_init_create_context_handle(
				nexus_create_context_handle_t *contexts,
                                int n_contexts);
extern void	nexus_create_context(nexus_global_pointer_t *node_gp,
				     char *executable_path,
				     int size,
				     nexus_global_pointer_t *new_context_gp,
				     int *return_code,
				     nexus_create_context_handle_t *contexts);
extern void	nexus_create_context_wait(
				nexus_create_context_handle_t *contexts);
extern void *	nexus_context_initial_segment();
extern void *	nexus_malloc(size_t size);
extern void	nexus_free(void *data_segment);

extern void	nexus_poll();
extern void	nexus_poll_profile(int id);
extern void	nexus_blocking_poll();
extern void	nexus_global_pointer(nexus_global_pointer_t *new_gp,
				     void *address);
extern void *	nexus_convert_global_pointer_address(
						nexus_global_pointer_t *gp);
extern void	nexus_destroy_global_pointer(nexus_global_pointer_t *gp);
extern void     nexus_copy_global_pointer(nexus_global_pointer_t *gp_out,
					  nexus_global_pointer_t *gp_in);
extern int	nexus_sizeof_global_pointer(nexus_buffer_t *buffer,
					    nexus_global_pointer_t *gp,
					    int count,
					    int *n_elements);
extern void	nexus_put_global_pointer(nexus_buffer_t *buffer,
					 nexus_global_pointer_t *gp,
					 int count);
extern void	nexus_get_global_pointer(nexus_buffer_t *buffer,
					 nexus_global_pointer_t *gp,
					 int count);
extern void	nexus_get_stashed_global_pointer(
						nexus_stashed_buffer_t *buffer,
						nexus_global_pointer_t *gp,
						int count);


extern int	nexus_handler_hash(char *name);
extern void	nexus_register_handlers(nexus_handler_t *handlers);
extern void	nexus_substitute_handler(char *name,
					 int id,
					 nexus_handler_type_t new_type,
					 void *new_func,
					 nexus_handler_type_t *old_type,
					 void **old_func);

extern nexus_arch_type_t	nexus_node_type(nexus_global_pointer_t *gp);
extern nexus_class_type_t	nexus_node_class(nexus_global_pointer_t *gp);
extern nexus_global_pointer_t *	nexus_node_exists(char *hostname,
						  char *path);
extern int nexus_same_context(nexus_global_pointer_t *gp1,
			      nexus_global_pointer_t *gp2);
extern int nexus_same_global_pointer(nexus_global_pointer_t *gp1,
				     nexus_global_pointer_t *gp2);
extern int nexus_global_pointer_to_current_context(nexus_global_pointer_t *gp);

extern int nexus_master_node(void);

/*
 * args.c
 */
extern int	nexus_find_argument(int *argc,
				    char ***argv,
				    char *arg,
				    int count);
extern void	nexus_remove_arguments(int *argc,
				       char ***argv,
				       int arg_num,
				       int count);
extern void	nexus_get_retained_arguments(int **argc, char ****argv);

/*
 * db_iface.c
 */
extern char *   nexus_database_lookup(char *node_name,
				      int node_number,
				      char *key);
extern void     nexus_database_free(char *value);

/*
 * error.c
 */
extern void		nexus_silent_fatal();
#if defined(__STDC__) || defined(__cplusplus)
extern void		nexus_fatal(char *msg, ...);
extern void		nexus_error(char *msg, ...);
extern void		nexus_warning(char *msg, ...);
extern void		nexus_notice(char *msg, ...);
extern void		nexus_printf(char *msg, ...);
extern void		nexus_perror(char *msg, ...);
extern void		nexus_fatal_perror(char *msg, ...);
extern char *		nexus_assert_sprintf(char *msg, ...);
#else /* __STDC__ */
extern void		nexus_fatal();
extern void		nexus_error();
extern void		nexus_warning();
extern void		nexus_notice();
extern void		nexus_printf();
extern void		nexus_perror();
extern void		nexus_fatal_perror();
extern char *		nexus_assert_sprintf();
#endif /* __STDC__ */


/*
 * md_*.c
 */
extern void		nexus_kill(int pid);
extern void 		nexus_usleep(long usecs);
extern double		nexus_wallclock();
extern void		nexus_pause();


/*
 * pr_tcp.c
 */
extern int		nexus_allow_attach(unsigned short *port,
					   char **host,
					   int (*approval_func)(void *user_arg,
								char *url,
						   nexus_global_pointer_t *gp),
					   void *approval_func_user_arg
					   );
extern void		nexus_disallow_attach(unsigned short port);
extern int		nexus_attach_with_url(char *url,
					      nexus_global_pointer_t *gp);
extern int		nexus_split_nexus_url(char *url,
					      char **host,
					      unsigned short *port,
					      char ***specifiers);
extern void		nexus_split_nexus_url_free(char **host,
						   char ***specifiers);

/*
 * nexus_register_fd_notifiers() registers functions that
 * Nexus should call to register any fd's that it uses.
 *
 * Note: This is _not_ a general solution.  It exists so that nperl
 * does not have to use a patched version of nexus.  It _will_ change
 * sometime in the future.  You should not rely on this function.
 */
typedef void (*nexus_fd_notifier_func_t)(int fd);
extern void	nexus_register_fd_notifiers(nexus_fd_notifier_func_t add,
					    nexus_fd_notifier_func_t remove);



/*
 * th_*.c
 */
extern nexus_bool_t		nexus_preemptive_threads();


/*
 * utils.c
 */
extern void		nexus_ids(int *node_id,
				  int *context_id,
				  int *thread_id);



/*********************************************************************
 * 		Exported interface routines that can be macros
 *********************************************************************/

#define nexus_macro_set_package_id(package_id_base) \
    ports0_set_package_id(package_id_base)
#define nexus_macro_get_package_id() \
    ports0_get_package_id()
#define	nexus_macro_null_global_pointer(gp) \
{ \
    (gp)->proto = (nexus_proto_t *) NULL; \
    (gp)->all_protos = (nexus_mi_proto_t *) NULL; \
}
#define nexus_macro_is_null_global_pointer(gp) \
    ((gp)->proto == (nexus_proto_t *) NULL)
#define nexus_macro_init_remote_service_request(buf, gp, h_name, h_id) \
    ((gp)->proto->funcs->init_remote_service_request)(buf, gp, \
						      h_name, h_id)
#ifndef NEXUS_USE_MACROS
extern int	nexus_set_package_id(char *package_id_base);
extern char *	nexus_get_package_id();
extern void	nexus_null_global_pointer(nexus_global_pointer_t *gp);
extern nexus_bool_t	nexus_is_null_global_pointer(nexus_global_pointer_t *gp);
extern void nexus_init_remote_service_request(nexus_buffer_t *buffer,
					      nexus_global_pointer_t *gp,
					      char *handler_name,
					      int handler_id);
#else  /* NEXUS_USE_MACROS */
#define nexus_set_package_id(package_id_base) \
    nexus_macro_set_package_id(package_id_base)
#define nexus_get_package_id() \
    nexus_macro_get_package_id()
#define	nexus_null_global_pointer(gp) \
    nexus_macro_null_global_pointer(gp)
#define nexus_is_null_global_pointer(gp) \
    nexus_macro_is_null_global_pointer(gp)
#define nexus_init_remote_service_request(buf, gp, h_name, h_id) \
    nexus_macro_init_remote_service_request(buf, gp, h_name, h_id)
#endif /* NEXUS_USE_MACROS */



#define nexus_macro_set_buffer_size(buffer, size, n_elements) \
    ((*buffer)->funcs->set_buffer_size)(buffer, size, n_elements)
#define nexus_macro_check_buffer_size(buffer, slack, increment) \
    ((*buffer)->funcs->check_buffer_size)(buffer, slack, increment)
#define nexus_macro_send_remote_service_request(buffer) \
    ((*buffer)->funcs->send_remote_service_request)(buffer)
#define nexus_macro_send_urgent_remote_service_request(buffer) \
    ((*buffer)->funcs->send_urgent_remote_service_request)(buffer)
#define nexus_macro_free_buffer(buffer) \
{ \
    if (*buffer) /* *buffer==NULL if the buffer has been stashed */ \
    { \
	((*buffer)->funcs->free_buffer)(buffer); \
    } \
}
#define nexus_macro_stash_buffer(buffer, stashed_buffer) \
    ((*buffer)->funcs->stash_buffer)(buffer, stashed_buffer)
#define nexus_macro_free_stashed_buffer(stashed_buffer) \
    ((*stashed_buffer)->funcs->free_stashed_buffer)(stashed_buffer)
#ifndef NEXUS_USE_MACROS
extern void		nexus_set_buffer_size(nexus_buffer_t *buffer,
					      int size,
					      int n_elements);
extern int		nexus_check_buffer_size(nexus_buffer_t *buffer,
						int slack,
						int increment);
extern void		nexus_send_remote_service_request(
					       nexus_buffer_t *buffer);
extern void		nexus_send_urgent_remote_service_request(
					       nexus_buffer_t *buffer);
extern void		nexus_free_buffer(nexus_buffer_t *buffer);
extern void		nexus_stash_buffer(nexus_buffer_t *buffer,
				   nexus_stashed_buffer_t *stashed_buffer);
extern void		nexus_free_stashed_buffer(
				   nexus_stashed_buffer_t *stashed_buffer);
#else  /* NEXUS_USE_MACROS */
#define nexus_set_buffer_size(buffer, size, n_elements) \
    nexus_macro_set_buffer_size(buffer, size, n_elements)
#define nexus_check_buffer_size(buffer, slack, increment) \
    nexus_macro_check_buffer_size(buffer, slack, increment)
#define nexus_send_remote_service_request(buffer) \
    nexus_macro_send_remote_service_request(buffer)
#define nexus_send_urgent_remote_service_request(buffer) \
    nexus_macro_send_urgent_remote_service_request(buffer)
#define nexus_free_buffer(buffer) \
    nexus_macro_free_buffer(buffer)
#define nexus_stash_buffer(buffer, stashed_buffer) \
    nexus_macro_stash_buffer(buffer, stashed_buffer)
#define nexus_free_stashed_buffer(stashed_buffer) \
    nexus_macro_free_stashed_buffer(stashed_buffer)
#endif /* NEXUS_USE_MACROS */



#define nexus_macro_sizeof_float(buffer, count) \
    ((*buffer)->funcs->sizeof_float)(buffer, count)
#define nexus_macro_sizeof_double(buffer, count) \
    ((*buffer)->funcs->sizeof_double)(buffer, count)
#define nexus_macro_sizeof_short(buffer, count) \
    ((*buffer)->funcs->sizeof_short)(buffer, count)
#define nexus_macro_sizeof_u_short(buffer, count) \
    ((*buffer)->funcs->sizeof_u_short)(buffer, count)
#define nexus_macro_sizeof_int(buffer, count) \
    ((*buffer)->funcs->sizeof_int)(buffer, count)
#define nexus_macro_sizeof_u_int(buffer, count) \
    ((*buffer)->funcs->sizeof_u_int)(buffer, count)
#define nexus_macro_sizeof_long(buffer, count) \
    ((*buffer)->funcs->sizeof_long)(buffer, count)
#define nexus_macro_sizeof_u_long(buffer, count) \
    ((*buffer)->funcs->sizeof_u_long)(buffer, count)
#define nexus_macro_sizeof_char(buffer, count) \
    ((*buffer)->funcs->sizeof_char)(buffer, count)
#define nexus_macro_sizeof_u_char(buffer, count) \
    ((*buffer)->funcs->sizeof_u_char)(buffer, count)
#ifndef NEXUS_USE_MACROS
extern int	nexus_sizeof_float(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_double(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_short(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_short(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_int(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_int(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_long(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_long(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_char(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_char(nexus_buffer_t *buffer, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_sizeof_float(buffer, count) \
    nexus_macro_sizeof_float(buffer, count)
#define nexus_sizeof_double(buffer, count) \
    nexus_macro_sizeof_double(buffer, count)
#define nexus_sizeof_short(buffer, count) \
    nexus_macro_sizeof_short(buffer, count)
#define nexus_sizeof_u_short(buffer, count) \
    nexus_macro_sizeof_u_short(buffer, count)
#define nexus_sizeof_int(buffer, count) \
    nexus_macro_sizeof_int(buffer, count)
#define nexus_sizeof_u_int(buffer, count) \
    nexus_macro_sizeof_u_int(buffer, count)
#define nexus_sizeof_long(buffer, count) \
    nexus_macro_sizeof_long(buffer, count)
#define nexus_sizeof_u_long(buffer, count) \
    nexus_macro_sizeof_u_long(buffer, count)
#define nexus_sizeof_char(buffer, count) \
    nexus_macro_sizeof_char(buffer, count)
#define nexus_sizeof_u_char(buffer, count) \
    nexus_macro_sizeof_u_char(buffer, count)
#endif /* NEXUS_USE_MACROS */


#define nexus_macro_put_float(buffer, data, count) \
    ((*buffer)->funcs->put_float)(buffer, data, count)
#define nexus_macro_put_double(buffer, data, count) \
    ((*buffer)->funcs->put_double)(buffer, data, count)
#define nexus_macro_put_short(buffer, data, count) \
    ((*buffer)->funcs->put_short)(buffer, data, count)
#define nexus_macro_put_u_short(buffer, data, count) \
    ((*buffer)->funcs->put_u_short)(buffer, data, count)
#define nexus_macro_put_int(buffer, data, count) \
    ((*buffer)->funcs->put_int)(buffer, data, count)
#define nexus_macro_put_u_int(buffer, data, count) \
    ((*buffer)->funcs->put_u_int)(buffer, data, count)
#define nexus_macro_put_long(buffer, data, count) \
    ((*buffer)->funcs->put_long)(buffer, data, count)
#define nexus_macro_put_u_long(buffer, data, count) \
    ((*buffer)->funcs->put_u_long)(buffer, data, count)
#define nexus_macro_put_char(buffer, data, count) \
    ((*buffer)->funcs->put_char)(buffer, data, count)
#define nexus_macro_put_u_char(buffer, data, count) \
    ((*buffer)->funcs->put_u_char)(buffer, data, count)
#ifndef NEXUS_USE_MACROS
extern void nexus_put_float(nexus_buffer_t *buffer,
			    float *data, int count);
extern void nexus_put_double(nexus_buffer_t *buffer,
			     double *data, int count);
extern void nexus_put_short(nexus_buffer_t *buffer,
			    short *data, int count);
extern void nexus_put_u_short(nexus_buffer_t *buffer,
			      unsigned short *data, int count);
extern void nexus_put_int(nexus_buffer_t *buffer,
			  int *data, int count);
extern void nexus_put_u_int(nexus_buffer_t *buffer,
			    unsigned int *data, int count);
extern void nexus_put_long(nexus_buffer_t *buffer,
			   long *data, int count);
extern void nexus_put_u_long(nexus_buffer_t *buffer,
			     unsigned long *data, int count);
extern void nexus_put_char(nexus_buffer_t *buffer,
			   char *data, int count);
extern void nexus_put_u_char(nexus_buffer_t *buffer,
			     unsigned char *data, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_put_float(buffer, data, count) \
    nexus_macro_put_float(buffer, data, count)
#define nexus_put_double(buffer, data, count) \
    nexus_macro_put_double(buffer, data, count)
#define nexus_put_short(buffer, data, count) \
    nexus_macro_put_short(buffer, data, count)
#define nexus_put_u_short(buffer, data, count) \
    nexus_macro_put_u_short(buffer, data, count)
#define nexus_put_int(buffer, data, count) \
    nexus_macro_put_int(buffer, data, count)
#define nexus_put_u_int(buffer, data, count) \
    nexus_macro_put_u_int(buffer, data, count)
#define nexus_put_long(buffer, data, count) \
    nexus_macro_put_long(buffer, data, count)
#define nexus_put_u_long(buffer, data, count) \
    nexus_macro_put_u_long(buffer, data, count)
#define nexus_put_char(buffer, data, count) \
    nexus_macro_put_char(buffer, data, count)
#define nexus_put_u_char(buffer, data, count) \
    nexus_macro_put_u_char(buffer, data, count)
#endif /* NEXUS_USE_MACROS */


#define nexus_macro_get_float(buffer, dest, count) \
    ((*buffer)->funcs->get_float)(buffer, dest, count)
#define nexus_macro_get_double(buffer, dest, count) \
    ((*buffer)->funcs->get_double)(buffer, dest, count)
#define nexus_macro_get_short(buffer, dest, count) \
    ((*buffer)->funcs->get_short)(buffer, dest, count)
#define nexus_macro_get_u_short(buffer, dest, count) \
    ((*buffer)->funcs->get_u_short)(buffer, dest, count)
#define nexus_macro_get_int(buffer, dest, count) \
    ((*buffer)->funcs->get_int)(buffer, dest, count)
#define nexus_macro_get_u_int(buffer, dest, count) \
    ((*buffer)->funcs->get_u_int)(buffer, dest, count)
#define nexus_macro_get_long(buffer, dest, count) \
    ((*buffer)->funcs->get_long)(buffer, dest, count)
#define nexus_macro_get_u_long(buffer, dest, count) \
    ((*buffer)->funcs->get_u_long)(buffer, dest, count)
#define nexus_macro_get_char(buffer, dest, count) \
    ((*buffer)->funcs->get_char)(buffer, dest, count)
#define nexus_macro_get_u_char(buffer, dest, count) \
    ((*buffer)->funcs->get_u_char)(buffer, dest, count)
#ifndef NEXUS_USE_MACROS
extern void nexus_get_float(nexus_buffer_t *buffer,
			    float *dest, int count);
extern void nexus_get_double(nexus_buffer_t *buffer,
			     double *dest, int count);
extern void nexus_get_short(nexus_buffer_t *buffer,
			    short *dest, int count);
extern void nexus_get_u_short(nexus_buffer_t *buffer,
			      unsigned short *dest, int count);
extern void nexus_get_int(nexus_buffer_t *buffer,
			  int *dest, int count);
extern void nexus_get_u_int(nexus_buffer_t *buffer,
			    unsigned int *dest, int count);
extern void nexus_get_long(nexus_buffer_t *buffer,
			   long *dest, int count);
extern void nexus_get_u_long(nexus_buffer_t *buffer,
			     unsigned long *dest, int count);
extern void nexus_get_char(nexus_buffer_t *buffer,
			   char *dest, int count);
extern void nexus_get_u_char(nexus_buffer_t *buffer,
			     unsigned char *dest, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_get_float(buffer, dest, count) \
    nexus_macro_get_float(buffer, dest, count)
#define nexus_get_double(buffer, dest, count) \
    nexus_macro_get_double(buffer, dest, count)
#define nexus_get_short(buffer, dest, count) \
    nexus_macro_get_short(buffer, dest, count)
#define nexus_get_u_short(buffer, dest, count) \
    nexus_macro_get_u_short(buffer, dest, count)
#define nexus_get_int(buffer, dest, count) \
    nexus_macro_get_int(buffer, dest, count)
#define nexus_get_u_int(buffer, dest, count) \
    nexus_macro_get_u_int(buffer, dest, count)
#define nexus_get_long(buffer, dest, count) \
    nexus_macro_get_long(buffer, dest, count)
#define nexus_get_u_long(buffer, dest, count) \
    nexus_macro_get_u_long(buffer, dest, count)
#define nexus_get_char(buffer, dest, count) \
    nexus_macro_get_char(buffer, dest, count)
#define nexus_get_u_char(buffer, dest, count) \
    nexus_macro_get_u_char(buffer, dest, count)
#endif /* NEXUS_USE_MACROS */


#define nexus_macro_get_stashed_float(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_float)(buffer, dest, count)
#define nexus_macro_get_stashed_double(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_double)(buffer, dest, count)
#define nexus_macro_get_stashed_short(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_short)(buffer, dest, count)
#define nexus_macro_get_stashed_u_short(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_short)(buffer, dest, count)
#define nexus_macro_get_stashed_int(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_int)(buffer, dest, count)
#define nexus_macro_get_stashed_u_int(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_int)(buffer, dest, count)
#define nexus_macro_get_stashed_long(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_long)(buffer, dest, count)
#define nexus_macro_get_stashed_u_long(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_long)(buffer, dest, count)
#define nexus_macro_get_stashed_char(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_char)(buffer, dest, count)
#define nexus_macro_get_stashed_u_char(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_char)(buffer, dest, count)
#ifndef NEXUS_USE_MACROS
extern void nexus_get_stashed_float(nexus_stashed_buffer_t *buffer,
				    float *dest, int count);
extern void nexus_get_stashed_double(nexus_stashed_buffer_t *buffer,
				     double *dest, int count);
extern void nexus_get_stashed_short(nexus_stashed_buffer_t *buffer,
				    short *dest, int count);
extern void nexus_get_stashed_u_short(nexus_stashed_buffer_t *buffer,
				      unsigned short *dest, int count);
extern void nexus_get_stashed_int(nexus_stashed_buffer_t *buffer,
				  int *dest, int count);
extern void nexus_get_stashed_u_int(nexus_stashed_buffer_t *buffer,
				    unsigned int *dest, int count);
extern void nexus_get_stashed_long(nexus_stashed_buffer_t *buffer,
				   long *dest, int count);
extern void nexus_get_stashed_u_long(nexus_stashed_buffer_t *buffer,
				     unsigned long *dest, int count);
extern void nexus_get_stashed_char(nexus_stashed_buffer_t *buffer,
				   char *dest, int count);
extern void nexus_get_stashed_u_char(nexus_stashed_buffer_t *buffer,
				     unsigned char *dest, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_get_stashed_float(buffer, dest, count) \
    nexus_macro_get_stashed_float(buffer, dest, count)
#define nexus_get_stashed_double(buffer, dest, count) \
    nexus_macro_get_stashed_double(buffer, dest, count)
#define nexus_get_stashed_short(buffer, dest, count) \
    nexus_macro_get_stashed_short(buffer, dest, count)
#define nexus_get_stashed_u_short(buffer, dest, count) \
    nexus_macro_get_stashed_u_short(buffer, dest, count)
#define nexus_get_stashed_int(buffer, dest, count) \
    nexus_macro_get_stashed_int(buffer, dest, count)
#define nexus_get_stashed_u_int(buffer, dest, count) \
    nexus_macro_get_stashed_u_int(buffer, dest, count)
#define nexus_get_stashed_long(buffer, dest, count) \
    nexus_macro_get_stashed_long(buffer, dest, count)
#define nexus_get_stashed_u_long(buffer, dest, count) \
    nexus_macro_get_stashed_u_long(buffer, dest, count)
#define nexus_get_stashed_char(buffer, dest, count) \
    nexus_macro_get_stashed_char(buffer, dest, count)
#define nexus_get_stashed_u_char(buffer, dest, count) \
    nexus_macro_get_stashed_u_char(buffer, dest, count)
#endif /* NEXUS_USE_MACROS */


#ifdef NEXUS_LITE
/*
 * For NexusLite we need to dummy up some simple versions
 * of the mutex and cond stuff.
 */

#define nexus_macro_mutex_init(M,A)	(*(M) = 0)
#define nexus_macro_mutex_destroy(M)	(*(M) = 0)
#define nexus_macro_mutex_lock(M) \
    ( (*(M)) \
     ? (nexus_fatal("nexus_mutex_lock(): Deadlock detected in file %s at line %d. The mutex at address 0x%lu is already locked.\n", __FILE__, __LINE__, (unsigned long) (M)), 1) \
     : ( ((*(M)) = 1), 0 ) )
#define nexus_macro_mutex_unlock(M)	( ((*(M)) = 0), 0 )
#define nexus_macro_cond_init(C,A)	(*(C) = 0)
#define nexus_macro_cond_destroy(C)	(*(C) = 0)
#define nexus_macro_cond_wait(C,M) \
    ( ((*(M)) = 0), nexus_blocking_poll(), ((*(M)) = 1), 0 )
#define nexus_macro_cond_signal(C)	(*(C) = 0)
#define nexus_macro_cond_broadcast(C)	(*(C) = 0)

#ifndef NEXUS_USE_MACROS
extern int		nexus_mutex_init(nexus_mutex_t *mutex,
					 nexus_mutexattr_t *attr);
extern int		nexus_mutex_destroy(nexus_mutex_t *mutex);
extern int		nexus_mutex_lock(nexus_mutex_t *mutex);
extern int		nexus_mutex_unlock(nexus_mutex_t *mutex);
extern int		nexus_cond_init(nexus_cond_t *cond,
					nexus_condattr_t *attr);
extern int		nexus_cond_destroy(nexus_cond_t *cond);
extern int		nexus_cond_wait(nexus_cond_t *cond,
					nexus_mutex_t *mutex);
extern int		nexus_cond_signal(nexus_cond_t *cond);
extern int		nexus_cond_broadcast(nexus_cond_t *cond);
#else  /* NEXUS_USE_MACROS */
#define nexus_mutex_init(M,A) nexus_macro_mutex_init(M,A)
#define nexus_mutex_destroy(M) nexus_macro_mutex_destroy(M)
#define nexus_mutex_lock(M) nexus_macro_mutex_lock(M)
#define nexus_mutex_unlock(M) nexus_macro_mutex_unlock(M)
#define nexus_cond_init(C,A) nexus_macro_cond_init(C,A)
#define nexus_cond_destroy(C) nexus_macro_cond_destroy(C)
#define nexus_cond_wait(C,M) nexus_macro_cond_wait(C,M)
#define nexus_cond_signal(C) nexus_macro_cond_signal(C)
#define nexus_cond_broadcast(C) nexus_macro_cond_broadcast(C)
#endif /* NEXUS_USE_MACROS */

#endif /* NEXUS_LITE */


EXTERN_C_END

/* 
 * At this point, we have defined all of the macros that can effect
 * the size of structures in nexus.h
 */
#include "nx_ver.h"

#define NEXUS_NO_NODE_NUM    -1

#endif /* _NEXUS_INCLUDE_NEXUS_H */
