/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


#include <stdio.h>

#ifndef __GNUC__
# include <stdlib.h>
#else
// # pragma implementation
#endif

#include "macro.h"
#include "vpc.h"
#include "f90.h"

#include "extcxx_low.h"

#undef USER


//
// define for having the debugging
//
//define  DEBUGLIB 1
#define  MAX_FILES 1000
//
//
// Array to keep track of table for a file
//
//

void **tablebfnd[MAX_FILES];
void **tablellnd[MAX_FILES];
void **tabletype[MAX_FILES];
void **tablesymbol[MAX_FILES];
void **tablelabel[MAX_FILES];

int numtablebfnd[MAX_FILES];
int numtablellnd[MAX_FILES];
int numtabletype[MAX_FILES];
int numtablesymbol[MAX_FILES];
int numtablelabel[MAX_FILES];

static int CurrentFileNumber = 0;

//
// Table for making link between the nodes and the classes
// Take the id and return a pointer
//

void **fileTableClass;
int allocatedForfileTableClass;
void **bfndTableClass;
int allocatedForbfndTableClass;
void **llndTableClass;
int allocatedForllndTableClass;
void **typeTableClass;
int allocatedFortypeTableClass;
void **symbolTableClass;
int allocatedForsymbolTableClass;
void **labelTableClass;
int allocatedForlabelTableClass;

//
// Some definition for this module
//
#define ALLOCATECHUNK 1000

#define SORRY Message("Sorry, not implemented yet",0)


class SgProject;
class SgFile;
class SgStatement;
class SgExpression;
class SgLabel;
class SgSymbol;
class SgType;
class SgUnaryExp;
class SgClassSymb;
class SgVarDeclStmt; 

extern "C" char * funparse_llnd(PTR_LLND);
extern "C" char * funparse_bfnd(PTR_BFND);

//
// Set of function to care about the table management
//

void InitializeTable()
{
  int i;
  for (i=0 ; i < MAX_FILES; i++)
    {
      tablebfnd[i] = NULL;
      tablellnd[i]= NULL;
      tabletype[i]= NULL;
      tablesymbol[i]= NULL;
      tablelabel[i]= NULL;

      numtablebfnd[i]= 0;
      numtablellnd[i]= 0;
      numtabletype[i]= 0;
      numtablesymbol[i]= 0;
      numtablelabel[i]= 0;
    }

  
  fileTableClass =  NULL;
  bfndTableClass =  NULL;
  llndTableClass  = NULL;
  typeTableClass  = NULL;
  symbolTableClass  = NULL;
  labelTableClass  =  NULL;
  allocatedForfileTableClass = 0;
  allocatedForbfndTableClass = 0;
  allocatedForllndTableClass = 0;
  allocatedFortypeTableClass = 0;
  allocatedForsymbolTableClass =0;
  allocatedForlabelTableClass = 0;
}


void SwitchToFile(int i)
{
  if (i >= MAX_FILES)
    {
      Message("Too many files",0);
      exit(1);
    }
  
 tablebfnd[CurrentFileNumber] = bfndTableClass;
 tablellnd[CurrentFileNumber]= llndTableClass;
 tabletype[CurrentFileNumber]= typeTableClass;
 tablesymbol[CurrentFileNumber]= symbolTableClass;
 tablelabel[CurrentFileNumber]= labelTableClass;

 numtablebfnd[CurrentFileNumber]= allocatedForbfndTableClass;
 numtablellnd[CurrentFileNumber]= allocatedForllndTableClass;
 numtabletype[CurrentFileNumber]= allocatedFortypeTableClass;
 numtablesymbol[CurrentFileNumber]= allocatedForsymbolTableClass;
 numtablelabel[CurrentFileNumber]= allocatedForlabelTableClass;

 bfndTableClass =  tablebfnd[i];
 llndTableClass  = tablellnd[i];
 typeTableClass  = tabletype[i];
 symbolTableClass  = tablesymbol[i];
 labelTableClass  =  tablelabel[i];

 allocatedForbfndTableClass =  numtablebfnd[i];
 allocatedForllndTableClass = numtablellnd[i];
 allocatedFortypeTableClass = numtabletype[i];
 allocatedForsymbolTableClass =numtablesymbol[i];
 allocatedForlabelTableClass = numtablelabel[i];
 CurrentFileNumber = i;
}



// add a chunk to the size 
void ReallocatefileTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForfileTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForfileTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  
  for (i=0 ; i < allocatedForfileTableClass; i++)
    {
      pt[i] = fileTableClass[i];
    }
  if (allocatedForfileTableClass)
    delete fileTableClass;
  fileTableClass = pt;
  allocatedForfileTableClass = allocatedForfileTableClass + ALLOCATECHUNK;
}


void ReallocatebfndTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForbfndTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForbfndTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForbfndTableClass; i++)
    {
      pt[i] = bfndTableClass[i];
    }
  if (allocatedForbfndTableClass)
    delete bfndTableClass;
  bfndTableClass = pt;
  allocatedForbfndTableClass = allocatedForbfndTableClass + ALLOCATECHUNK;
}


void ResetbfndTableClass()
{
  int i;
  
  for (i=0 ; i < allocatedForbfndTableClass; i++)
    {
//      delete bfndTableClass[i];
      bfndTableClass[i] = NULL;
    }
}

void ReallocatellndTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForllndTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForllndTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForllndTableClass; i++)
    {
      pt[i] = llndTableClass[i];
    }
  if (allocatedForllndTableClass)
    delete llndTableClass;
  llndTableClass = pt;
  allocatedForllndTableClass = allocatedForllndTableClass + ALLOCATECHUNK;
}

void ReallocatesymbolTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForsymbolTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForsymbolTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForsymbolTableClass; i++)
    {
      pt[i] = symbolTableClass[i];
    }
  if (allocatedForsymbolTableClass)
    delete symbolTableClass;
  symbolTableClass = pt;
  allocatedForsymbolTableClass = allocatedForsymbolTableClass + ALLOCATECHUNK;
}


void ReallocatelabelTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForlabelTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForlabelTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForlabelTableClass; i++)
    {
      pt[i] = labelTableClass[i];
    }
  if (allocatedForlabelTableClass)
    delete labelTableClass;
  labelTableClass = pt;
  allocatedForlabelTableClass = allocatedForlabelTableClass + ALLOCATECHUNK;
}


void ReallocatetypeTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedFortypeTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedFortypeTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedFortypeTableClass; i++)
    {
      pt[i] = typeTableClass[i];
    }
  if (allocatedFortypeTableClass)
    delete typeTableClass;
  typeTableClass = pt;
  allocatedFortypeTableClass = allocatedFortypeTableClass + ALLOCATECHUNK;
}


void RemoveFromTableType(void * pt)
{
  int i;
  for (i=0 ; i < allocatedFortypeTableClass; i++)
    {
      if (typeTableClass[i] == pt)
        {
          typeTableClass[i] = NULL;
          return;
        }
    }
}


void RemoveFromTableSymb(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForsymbolTableClass; i++)
    {
      if (symbolTableClass[i] == pt)
        {
          symbolTableClass[i] = NULL;
          return;
        }
    }
}



void RemoveFromTableBfnd(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForbfndTableClass; i++)
    {
      if (bfndTableClass[i] == pt)
        {
          bfndTableClass[i] = NULL;
          return;
        }
    }
}


void RemoveFromTableFile(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForfileTableClass; i++)
    {
      if (fileTableClass[i] == pt)
        {
          fileTableClass[i] = NULL;
          return;
        }
    }
}


void RemoveFromTableLlnd(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForllndTableClass; i++)
    {
      if (llndTableClass[i] == pt)
        {
          llndTableClass[i] = NULL;
          return;
        }
    }
}



void RemoveFromTableLabel(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForlabelTableClass; i++)
    {
      if (labelTableClass[i] == pt)
        {
          labelTableClass[i] = NULL;
          return;
        }
    }
}



void 
SetMappingInTableForBfnd(PTR_BFND bif, void *pt)
{
  if (!bif)
    return ;
  while (allocatedForbfndTableClass <= BIF_ID(bif))
    {
      ReallocatebfndTableClass();
    }
  bfndTableClass[BIF_ID(bif)] = pt;
}


void 
SetMappingInTableForType(PTR_TYPE type, void *pt)
{
  if (!type)
    return ;
  while (allocatedFortypeTableClass <= TYPE_ID(type))
    {
      ReallocatetypeTableClass();
    }
  typeTableClass[TYPE_ID(type)] = pt;
}


void 
SetMappingInTableForSymb(PTR_SYMB symb, void *pt)
{
  if (!symb)
    return ;
  while (allocatedForsymbolTableClass <= SYMB_ID(symb))
    {
      ReallocatesymbolTableClass();
    }
  symbolTableClass[SYMB_ID(symb)] = pt;
}

void 
SetMappingInTableForLabel(PTR_LABEL lab, void *pt)
{
  if (!lab)
    return ;
  while (allocatedForlabelTableClass <= LABEL_ID(lab))
    {
      ReallocatelabelTableClass();
    }
  labelTableClass[SYMB_ID(lab)] = pt;
}



void 
SetMappingInTableForLlnd(PTR_LLND ll, void *pt)
{
  if (!ll)
    return ;
  while (allocatedForllndTableClass <= NODE_ID(ll))
    {
      ReallocatellndTableClass();
    }
  llndTableClass[NODE_ID(ll)] = pt;
}


void 
SetMappingInTableForFile(PTR_FILE file, void *pt)
{
  int id;
  if (!file)
    return ;
  id = GetFileNum(FILE_FILENAME(file));
  while (allocatedForfileTableClass <= id)
    {
      ReallocatefileTableClass();
    }
  fileTableClass[id] = pt;
}


SgSymbol *
GetMappingInTableForSymbol(PTR_SYMB symb)
{
  int id;
  if (!symb)
    return NULL;
  id = SYMB_ID(symb);
  if (allocatedForsymbolTableClass <= id)    
    {
      return NULL;
    }
  return (SgSymbol *) symbolTableClass[id];
}



SgLabel *
GetMappingInTableForLabel(PTR_LABEL lab)
{
  int id;
  if (!lab)
    return NULL;
  id = LABEL_ID(lab);
  if (allocatedForlabelTableClass <= id)    
    {
      return NULL;
    }
  return (SgLabel *) labelTableClass[id];
}


SgStatement *
GetMappingInTableForBfnd(PTR_BFND bf)
{
  int id;
  if (!bf)
    return NULL;
  id = BIF_ID(bf);
  if (allocatedForbfndTableClass <= id)    
    {
      return NULL;
    }
  return (SgStatement *) bfndTableClass[id];
}


SgType *
GetMappingInTableForType(PTR_TYPE t)
{
  int id;
  if (!t)
    return NULL;
  id = TYPE_ID(t);
  if (allocatedFortypeTableClass <= id)    
    {
      return NULL;
    }
  return (SgType *) typeTableClass[id];
}


SgExpression *
GetMappingInTableForLlnd(PTR_LLND ll)
{
  int id;
  if (!ll)
    return NULL;
  id = NODE_ID(ll);
  if (allocatedForllndTableClass <= id)    
    {
      return NULL;
    }
  return (SgExpression *) llndTableClass[id];
}


SgFile *
GetMappingInTableForFile(PTR_FILE file)
{
  int id;
  if (!file)
    return NULL;
  id = GetFileNum(FILE_FILENAME(file));
  if (allocatedForfileTableClass <= id)
    {
      return NULL;
    }
  return (SgFile *) fileTableClass[id];
}


//Fortran and C++ Structures
//
// There several families of classes here.
//   Projects-    which correspond to a collection of parsed
//                source files.
//   Files -      which corresponds to an individual source file
//   Statements-  Fortran or C statements
//   Expressions- Fortran or C expression trees.
//   Symbols-     Symbol Table entries.
//   Types-       Each symbol has a type which lives in a type table.
//   Labels-      Statement labels in fortran or C
//   Dependences- Data Dependence Class 
//
//  naming convention: Classnames begin with Sg (for Sage)
//  class functions begin with a lower case and have first letters
//  of words in Caps likeThisWord.
//  
//  In general functions return references when ever possible.  
//
//
// ************* Project and File Types ******************
// the sage fortran 90 and c++ parsers generate files with
// a .dep extension.  A project is a file with a .proj extension
// that consists of a list of .dep files that make the basis
// of the project.  The following describes the
// basic mechanisms to access and modify the structures
// The class hierarch is as follows:
//
//SgProject     = the class representing multi source file projects
//
//SgFile        = the basic source file object.
//   - SgFortranFile  = the subclass for Fortran sources
//   - SgCFile        = the subclass for C files.
//
// ******************************************************************

// forward ref
SgStatement * BfndMapping(PTR_BFND bif);
SgExpression * LlndMapping(PTR_LLND llin);
SgSymbol * SymbMapping(PTR_SYMB symb);
SgType * TypeMapping(PTR_TYPE ty);
SgLabel * LabelMapping(PTR_LABEL label);

// As you can see, some statements are specifically Fortran and
// some apply only to C and C++.
//

// the generic statement class has functions to access or modify any
// property of a given statement.

SgProject *CurrentProject;

#include "libSage++.h"

//
// Some Mapping stuff
//
SgStatement * BfndMapping(PTR_BFND bif)
{
  SgStatement *pt = NULL;
  if (!bif)
    {
      return pt;
    }
  pt = GetMappingInTableForBfnd(bif);
  if (pt)
    return pt;
  else
    {
      pt = new SgStatement(bif);      
    }  
  return pt;
}


//
// Some mapping stuff
//

SgExpression * LlndMapping(PTR_LLND llin)
{
  SgExpression *pt;
  if (!llin)
    return NULL;
  pt = GetMappingInTableForLlnd(llin);
  if (pt)
    return pt;
  else
    {
      pt = new SgExpression(llin);      
    } 
  return pt;
}


SgSymbol * SymbMapping(PTR_SYMB symb)
{
  SgSymbol *pt = NULL;
  if (!symb)
    {
      return pt;
    }
  pt = GetMappingInTableForSymbol(symb);
  if (pt)
    return pt;
  else
    {
      pt = new SgSymbol(symb);      
    }  
  return pt;
}

SgType * TypeMapping(PTR_TYPE ty)
{
  SgType *pt = NULL;
  
  if (!ty)
    return NULL;
  pt = GetMappingInTableForType(ty);
  if (pt)
    return pt;
  else
    {
      pt = new SgType(ty);      
    } 
  return pt;
}



SgLabel * LabelMapping(PTR_LABEL label)
{
  SgLabel *pt = NULL;
  if (!label)
    {
      return pt;
    }
  pt = GetMappingInTableForLabel(label);
  if (pt)
    return pt;
  else
    {
      pt = new SgLabel(label);      
    }  
  return pt;
}



SgValueExp * isSgValueExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case INT_VAL:
    case CHAR_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case STRING_VAL:
    case COMPLEX_VAL:
    case KEYWORD_VAL:
      return (SgValueExp *) pt;
    default:
      return NULL;
    }
}



SgKeywordValExp * isSgKeywordValExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case KEYWORD_VAL:
      return (SgKeywordValExp *) pt;
    default:
      return NULL;
    }
}


SgUnaryExp & makeAnUnaryExpression(int code,PTR_LLND ll1);

// I didn't understand what this function does.
// Should be modified to use LlndMapping.

SgExpression & SgUnaryExp::operand()
{
  PTR_LLND ll;
  SgExpression *pt = NULL;
  
  ll = NODE_OPERAND0(thellnd);
  if (!ll)
    ll = NODE_OPERAND1(thellnd);
  pt = GetMappingInTableForLlnd(ll);
  if (pt)
    return *pt;
  else
    {
      pt = new SgExpression(ll);      
    } 
  return *pt;
};

// Other handy constructors
SgUnaryExp &SgDerefOp(SgExpression &e)
  {return makeAnUnaryExpression(DEREF_OP,e.thellnd);};

SgUnaryExp &SgAddrOp(SgExpression &e)
  {return makeAnUnaryExpression(ADDRESS_OP,e.thellnd);};

SgUnaryExp &SgUMinusOp(SgExpression &e)
  {return makeAnUnaryExpression(MINUS_OP,e.thellnd);};

SgUnaryExp &SgUPlusOp(SgExpression &e)
  {return makeAnUnaryExpression(UNARY_ADD_OP,e.thellnd);};

SgUnaryExp &SgPrePlusPlusOp(SgExpression &e)
  {return makeAnUnaryExpression(PLUSPLUS_OP,e.thellnd);};

SgUnaryExp &SgPreMinusMinusOp(SgExpression &e)
  {return makeAnUnaryExpression(MINUSMINUS_OP,e.thellnd);};

SgUnaryExp &SgPostPlusPlusOp(SgExpression &e)
  { SgUnaryExp *pt;
    pt =  &makeAnUnaryExpression(PLUSPLUS_OP,e.thellnd);
    
    NODE_OPERAND1(pt->thellnd) = NODE_OPERAND0(pt->thellnd);
    NODE_OPERAND0(pt->thellnd) = 0;
    return *pt;
  };
SgUnaryExp &SgPostMinusMinusOp(SgExpression &e)
  {
    SgUnaryExp *pt;
    pt =  &makeAnUnaryExpression(MINUSMINUS_OP,e.thellnd);
    
    NODE_OPERAND1(pt->thellnd) = NODE_OPERAND0(pt->thellnd);
    NODE_OPERAND0(pt->thellnd) = 0;
    return *pt;
  };
SgUnaryExp &SgBitCompfOp(SgExpression &e)
  {return makeAnUnaryExpression(BIT_COMPLEMENT_OP,e.thellnd);};
SgUnaryExp &SgNotOp(SgExpression &e)
  {return makeAnUnaryExpression(NOT_OP,e.thellnd);};
SgUnaryExp &SgSizeOfOp(SgExpression &e)
  {return makeAnUnaryExpression(SIZE_OP,e.thellnd);};


// Add type-checking here.
SgUnaryExp &
makeAnUnaryExpression(int code,PTR_LLND ll1)
{
  PTR_LLND ll;
  SgUnaryExp *pt = NULL;

  ll = newExpr(code,NODE_TYPE(ll1),ll1);
  pt = new SgUnaryExp(ll);
  return *pt;
};

SgUnaryExp * isSgUnaryExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DEREF_OP:
    case ADDRESS_OP:
    case SIZE_OP:
    case MINUS_OP:
    case UNARY_ADD_OP:
    case PLUSPLUS_OP:
    case MINUSMINUS_OP:
    case BIT_COMPLEMENT_OP:
    case NOT_OP:
      return (SgUnaryExp *) pt;
    default:
      return NULL;
    }
}

SgCastExp * isSgCastExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case CAST_OP:
      return (SgCastExp *) pt;
    default:
      return NULL;
    }
}

SgDeleteExp * isSgDeleteExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DELETE_OP:
      return (SgDeleteExp *) pt;
    default:
      return NULL;
    }
}

SgNewExp * isSgNewExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case NEW_OP:
      return (SgNewExp *) pt;
    default:
      return NULL;
    }
}

SgExpression & SgExprIfExp::conditional()
{// expr 1
   PTR_LLND ll;
   SgExpression *pt = NULL;
 
 ll = NODE_OPERAND0(thellnd);
 pt = GetMappingInTableForLlnd(ll);
 if (pt)
   return *pt;
 else
   {
     pt = new SgExpression(ll);      
   } 
 return *pt;
}

SgExpression & SgExprIfExp::trueExp()
{// expr 2
   PTR_LLND ll = NULL,ll2;
 SgExpression *pt = NULL;
 ll2 = NODE_OPERAND1(thellnd);
 if (ll2)
   ll = NODE_OPERAND0(ll2);
 else
   Message("pb in SgExprIfExp",0);
 pt = GetMappingInTableForLlnd(ll);
 if (pt)
   return *pt;
 else
   {
     pt = new SgExpression(ll);      
   } 
 return *pt;
}

SgExpression & SgExprIfExp::falseExp()
{// expr 3
   PTR_LLND ll = NULL,ll2;
 SgExpression *pt = NULL;
 ll2 = NODE_OPERAND1(thellnd);
 if (ll2)
   ll = NODE_OPERAND1(ll2);
 else
   Message("pb in SgExprIfExp",0);
 pt = GetMappingInTableForLlnd(ll);
 if (pt)
   return *pt;
 else
   {
     pt = new SgExpression(ll);      
   } 
 return *pt;
}

void SgExprIfExp::setTrueExp(SgExpression &t)
{
  PTR_LLND ll;
  ll = NODE_OPERAND1(thellnd);
  if (ll)
    NODE_OPERAND0(ll) = t.thellnd;
  else
    {
      NODE_OPERAND1(thellnd)= newExpr(EXPR_IF_BODY,NULL,t.thellnd,NULL);
    }
}

void SgExprIfExp::setFalseExp(SgExpression &f)
{
  PTR_LLND ll;
  ll = NODE_OPERAND1(thellnd);
  if (ll)
    NODE_OPERAND1(ll) = f.thellnd;
  else
    {
      NODE_OPERAND1(thellnd)= newExpr(EXPR_IF_BODY,NULL,NULL,f.thellnd);
    }
}

SgExprIfExp * isSgExprIfExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case EXPR_IF:
      return (SgExprIfExp *) pt;
    default:
      return NULL;
    }
}

SgFunctionCallExp * isSgFunctionCallExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case FUNC_CALL:
      return (SgFunctionCallExp *) pt;
    default:
      return NULL;
    }
}

SgFuncPntrExp * isSgFuncPntrExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case FUNCTION_OP:
      return (SgFuncPntrExp *) pt;
    default:
      return NULL;
    }
}


void SgExprListExp::linkToEnd(SgExpression &arg)
{
  PTR_LLND  lptr;
  lptr = Follow_Llnd(thellnd,2);
  NODE_OPERAND1(lptr) = arg.thellnd;
}


SgExprListExp * isSgExprListExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case EXPR_LIST:
      return (SgExprListExp *) pt;
    default:
      return NULL;
    }
}


SgProject::SgProject(char * proj_file_name)
{
  // first let init the library we need
    if (!proj_file_name)
      {
        Message("Cannot open project: no file specified",0);
#if 1
        exit(1);
#endif
      }
    if( open_proj_toolbox(proj_file_name, proj_file_name) < 0)
      {
        fprintf(stderr,"%s   ",proj_file_name);
        Message("Cannot open project",0);
#if 1
        exit(1);
#endif
      }
    Init_Tool_Box();

// we have to initialize some specific data for this interface 
  CurrentProject = this;
}


SgFile &SgProject::file(int i)
{
  PTR_FILE file;
  SgFile *pt = NULL;
  file = GetFileWithNum(i);
  SetCurrentFileTo(file);
  SwitchToFile(GetFileNumWithPt(file));
  if (!file)
    {
      Message("SgProject::file; File not found",0);
      return *pt;
    }
  pt =  GetMappingInTableForFile(file);
  if (pt)
    return *pt;
  else
    {
      pt = new SgFile(FILE_FILENAME(file));
      return *pt;
    }
}





void SgProject::addFile(char * dep_file_name)
{
  SORRY;
  return;
  
}


void SgProject::deleteFile(SgFile * file)
{
   SORRY;
   return;
   
}

SgFile::SgFile(char * dep_file_name)
{
  filept =  GetPointerOnFile(dep_file_name);
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  if (! filept)
    {
      Message("File not found in SgFile; added",0);
      if (CurrentProject)
        CurrentProject->addFile(dep_file_name);
    }
  SetMappingInTableForFile(filept, (void *) this);
}


     

SgStatement * SgFile::functions(int i)
{
  PTR_BFND bif;
  SgStatement *pt = NULL;

  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  bif = getFunctionNumHeader(i);
  if (!bif)
    {
      Message("SgFile::functions; Function not found",0);
      return pt;
    }
  pt = GetMappingInTableForBfnd(bif);
  if (pt)
    return pt;
  else
    {
      pt = new SgStatement(bif);      
    }  
  return pt;
}



SgStatement *SgFile::getStruct(int i)
{  
  PTR_BFND bif;
  SgStatement *pt = NULL;

  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  bif = getStructNumHeader(i);
  if (!bif)
    {
      Message("SgFile::getStruct; Struct not found",0);
        return pt;
    }
  pt = GetMappingInTableForBfnd(bif);
  if (pt)
    return pt;
  else
    {
      pt = new SgStatement(bif);      
    }  
  return pt;
}



 SgStatement::SgStatement(int variant)
{
  if (!isABifNode(variant))
    {
      Message("Attempt to create a bif node with a variant that is not",0);
      // arbitrary choice for the variant
       thebif = (PTR_BFND) newNode (BASIC_BLOCK);
    } else
      thebif = (PTR_BFND) newNode (variant);
  SetMappingInTableForBfnd(thebif, (void *) this);
}

void SgStatement::insertStmtAfter(SgStatement &s)
{ insertBfndListIn(s.thebif,thebif,NULL); }

void SgStatement::insertStmtAfter(SgStatement &s,SgStatement &cp)
{ insertBfndListIn(s.thebif,thebif,cp.thebif); }
  

 SgStatement::SgStatement(PTR_BFND bif)
{
  thebif = bif;
  SetMappingInTableForBfnd(thebif, (void *) this);
}


SgExpression * SgStatement::expr(int i)
{
  PTR_LLND ll;
  SgExpression *pt = NULL;
  switch (i)
    {
    case 0:
      ll = BIF_LL1(thebif);
      break;
    case 1:
      ll = BIF_LL2(thebif);
      break;
    case 2:
      ll = BIF_LL3(thebif);
      break;
    default:
       ll = BIF_LL1(thebif);
      Message("A bif node can only have 3 expressions (0,1,2)",BIF_LINE(thebif));
    }
  return LlndMapping(ll);
}




SgLabel *SgStatement::label()
{
  PTR_LABEL lab;
  SgLabel *pt = NULL;
  lab = BIF_LABEL(thebif);
  if (!lab)
    {
//      Message("The bif has no label",BIF_LINE(thebif));
      return pt;
    }
  pt = GetMappingInTableForLabel(lab);
  if (pt)
    return pt;
  else
    {
      pt = new SgLabel(lab);      
    }  
  return pt;
}



void  SgStatement::setExpression (int i, SgExpression &e)
{
  switch (i)
    {
    case 0:
      BIF_LL1(thebif) = e.thellnd;
      break;
    case 1:
      BIF_LL2(thebif) = e.thellnd ;
      break;
    case 2:
      BIF_LL3(thebif) = e.thellnd ;
      break;
    default:
      Message("A bif node can only have 3 expressions (0,1,2)",BIF_LINE(thebif));
    }
}
 
SgStatement * SgStatement::nextInChildList()
{
  PTR_BLOB blob;
  SgStatement *x;

  if (BIF_CP(thebif))
    {
      blob = lookForBifInBlobList(BIF_BLOB1(BIF_CP(thebif)), thebif);
      if (!blob)
        blob = lookForBifInBlobList(BIF_BLOB2(BIF_CP(thebif)), thebif);
      if (blob)
        blob = BLOB_NEXT(blob);
      if (blob)
        x = BfndMapping(BLOB_VALUE(blob));
      else x = NULL;
    }
  else
    x = NULL;

  return x;
   
}


char*  SgStatement::unparse()
{
//  UnparseBif(thebif);
//  SORRY;
//  return NULL;
    return(funparse_bfnd(thebif));
}


void  SgStatement::sunparse(char *buffer)
{
  UnparseBif(thebif);
  SORRY;
  return;
}



int  SgStatement::numberOfComments()  
{
  SORRY;
  return 0;
}     

void SgStatement::addComment(char *com)
{
  LibAddComment(thebif,com);
}

 
int  SgStatement::hasAnnotations()
{
  SORRY;
  return 0;
}

int  SgStatement::IsSymbolInScope(SgSymbol &symb)
{
 return  LibIsSymbolInScope(thebif,symb.thesymb);
}

int  SgStatement::IsSymbolReferenced(SgSymbol &symb)   
{
  return LibIsSymbolReferenced(thebif,symb.thesymb);
}   
  
SgExpression::SgExpression(int variant)
{
  if (!isALoNode(variant))
    {
      Message("Attempt to create a low level node with a variant that is not",0);
      // arbitrary choice for the variant
        thellnd = (PTR_LLND) newNode (EXPR_LIST);
    } else
      thellnd = (PTR_LLND) newNode (variant);
  SetMappingInTableForLlnd(thellnd, (void *) this);
}


SgExpression::SgExpression(PTR_LLND ll)
{
  thellnd = ll;
  SetMappingInTableForLlnd(thellnd, (void *) this);
}

 SgExpression::SgExpression(int variant, SgExpression &lhs, SgExpression &rhs,
                            SgSymbol &s, SgType &type)
{
  if (!isALoNode(variant))
    {
      Message("Attempt to create a low level node with a variant that is not",0);
      // arbitrary choice for the variant
        thellnd = (PTR_LLND) newNode (EXPR_LIST);
    } else
      thellnd = (PTR_LLND) newNode (variant);
  SetMappingInTableForLlnd(thellnd, (void *) this);
  NODE_OPERAND0(thellnd) = lhs.thellnd;
  NODE_OPERAND1(thellnd) = rhs.thellnd;
  NODE_SYMB(thellnd) = s.thesymb;
  NODE_TYPE(thellnd) = type.thetype;
    
}

SgExpression *SgExpression::operand(int i)
{
  PTR_LLND ll;
  SgExpression *pt = NULL;
  switch (i)
    {
    case 1:
      ll = NODE_OPERAND0(thellnd);
      break;
    case 2:
      ll =  NODE_OPERAND1(thellnd);
      break;
    default:
      ll = NODE_OPERAND0(thellnd);
      Message("A ll node can only have 2 child (1,2)",0);
    }
  return LlndMapping(ll);
}

char *SgExpression::unparse()
{
    char *string = funparse_llnd(thellnd);
    char *nl = strrchr((const char*)string, '\n');

    if (nl) {
        *nl = '\000';
    } 
    return (string);
}

#if 0
char *SgExpression::unparse()
{
  
  UnparseLLND(thellnd);
  SORRY;
  return NULL;
}
#endif /* 0 */

void SgExpression::sunparse(char *buffer)
{
  SORRY;
}

#define ERR_TOOMANYSYMS -1

int  SgExpression::linearRepresentation(int *coeff, SgSymbol **symb,int *cst, int size)
{
  PTR_SYMB ts[100];
  int i;
  if (!symb || !coeff || !cst)
    return 0;
  if (size > 100)
    {
      Message (" Too many symbols in linearRepresentation ",0);
      return ERR_TOOMANYSYMS;
    }
  for (i=0 ; i < size; i++)
    ts[i] = symb[i]->thesymb;
  return buildLinearRep(thellnd,coeff,ts,size,cst);
}



SgExpression *SgExpression::normalForm(int n, SgSymbol *s)
{
 SORRY;
 return (SgExpression *) NULL;
 
}
SgExpression *SgExpression::coefficient(SgSymbol &s)
{
  SORRY;
  return (SgExpression *) NULL;
  
}

int SgExpression::isInteger()
{
  int *res;
  int resul = 0;
  res = evaluateExpression (thellnd);
  if (res[0] != -1)
    {
      resul = 1;
    }
  delete res;
  return resul;
}

int SgExpression::valueInteger()
{
  int *res;
  int resul = 0;
  res = evaluateExpression (thellnd);
  if (res[0] != -1)
    {
      resul = res[1];
    }
  delete res;
  return resul;
}

SgExpression &
makeAnBinaryExpression(int code,SgExpression *ll1,SgExpression *ll2)
{
  SgExpression *resul = NULL;
  if (ll1 && ll2)
    return *LlndMapping(newExpr(code,NODE_TYPE(ll1->thellnd),ll1->thellnd,ll2->thellnd));
  else
    if (ll1) 
      return *LlndMapping(newExpr(code,NODE_TYPE(ll1->thellnd),ll1->thellnd,NULL));
    else
      if (ll2)
        return *LlndMapping(newExpr(code,NODE_TYPE(ll2->thellnd),NULL,ll2->thellnd));
      else
        return *LlndMapping(newExpr(code,NULL,NULL,NULL));
  return *resul;
}


SgExpression &
makeAnBinaryExpression(int code,PTR_LLND ll1,PTR_LLND ll2)
{

  return *LlndMapping(newExpr(code,NODE_TYPE(ll1),ll1,ll2));
}

SgExpression &operator + ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(ADD_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator - ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(SUBT_OP,lhs.thellnd,rhs.thellnd);}
 
SgExpression &operator * ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(MULT_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator / ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(DIV_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator % ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(MOD_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator <<( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(LSHIFT_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator >>( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(RSHIFT_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator < ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(LT_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator > ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(GT_OP,lhs.thellnd,rhs.thellnd);} 


SgExpression &operator <= ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(LE_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator >= ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(GE_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator & ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(BITAND_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator | ( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(BITOR_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator &&( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(AND_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator ||( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(OR_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator +=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(PLUS_ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator &=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(AND_ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator *=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(MULT_ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator /=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(DIV_ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator %=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(MOD_ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator ^=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(XOR_ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator <<=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(LSHIFT_ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &operator >>=( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(RSHIFT_ASSGN_OP,lhs.thellnd,rhs.thellnd);}

SgExpression &SgAssignOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(ASSGN_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &    SgEqOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(EQ_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &   SgNeqOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(NE_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &SgExprListOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(EXPR_LIST,lhs.thellnd,rhs.thellnd);} 

SgExpression &  SgRecRefOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(RECORD_REF,lhs.thellnd,rhs.thellnd);} 

SgExpression & SgPointStOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(POINTST_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &   SgScopeOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(SCOPE_OP,lhs.thellnd,rhs.thellnd);} 

SgExpression &    SgDDotOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(DDOT,lhs.thellnd,rhs.thellnd);} 

SgExpression & SgBitNumbOp( SgExpression &lhs, SgExpression &rhs)
{return makeAnBinaryExpression(BIT_NUMBER,lhs.thellnd,rhs.thellnd);} 






// For correctness of symbol creation, it is
// necessary to have a symbol table of some form to
// ensure there are no duplicate symbols being 
// created.
SgSymbol::SgSymbol(int variant, char *name)
{
  if (!isASymbNode(variant))
    {
      Message("Attempt to create a symbol node with a variant that is not",0);
      // arbitrary choice for the variant
        thesymb = newSymbol (VARIABLE_NAME, name, NULL);
    } else
      thesymb = newSymbol (variant, name, NULL);

  SYMB_TYPE(thesymb) = GetAtomicType(T_INT);
  SetMappingInTableForSymb(thesymb, (void *) this);
}

SgSymbol::SgSymbol(int variant)
{
  if (!isASymbNode(variant))
    {
      Message("Attempt to create a symbol node with a variant that is not",0);
      // arbitrary choice for the variant
        thesymb = newSymbol (VARIABLE_NAME, NULL, NULL);
    } else
      thesymb = newSymbol (variant, NULL, NULL);
  SYMB_TYPE(thesymb) = GetAtomicType(T_INT);
  SetMappingInTableForSymb(thesymb, (void *) this);
}

 SgSymbol::SgSymbol(PTR_SYMB symb)
{
  thesymb = symb;
  SetMappingInTableForSymb(thesymb, (void *) this);
}

SgSymbol::SgSymbol(int variant, char *identifier, SgType &t, SgStatement &scope)
{
  if (!isASymbNode(variant))
    {
      Message("Attempt to create a symbol node with a variant that is not",0);
      // arbitrary choice for the variant
        thesymb = newSymbol (VARIABLE_NAME, identifier, NULL);
    } else
      thesymb = newSymbol (variant, identifier, NULL);

  SYMB_TYPE(thesymb) = t.thetype;
  SYMB_SCOPE(thesymb) = scope.thebif;
  SetMappingInTableForSymb(thesymb, (void *) this);
}

SgSymbol::SgSymbol(int variant, char *identifier, SgStatement &scope)
{
  if (!isASymbNode(variant))
    {
      Message("Attempt to create a symbol node with a variant that is not",0);
      // arbitrary choice for the variant
        thesymb = newSymbol (VARIABLE_NAME, identifier, NULL);
    } else
      thesymb = newSymbol (variant, identifier, NULL);

  SYMB_TYPE(thesymb) = GetAtomicType(T_INT);
  SYMB_SCOPE(thesymb) = scope.thebif;
  SetMappingInTableForSymb(thesymb, (void *) this);
}


SgStatement *SgSymbol::declaredInStmt()
{
  return BfndMapping(LibWhereIsSymbDeclare(thesymb));
  
}

int SgSymbol::attributes()
{
  return SYMB_ATTR(thesymb);
}

void SgSymbol::setAttribute(int attribute)
{
  SORRY;
  return;
  
}
void SgSymbol::removeAttribute(int attribute)
{
  SORRY;
  return;
  
}

SgStatement *SgSymbol::body()
{
  PTR_BFND  bif = NULL;
  PTR_TYPE type;
  // there is a function low_level.c that does it.
  if ((SYMB_CODE(thesymb) == COLLECTION_NAME) ||
      (SYMB_CODE(thesymb) == CLASS_NAME) ||
      (SYMB_CODE(thesymb) == TECLASS_NAME))
    {
      type = SYMB_TYPE(thesymb);
      if (type)
        {
          bif = TYPE_COLL_ORI_CLASS(type);
        } else
          {
            Message("Body of collection or class not found",0);
          }
    } else
      if ((SYMB_CODE(thesymb) == FUNCTION_NAME) ||
          (SYMB_CODE(thesymb) == PROGRAM_NAME) ||
          (SYMB_CODE(thesymb) == PROCEDURE_NAME) ||
          (SYMB_CODE(thesymb) == MEMBER_FUNC))
        {
          bif = SYMB_FUNC_HEDR(thesymb);
        } else
          {
            Message("Body not found, may not be implemented yet",0);
            SORRY;
          }	
  return BfndMapping(bif);
}





SgType::SgType(int variant)
{
  if (!isATypeNode(variant))
    {
      Message("Attempt to create a type node with a variant that is not",0);
      // arbitrary choice for the variant
       thetype = (PTR_TYPE) newNode (T_INT);
    } else
      thetype = (PTR_TYPE) newNode (variant);
  SetMappingInTableForType(thetype, (void *) this);
}



SgType *SgTypeInt()
{
  return TypeMapping(GetAtomicType(T_INT));
}


SgType *SgTypeChar()
{
  return TypeMapping(GetAtomicType(T_CHAR));
}

SgType *SgTypeFloat()
{
   return TypeMapping(GetAtomicType(T_FLOAT));
}

SgType *SgTypeDouble()
{
  return TypeMapping(GetAtomicType(T_DOUBLE));
}

SgType *SgTypeVoid()
{
  return TypeMapping(GetAtomicType(T_VOID));
}

SgType *SgTypeBool()
{
  return TypeMapping(GetAtomicType(T_BOOL));
}

SgType *SgTypeDefault()
{
  return TypeMapping(GetAtomicType(DEFAULT));
}



//
//
// Subclass for reference to symbol
//
//


SgRefExp * isSgRefExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case CONST_REF:
    case TYPE_REF:
    case INTERFACE_REF:
      return (SgRefExp *) pt;
    default:
      return NULL;
    }
}

SgExpression * SgVarRefExp::progatedValue()
 {
   SORRY;     // if scalar propogation worked
   return (SgExpression *) NULL;
 }


SgVarRefExp * isSgVarRefExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case VAR_REF:
      return (SgVarRefExp *) pt;
    default:
      return NULL;
    }
}

SgThisExp * isSgThisExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case THIS_NODE:
      return (SgThisExp *) pt;
    default:
      return NULL;
    }
}


SgArrayRefExp * isSgArrayRefExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case ARRAY_REF:
      return (SgArrayRefExp *) pt;
    default:
      return NULL;
    }
}



SgPntrArrRefExp * isSgPntrArrRefExp(SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case ARRAY_OP:
      return (SgPntrArrRefExp *) pt;
    default:
      return NULL;
    }
}

SgPointerDerefExp * isSgPointerDerefExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DEREF_OP:
      return (SgPointerDerefExp *) pt;
    default:
      return NULL;
    }
}


SgRecordRefExp * isSgRecordRefExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case RECORD_REF:
      return (SgRecordRefExp *) pt;
    default:
      return NULL;
    }
}

SgStructConstExp* isSgStructConstExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case STRUCTURE_CONSTRUCTOR:
      return (SgStructConstExp *) pt;
    default:
      return NULL;
    }
}

SgConstExp* isSgConstExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case CONSTRUCTOR_REF:
      return (SgConstExp *) pt;
    default:
      return NULL;
    }
}


SgVecConstExp * isSgVecConstExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case VECTOR_CONST:
      return (SgVecConstExp *) pt;
    default:
      return NULL;
    }
}

SgInitListExp * isSgInitListExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case INIT_LIST:
      return (SgInitListExp *) pt;
    default:
      return NULL;
    }
}

SgObjectListExp * isSgObjectListExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case EQUI_LIST:
    case NAMELIST_LIST:
    case COMM_LIST:
      return (SgObjectListExp *) pt;
    default:
      return NULL;
    }
}


SgAttributeExp * isSgAttributeExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case PARAMETER_OP:
    case PUBLIC_OP:
    case PRIVATE_OP:
    case ALLOCATABLE_OP:
    case DIMENSION_OP:
    case EXTERNAL_OP:
    case IN_OP:
    case OUT_OP:
    case INOUT_OP:
    case INTRINSIC_OP:
    case POINTER_OP:
    case OPTIONAL_OP:
    case SAVE_OP:
    case TARGET_OP:
      return (SgAttributeExp *) pt;
    default:
      return NULL;
    }
}



SgKeywordArgExp * isSgKeywordArgExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case KEYWORD_ARG:
      return (SgKeywordArgExp *) pt;
    default:
      return NULL;
    }
}

SgSubscriptExp* isSgSubscriptExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DDOT:
      return (SgSubscriptExp *) pt;
    default:
      return NULL;
    }
}

SgUseOnlyExp * isSgUseOnlyExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case ONLY_NODE:
      return (SgUseOnlyExp *) pt;
    default:
      return NULL;
    }
}

SgUseRenameExp * isSgUseRenameExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case RENAME_NODE:
      return (SgUseRenameExp *) pt;
    default:
      return NULL;
    }
}


SgSpecPairExp * isSgSpecPairExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case SPEC_PAIR:
      return (SgSpecPairExp *) pt;
    default:
      return NULL;
    }
}

SgIOAccessExp * isSgIOAccessExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case IOACCESS:
      return (SgIOAccessExp *) pt;
    default:
      return NULL;
    }
}


SgImplicitTypeExp * isSgImplicitTypeExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case IMPL_TYPE:
      return (SgImplicitTypeExp *) pt;
    default:
      return NULL;
    }
}

SgTypeExp * isSgTypeExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case TYPE_OP:
      return (SgTypeExp *) pt;
    default:
      return NULL;
    }
}

SgSeqExp * isSgSeqExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case SEQ:
      return (SgSeqExp *) pt;
    default:
      return NULL;
    }
}

SgStringLengthExp * isSgStringLengthExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case LEN_OP:
      return (SgStringLengthExp *) pt;
    default:
      return NULL;
    }
}

SgDefaultExp * isSgDefaultExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DEFAULT:
      return (SgDefaultExp *) pt;
    default:
      return NULL;
    }
}


SgLabelRefExp * isSgLabelRefExp (SgExpression *pt)
{

  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case LABEL_REF:
      return (SgLabelRefExp *) pt;
    default:
      return NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                                                                           //
// We add the subclass for statements  here.                                 //
// Need more comment and so on ........                                      //
// Reorganizing that file may be necessary sometimes                         //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



SgProgHedrStmt * isSgProgHedrStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROG_HEDR:
      return (SgProgHedrStmt *) pt;
    default:
      return NULL;
    }
}

SgProcHedrStmt * isSgProcHedrStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROC_HEDR:
      return (SgProcHedrStmt *) pt;
    default:
      return NULL;
    }
}


SgFuncHedrStmt * isSgFuncHedrStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case FUNC_HEDR:
      return (SgFuncHedrStmt *) pt;
    default:
      return NULL;
    }
}

#if 0
class  SgModuleStmt: public SgStatement{
        // Fortran 90 Module statement
        // variant ==  MODULE_STMT
     public:
        SgModuleStmt(SgSymbol &moduleName, SgStatement &body):SgStatement(MODULE_STMT)
          {
            SORRY;
          };
        SgModuleStmt(SgSymbol &moduleName):SgStatement(PROG_HEDR)
          {
            SORRY;
          };
        ~SgModuleStmt(){RemoveFromTableBfnd((void *) this);};

        SgSymbol *moduleName()
          {
            SORRY;
          };               // module name 
        void setName(SgSymbol &symbol)
          {
            SORRY;
          };        // set module name 

        int numberOfSpecificationStmts()
          {
            SORRY;
          };
        int numberOfRoutinesDefined()
          {
            SORRY;
          };
        int numberOfFunctionsDefined()
          {
            SORRY;
          };
        int numberOfSubroutinesDefined()
          {
            SORRY;
          };

        SgStatement *specificationStmt(int i)
          {
            SORRY;
          };
        SgStatement *routine(int i)
          {
            SORRY;
          };
        SgStatement *function(int i)
          {
            SORRY;
          };
        SgStatement *subroutine(int i)
          {
            SORRY;
          };

        int isSymbolInScope(SgSymbol &symbol)
          {
            SORRY;
          };
        int isSymbolDeclaredHere(SgSymbol &symbol)
          {
            SORRY;
          };

        SgSymbol &addVariable(SgType &T, char *name)
          {
            SORRY;
          }; 
                                        //add a declaration for new variable

        SgStatement *addCommonBlock(char *blockname, int noOfVars,
                                    SgSymbol *Vars)
          {
            SORRY;
          }; // add a new common block
};


SgModuleStmt * isSgModuleStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case MODULE_STMT:
      return (SgModuleStmt *) pt;
    default:
      return NULL;
    }
}


class  SgInterfaceStmt: public SgStatement{
        // Fortran 90 Operator Interface Statement
        // variant == INTERFACE_STMT
     public:
        SgInterfaceStmt(SgSymbol &name, SgStatement &body, SgStatement &scope):SgStatement(INTERFACE_STMT)
          {
            SORRY;
          };
        ~SgInterfaceStmt(){RemoveFromTableBfnd((void *) this);};

        SgSymbol *interfaceName()
          {
            SORRY;
          };               // interface name if given
        int setName(SgSymbol &symbol)
          {
            SORRY;
          };           // set interface name 

        int numberOfSpecificationStmts()
          {
            SORRY;
          };

        SgStatement *specificationStmt(int i)
          {
            SORRY;
          };

        int isSymbolInScope(SgSymbol &symbol)
          {
            SORRY;
          };
        int isSymbolDeclaredHere(SgSymbol &symbol)
          {
            SORRY;
          };
};


SgInterfaceStmt * isSgInterfaceStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case INTERFACE_STMT:
      return (SgInterfaceStmt *) pt;
    default:
      return NULL;
    }
}


class  SgBlockDataStmt: public SgStatement{
        // Fortran Block Data statement
        // variant == BLOCK_DATA
     public:
        SgBlockDataStmt(SgSymbol &name, SgStatement &body):SgStatement(BLOCK_DATA)
          {
            BIF_SYMB(thebif) = name.thesymb;
            insertBfndListIn(body.thebif,thebif,thebif);            
          };
        ~SgBlockDataStmt(){RemoveFromTableBfnd((void *) this);};

        SgSymbol *name()  // block data name if given
          { return SymbMapping(BIF_SYMB(thebif)); };
        int setName(SgSymbol &symbol)
          { 
            BIF_SYMB(thebif) = symbol.thesymb;
            return 1;
          };           // set block data name 

        int isSymbolInScope(SgSymbol &symbol)
          {
            SORRY;
          };
        int isSymbolDeclaredHere(SgSymbol &symbol)
          {
            SORRY;
          };
};



SgBlockDataStmt * isSgBlockDataStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case BLOCK_DATA:
      return (SgBlockDataStmt *) pt;
    default:
      return NULL;
    }
}
#endif

SgClassStmt * isSgClassStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case CLASS_DECL:
    case TECLASS_DECL:
    case STRUCT_DECL:
    case UNION_DECL:
    case ENUM_DECL:
    case COLLECTION_DECL:
      return (SgClassStmt *) pt;
    default:
      return NULL;
    }
}


SgStructStmt * isSgStructStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case STRUCT_DECL:
      return (SgStructStmt *) pt;
    default:
      return NULL;
    }
}


SgUnionStmt * isSgUnionStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case UNION_DECL:
      return (SgUnionStmt *) pt;
    default:
      return NULL;
    }
}

SgEnumStmt * isSgEnumStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case ENUM_DECL:
      return (SgEnumStmt *) pt;
    default:
      return NULL;
    }
}

SgCollectionStmt * isSgCollectionStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case COLLECTION_DECL:
      return (SgCollectionStmt *) pt;
    default:
      return NULL;
    }
}


SgBasicBlockStmt * isSgBasicBlockStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case BASIC_BLOCK:
      return (SgBasicBlockStmt *) pt;
    default:
      return NULL;
    }
}



SgForStmt * isSgForStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case FOR_NODE :
      return (SgForStmt *) pt;
    default:
      return NULL;
    }
}

SgProcessDoStmt * isSgProcessDoStmt (SgStatement *pt)
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROCESS_DO_STAT :
      return (SgProcessDoStmt *) pt;
    default:
      return NULL;
    }
}

SgWhileStmt * isSgWhileStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case WHILE_NODE:
      return (SgWhileStmt *) pt;
    default:
      return NULL;
    }
}

SgDoWhileStmt * isSgDoWhileStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case DO_WHILE_NODE:
      return (SgDoWhileStmt *) pt;
    default:
      return NULL;
    }
}

SgLogIfStmt * isSgLogIfStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case LOGIF_NODE:
      return (SgLogIfStmt *) pt;
    default:
      return NULL;
    }
}


SgIfStmt * isSgIfStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case IF_NODE:
      return (SgIfStmt *) pt;
    default:
      return NULL;
    }
}
#if 0
class  SgIfElseIfStmt: public SgIfStmt {
        // For Fortran if then elseif .. elseif ... case
        // variant == ELSEIF_NODE
     public:
        SgIfElseIfStmt(SgExpression &condList, SgStatement &blockList,
                       SgSymbol &constructName):SgIfStmt(ELSEIF_NODE)
          {
            SORRY;
          }; 
        int numberOfConditionals()
          {
            SORRY;
          };       // the number of conditionals
        SgStatement *body(int b)
          {
            SORRY;
          };          // block b
        void setBody(int b)
          {
            SORRY;
          };              // sets block 
        SgExpression *conditional(int i)
          {
            SORRY;
          }; // the i-th conditional
        void setConditional(int i)
          {
            SORRY;
          };       // sets the i-th conditional
        void addClause(SgExpression &cond, SgStatement &block)
          {
            SORRY;
          };
        void removeClause(int b)
          {
            SORRY;
          };          // removes block b and it's conditional

};


SgIfElseIfStmt * isSgIfElseIfStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case ELSEIF_NODE:
      return (SgIfElseIfStmt *) pt;
    default:
      return NULL;
    }
}
#endif

SgArithIfStmt * isSgArithIfStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case ARITHIF_NODE:
      return (SgArithIfStmt *) pt;
    default:
      return NULL;
    }
}

SgWhereStmt * isSgWhereStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case WHERE_NODE:
      return (SgWhereStmt *) pt;
    default:
      return NULL;
    }
}


SgWhereBlockStmt * isSgWhereBlockStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case WHERE_BLOCK_STMT:
      return (SgWhereBlockStmt *) pt;
    default:
      return NULL;
    }
}


SgSwitchStmt * isSgSwitchStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case SWITCH_NODE:
      return (SgSwitchStmt *) pt;
    default:
      return NULL;
    }
}



SgCaseOptionStmt * isSgCaseOptionStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case CASE_NODE:
      return (SgCaseOptionStmt *) pt;
    default:
      return NULL;
    }
}

// ******************** Leaf Executable Nodes ***********************


SgExecutableStatement * isSgExecutableStatement (SgStatement *pt)
{
  if (!pt)
    return NULL;
  if (!isADeclBif(BIF_CODE(pt->thebif)))
    return (SgExecutableStatement *) pt;
  else
    return NULL;
}

SgAssignStmt * isSgAssignStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case ASSIGN_STAT:
      return (SgAssignStmt *) pt;
    default:
      return NULL;
    }
}


SgCExpStmt * isSgCExpStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case EXPR_STMT_NODE:
      return (SgCExpStmt *) pt;
    default:
      return NULL;
    }
}


SgPointerAssignStmt * isSgPointerAssignStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case POINTER_ASSIGN_STAT:
      return (SgPointerAssignStmt *) pt;
    default:
      return NULL;
    }
}

SgHeapStmt * isSgHeapStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case ALLOCATE_STMT:
    case DEALLOCATE_STMT:
      return (SgHeapStmt *) pt;
    default:
      return NULL;
    }
}
   
SgNullifyStmt * isSgNullifyStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case NULLIFY_STMT:
      return (SgNullifyStmt *) pt;
    default:
      return NULL;
    }
}

SgContinueStmt * isSgContinueStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case CONT_STAT:
      return (SgContinueStmt *) pt;
    default:
      return NULL;
    }
}
   

SgControlEndStmt * isSgControlEndStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case CONTROL_END :
      return (SgControlEndStmt *) pt;
    default:
      return NULL;
    }
}


SgBreakStmt * isSgBreakStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case BREAK_NODE:
      return (SgBreakStmt *) pt;
    default:
      return NULL;
    }
}


SgCycleStmt * isSgCycleStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case CYCLE_STMT:
      return (SgCycleStmt *) pt;
    default:
      return NULL;
    }
}


SgReturnStmt * isSgReturnStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case RETURN_NODE:
    case RETURN_STAT:
      return (SgReturnStmt *) pt;
    default:
      return NULL;
    }
}

SgExitStmt * isSgExitStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case EXIT_STMT:
      return (SgExitStmt *) pt;
    default:
      return NULL;
    }
}

SgGotoStmt * isSgGotoStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case GOTO_NODE:
      return (SgGotoStmt *) pt;
    default:
      return NULL;
    }
}


SgLabelListStmt * isSgLabelListStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case COMGOTO_NODE:
    case  ASSGOTO_NODE:
      return (SgLabelListStmt *) pt;
    default:
//      SORRY;
      return NULL;
    }
}


SgAssignedGotoStmt * isSgAssignedGotoStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case ASSGOTO_NODE:
      return (SgAssignedGotoStmt *) pt;
    default:
      return NULL;
    }
}

SgComputedGotoStmt * isSgComputedGotoStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case COMGOTO_NODE:
      return (SgComputedGotoStmt *) pt;
    default:
      return NULL;
    }
}

SgStopOrPauseStmt * isSgStopOrPauseStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case STOP_STAT:
      return (SgStopOrPauseStmt *) pt;
    default:
      return NULL;
    }
}

SgCallStmt* isSgCallStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROC_STAT:
      return (SgCallStmt *) pt;
    default:
      return NULL;
    }
}

SgProsHedrStmt* isSgProsHedrStmt (SgStatement *pt) /* Fortran M */
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROS_HEDR:
      return (SgProsHedrStmt *) pt;
    default:
      return NULL;
    }
}

SgProsCallStmt* isSgProsCallStmt (SgStatement *pt) /* Fortran M */
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROS_STAT:
      return (SgProsCallStmt *) pt;
    default:
      return NULL;
    }
}

SgProsCallLctn* isSgProsCallLctn (SgStatement *pt) /* Fortran M */
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROS_STAT_LCTN:
      return (SgProsCallLctn *) pt;
    default:
      return NULL;
    }
}

SgProsCallSubm* isSgProsCallSubm (SgStatement *pt) /* Fortran M */
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROS_STAT_SUBM:
      return (SgProsCallSubm *) pt;
    default:
      return NULL;
    }
}

SgMPCallStmt* isSgMPCallStmt (SgStatement *pt) /* Fortran M */
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case MP_STAT:
      return (SgMPCallStmt *) pt;
    default:
      return NULL;
    }
}

SgMPCallLctn* isSgMPCallLctn (SgStatement *pt) /* Fortran M */
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case MP_STAT_LCTN:
      return (SgMPCallLctn *) pt;
    default:
      return NULL;
    }
}

SgMPCallSubm* isSgMPCallSubm (SgStatement *pt) /* Fortran M */
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case MP_STAT_SUBM:
      return (SgMPCallSubm *) pt;
    default:
      return NULL;
    }
}

SgIOStmt * isSgIOStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case 0:
      return (SgIOStmt *) pt;
    default:
      SORRY;
      return NULL;
    }
}


SgInputOutputStmt * isSgInputOutputStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case READ_STAT:
    case WRITE_STAT:
    case PRINT_STAT:
      return (SgInputOutputStmt *) pt;
    default:
      return NULL;
    }
}

SgIOControlStmt::SgIOControlStmt(int variant, SgExpression &controlSpecifierList):SgExecutableStatement(variant)
{
  switch (variant){
  case OPEN_STAT:
  case CLOSE_STAT:
  case INQUIRE_STAT:
  case BACKSPACE_STAT:
  case REWIND_STAT:
  case ENDFILE_STAT:
  case FORMAT_STAT:
    break;
  default:
    Message("illegal variant for SgIOControlStmt",0);
  }
  
  BIF_LL2(thebif) = controlSpecifierList.thellnd;
}

SgIOControlStmt * isSgIOControlStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
      case OPEN_STAT:
    case CLOSE_STAT:
    case INQUIRE_STAT:
    case BACKSPACE_STAT:
    case REWIND_STAT:
    case ENDFILE_STAT:
    case FORMAT_STAT:
      return (SgIOControlStmt *) pt;
    default:
      return NULL;
    }
}

// ******************** Declaration Nodes ***************************

SgDeclarationStatement * isSgDeclarationStatement (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case VAR_DECL:
    case ENUM_DECL:
    case STRUCT_DECL:
    case CLASS_DECL:
    case TECLASS_DECL:
    case COLLECTION_DECL:
      return (SgDeclarationStatement *) pt;
    default:
      return NULL;
    }
}


// sets the initial value ofthe i-th variable
// an alternative way to initialize variables. The low-level node 
// (VAR_REF or ARRAY_REF) is replaced by a ASSIGN_OP low-level node.
void SgVarDeclStmt::setInitialValue(int i, SgExpression &initVal) // sets the initial value ofthe i-th variable
{
  int j;
  SgExpression *list, *varRef;
  list = this->expr(0);
  for(j = 0; j < i; j++) if(list) list = list->rhs();
  if(!list) return;
  varRef = list->lhs();
  if(!varRef) return;
  if(varRef->variant() == ASSGN_OP){
	 varRef->setRhs(initVal);
	 return;
	}
  SgExpression &e = SgAssignOp(*varRef, initVal);
  list->setLhs(e);
  return;
}
  
// method below contributed by Michael Golden 
// removes the initial value of the i-ith declaration
 void SgVarDeclStmt::clearInitialValue(int i)
 {
     int j;
     SgExpression *list, *varRef;  
 
     list = this->expr(0);
     for(j = 0; j < i; j++) 
       if (list) 
 	list = list->rhs();
     if(!list) 
       return;
     varRef = list->lhs();
     if(!varRef) 
       return;
     
     /* If there is an assignment here, then change it to just the LHS */
     /* Which is the variable itself                                   */
     if (varRef->variant() == ASSGN_OP) 
       list->setLhs(*(varRef->lhs()));
 
 
 }


SgVarDeclStmt * isSgVarDeclStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case VAR_DECL:
      return (SgVarDeclStmt *) pt;
    default:
      return NULL;
    }
}


SgIntentStmt * isSgIntentStmt (SgStatement *pt) /* Fortran M */
{
 
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case INTENT_STMT:
      return (SgIntentStmt *) pt;
    default:
      return NULL;
    }
}


SgVarListDeclStmt::SgVarListDeclStmt(int variant, SgExpression &symbolRefList):SgDeclarationStatement(variant)
          {
            switch (variant) {
              case INTENT_STMT:
              case OPTIONAL_STMT:
              case SAVE_DECL:
              case PUBLIC_STMT:
              case PRIVATE_STMT:
              case EXTERN_STAT:
              case INTRIN_STAT:
              case DIM_STAT:
              case ALLOCATABLE_STMT:
              case POINTER_STMT:
              case TARGET_STMT:
              case MODULE_PROC_STMT:
	      case PROCESSORS_STAT:
	      case DISTRIBUTE_STMT:
                   break;
              default:
                   Message("Illegal variant for SgVarListDeclStmt",0);
            };

//            findStatementAttribute(variant, attribute);
//            BIF_LL1(thesymb) = symbolrefList.thellnd;
//            setSymbolAttributesInVarRefList(BIF_LL1(thesymb));
            SORRY;
          }

SgVarListDeclStmt::SgVarListDeclStmt(int variant, SgSymbol &symbolList, SgStatement &scope):SgDeclarationStatement(variant)
          {
            switch (variant) {
              case INTENT_STMT:
              case OPTIONAL_STMT:
              case SAVE_DECL:
              case PUBLIC_STMT:
              case PRIVATE_STMT:
              case EXTERN_STAT:
              case INTRIN_STAT:
              case DIM_STAT:
              case ALLOCATABLE_STMT:
              case POINTER_STMT:
              case TARGET_STMT:
              case MODULE_PROC_STMT:
                   break;
              default:
                   Message("Illegal variant for SgVarListDeclStmt",0);
            };

//            findStatementAttribute(variant,attribute);
//            BIF_LL1(thesymb) = symbolList.thellnd;
//            setSymbolAttributesInVarRefList(BIF_LL1(thesymb));
            SORRY;
          }

SgDataStmt * isSgDataStmt(SgStatement *pt)
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case DATA_DECL:
      return (SgDataStmt *) pt;
    default:
      return NULL;
    }
}

SgDataPairExp * isSgDataPairExp(SgExpression *pt)
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DATA_PAIR:
      return (SgDataPairExp *) pt;
    default:
      return NULL;
    }
}


SgVarListDeclStmt * isSgVarListDeclStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case INTENT_STMT:
    case OPTIONAL_STMT:
    case SAVE_DECL:
    case PUBLIC_STMT:
    case PRIVATE_STMT:
    case EXTERN_STAT:
    case INTRIN_STAT:
    case DIM_STAT:
    case ALLOCATABLE_STMT:
    case POINTER_STMT:
    case TARGET_STMT:
    case MODULE_PROC_STMT:
    case PROCESSORS_STAT:
    case DISTRIBUTE_STMT:
      return (SgVarListDeclStmt *) pt;
    default:
      return NULL;
    }
}
     
     

SgStructureDeclStmt * isSgStructureDeclStmtSgStructureDeclStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case STRUCT_DECL:
      return (SgStructureDeclStmt *) pt;
    default:
      return NULL;
    }
}

SgNestedVarListDeclStmt::SgNestedVarListDeclStmt(int variant, SgExpression &listOfVarList):SgDeclarationStatement(VAR_DECL)
{
  int listVariant;
  
  switch (variant) {
  case NAMELIST_STAT:
    listVariant = NAMELIST_LIST;
    break;
  case EQUI_STAT:
    listVariant = EQUI_LIST;
    break;
  case COMM_STAT:
    listVariant = COMM_LIST;
    break;
  case PROS_COMM:  /* Fortran M */
    listVariant = COMM_LIST;
    break;
  default:
    Message("Illegal variant in SgNestedVarListDeclStmt",0);
  };
  BIF_CODE(thebif) = variant;
//            checkIfListOfVariant(listVariant, listOfVarList);            
//  SORRY;
  BIF_LL1(thebif) = listOfVarList.thellnd;
}

SgNestedVarListDeclStmt * isSgNestedVarListDeclStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case NAMELIST_STAT:
    case EQUI_STAT:
    case PROS_COMM:
    case COMM_STAT:
      return (SgNestedVarListDeclStmt *) pt;
    default:
      return NULL;
    }
}

     

SgParameterStmt * isSgParameterStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PARAM_DECL:
      return (SgParameterStmt *) pt;
    default:
      return NULL;
    }
}
    

SgImplicitStmt * isSgImplicitStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case IMPL_DECL:
      return (SgImplicitStmt *) pt;
    default:
      return NULL;
    }
}


SgInportStmt * isSgInportStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case INPORT_DECL:
      return (SgInportStmt *) pt;
    default:
      return NULL;
    }
}
 
 
SgOutportStmt * isSgOutportStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case OUTPORT_DECL:
      return (SgOutportStmt *) pt;
    default:
      return NULL;
    }
}
 
 
SgChannelStmt * isSgChannelStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case CHANNEL_STAT:
      return (SgChannelStmt *) pt;
    default:
      return NULL;
    }
}


SgMergerStmt * isSgMergerStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case MERGER_STAT:
      return (SgMergerStmt *) pt;
    default:
      return NULL;
    }
}
 

SgMoveportStmt * isSgMoveportStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case MOVE_PORT:
      return (SgMoveportStmt *) pt;
    default:
      return NULL;
    }
}
 

SgSendStmt * isSgSendStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case SEND_STAT:
      return (SgSendStmt *) pt;
    default:
      return NULL;
    }
}
 

SgReceiveStmt * isSgReceiveStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case RECEIVE_STAT:
      return (SgReceiveStmt *) pt;
    default:
      return NULL;
    }
}
 

SgEndchannelStmt * isSgEndchannelStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case ENDCHANNEL_STAT:
      return (SgEndchannelStmt *) pt;
    default:
      return NULL;
    }
}
 

SgProbeStmt * isSgProbeStmt(SgStatement *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case PROBE_STAT:
      return (SgProbeStmt *) pt;
    default:
      return NULL;
    }
}

SgProcessorsRefExp * isSgProcessorsRefExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case PROCESSORS_REF:
      return (SgProcessorsRefExp *) pt;
    default:
      return NULL;
    }
}


SgPortTypeExp * isSgPortTypeExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case PORT_TYPE_OP:
    case INPORT_TYPE_OP:
    case OUTPORT_TYPE_OP:
      return (SgPortTypeExp *) pt;
    default:
      return NULL;
    }
}

SgInportExp * isSgInportExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case INPORT_NAME:
      return (SgInportExp *) pt;
    default:
      return NULL;
    }
}

SgOutportExp * isSgOutportExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case OUTPORT_NAME:
      return (SgOutportExp *) pt;
    default:
      return NULL;
    }
}

SgFromportExp * isSgFromportExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case FROMPORT_NAME:
      return (SgFromportExp *) pt;
    default:
      return NULL;
    }
}

SgToportExp * isSgToportExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case TOPORT_NAME:
      return (SgToportExp *) pt;
    default:
      return NULL;
    }
}

SgIO_statStoreExp * isSgIO_statStoreExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case IOSTAT_STORE:
      return (SgIO_statStoreExp *) pt;
    default:
      return NULL;
    }
}

SgEmptyStoreExp * isSgEmptyStoreExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case EMPTY_STORE:
      return (SgEmptyStoreExp *) pt;
    default:
      return NULL;
    }
}

SgErrLabelExp * isSgErrLabelExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case ERR_LABEL:
      return (SgErrLabelExp *) pt;
    default:
      return NULL;
    }
}

SgEndLabelExp * isSgEndLabelExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case END_LABEL:
      return (SgEndLabelExp *) pt;
    default:
      return NULL;
    }
}

SgDataImpliedDoExp * isSgDataImpliedDoExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DATA_IMPL_DO:
      return (SgDataImpliedDoExp *) pt;
    default:
      return NULL;
    }
}

SgDataEltExp * isSgDataEltExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DATA_ELT:
      return (SgDataEltExp *) pt;
    default:
      return NULL;
    }
}

SgDataSubsExp * isSgDataSubsExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DATA_SUBS:
      return (SgDataSubsExp *) pt;
    default:
      return NULL;
    }
}

SgDataRangeExp * isSgDataRangeExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case DATA_RANGE:
      return (SgDataRangeExp *) pt;
    default:
      return NULL;
    }
}

SgIconExprExp * isSgIconExprExp(SgExpression *pt) /* Fortran M */
{
  if (!pt)
    return NULL;
  switch(NODE_CODE(pt->thellnd))
    {
    case ICON_EXPR:
      return (SgIconExprExp *) pt;
    default:
      return NULL;
    }
}




#if 0
class SgUseStmt: public SgDeclarationStatement{
       // Fortran 90 module usuage statement
       // variant = USE_STMT
   public:
     SgUseStmt(SgSymbol &moduleName, SgExpression &renameList, SgStatement &scope):SgDeclarationStatement(USE_STMT)
          {
            SORRY;
          };
          // renameList must be a list of low-level nodes of variant RENAME_NODE
     ~SgUseStmt(){RemoveFromTableBfnd((void *) this);};

     int isOnly()
          {
            SORRY;
          };
     SgSymbol *moduleName()
          {
            SORRY;
          };
     void setModuleName(SgSymbol &moduleName)
          {
            SORRY;
          };
     int numberOfRenames()
          {
            SORRY;
          };
     SgExpression *renameNode(int i)
          {
            SORRY;
          };
     void  addRename(SgSymbol &localName, SgSymbol &useName)
          {
            SORRY;
          };
     void  addRenameNode(SgExpression &renameNode)
          {
            SORRY;
          };
     void  deleteRenameNode(int i)
          {
            SORRY;
          };
     void deleteTheRenameNode(SgExpression &renameNode)
          {
            SORRY;
          };
};


SgUseStmt * isSgUseStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case USE_STMT:
      return (SgUseStmt *) pt;
    default:
      return NULL;
    }
}



class  SgStmtFunctionStmt: public SgDeclarationStatement{
        // Fortran statement function declaration
        // variant == STMTFN_DECL
     public:        
        SgStmtFunctionStmt(SgSymbol &name, SgExpression &args, SgStatement Body):SgDeclarationStatement(STMTFN_DECL)
          {
            SORRY;
          };
        ~SgStmtFunctionStmt(){RemoveFromTableBfnd((void *) this);};

        SgSymbol *name()
          {
            SORRY;
          };
        void setName(SgSymbol &name)
          {
            SORRY;
          };
        SgType *type()
          {
            SORRY;
          };
        int numberOfParameters()
          {
            SORRY;
          };       // the number of parameters
        SgSymbol *parameter(int i)
          {
            SORRY;
          };     // the i-th parameter
};      

class  SgMiscellStmt: public SgDeclarationStatement{
        // Fortran 90 simple miscellaneous statements
        // variant == CONTAINS_STMT, PRIVATE_STMT, SEQUENCE_STMT
     public:        
        SgMiscellStmt(int variant):SgDeclarationStatement(variant) {}
        ~SgMiscellStmt(){RemoveFromTableBfnd((void *) this);};
};      



SgStmtFunctionStmt * isSgStmtFunctionStmt (SgStatement *pt)
{

  if (!pt)
    return NULL;
  switch(BIF_CODE(pt->thebif))
    {
    case STMTFN_DECL:
      return (SgStmtFunctionStmt *) pt;
    default:
      return NULL;
    }
}
#endif

//
//
// More stuffs for types and symbols
//
//


SgVariableSymb * isSgVariableSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case VARIABLE_NAME:
      return (SgVariableSymb *) pt;
    default:
      return NULL;
    }
}


SgConstantSymb * isSgConstantSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case CONST_NAME :
      return (SgConstantSymb *) pt;
    default:
      return NULL;
    }
}

SgFunctionSymb::SgFunctionSymb(int variant):SgSymbol(variant)
{
  switch (variant) {
  case PROGRAM_NAME:
  case PROCEDURE_NAME:
  case FUNCTION_NAME:
  case MEMBER_FUNC:
    break;
  default:
    Message("SgFunctionSymb variant invalid",0);
  }
}

SgFunctionSymb::SgFunctionSymb(int variant, char *identifier, SgType &t, 
                               SgStatement &scope):SgSymbol(variant,identifier,t,scope)
{
  switch (variant) {
  case PROGRAM_NAME:
  case PROCEDURE_NAME:
  case FUNCTION_NAME:
  case MEMBER_FUNC:
    break;
  default:
    Message("SgFunctionSymb variant invalid",0);
  }
  SYMB_TYPE(thesymb) = t.thetype;
}

SgFunctionSymb * isSgFunctionSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case PROGRAM_NAME:
    case PROCEDURE_NAME:
    case FUNCTION_NAME:
    case MEMBER_FUNC:
      return (SgFunctionSymb *) pt;
    default:
      return NULL;
    }
}
            

SgMemberFuncSymb * isSgMemberFuncSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case MEMBER_FUNC:
      return (SgMemberFuncSymb *) pt;
    default:
      return NULL;
    }
}

SgFieldSymb * isSgFieldSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case ENUM_NAME:
    case FIELD_NAME:
      return (SgFieldSymb *) pt;
    default:
      return NULL;
    }
}


SgClassSymb * isSgClassSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case CLASS_NAME:
    case TECLASS_NAME:
    case UNION_NAME:
    case STRUCT_NAME:
    case COLLECTION_NAME:
      return (SgClassSymb *) pt;
    default:
      return NULL;
    }
}

#if 0
class SgTypeSymb: public SgSymbol{
        // a C typedef.  the type() function returns the base type.
        // variant == TYPE_NAME
      public:
        SgTypeSymb(char *name, SgType &baseType):SgSymbol(TYPE_NAME)
          {
            SORRY;
          };
        SgType &baseType()
          {
            SORRY;
          };
        ~SgTypeSymb(){RemoveFromTableSymb((void *) this);};
};


SgTypeSymb * isSgTypeSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case TYPE_NAME:
      return (SgTypeSymb *) pt;
    default:
      return NULL;
    }
}
#endif

SgLabelSymb * isSgLabelSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case LABEL_NAME:
      return (SgLabelSymb *) pt;
    default:
      return NULL;
    }
}

SgLabelVarSymb * isSgLabelVarSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case LABEL_NAME:
      return (SgLabelVarSymb *) pt;
    default:
      return NULL;
    }
}


SgExternalSymb * isSgExternalSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case ROUTINE_NAME:
      return (SgExternalSymb *) pt;
    default:
      return NULL;
    }
}

SgConstructSymb * isSgConstructSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case CONSTRUCT_NAME:
      return (SgConstructSymb *) pt;
    default:
      return NULL;
    }
}

SgInterfaceSymb * isSgInterfaceSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case INTERFACE_NAME:
      return (SgInterfaceSymb *) pt;
    default:
      return NULL;
    }
}



SgModuleSymb * isSgModuleSymb (SgSymbol *pt)
{

  if (!pt)
    return NULL;
  switch(SYMB_CODE(pt->thesymb))
    {
    case MODULE_NAME:
      return (SgModuleSymb *) pt;
    default:
      return NULL;
    }
}

// ********************* Types *******************************


SgArrayType * isSgArrayType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_ARRAY:
      return (SgArrayType *) pt;
    default:
      return NULL;
    }
}

#if 0
class SgClassType: public SgType{
        // a C struct or Fortran Record, a C++ class, a C Union and a C Enum
        // and a pC++ collection.  note: derived classes are another type.
        // this type is very simple.  it only contains the standard type
        // info from SgType and a pointer to the class declaration stmt
        // and a pointer to the symbol that is the first field in the struct.
        // variant == T_STRUCT, T_ENUM, T_CLASS, T_TECLASS T_ENUM, T_COLLECTION
    public:
        // why is struct_decl needed. No appropriate field found.
        // assumes that first_field has been declared as
        // FIELD_NAME and the remaining fields have been stringed to it.
        SgClassType(int variant, char *name, SgStatement &struct_decl, int num_fields,
                     SgSymbol &first_field):SgType(variant)
          { 
            
            SORRY;
          };
        SgStatement &structureDecl()
          {
            SORRY;
          };
        SgSymbol *firstFieldSymb()
          { return SymbMapping(TYPE_FIRST_FIELD(thetype)); };
        SgSymbol *fieldSymb(int i)
          { return SymbMapping(GetThOfFieldListForType(thetype, i)); }
        int numberOfFields()
          { return lenghtOfFieldListForType(thetype); }
        ~SgClassType(){RemoveFromTableType((void *) this);};
};
        

SgClassType * isSgClassType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_STRUCT:
    case T_ENUM:
    case T_CLASS:
    case T_TECLASS:
    case T_COLLECTION:
      return (SgClassType *) pt;
    default:
      return NULL;
    }
}
#endif

SgPointerType * isSgPointerType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_POINTER:
      return (SgPointerType *) pt;
    default:
      return NULL;
    }
}

SgReferenceType * isSgReferenceType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_REFERENCE:
      return (SgReferenceType *) pt;
    default:
      return NULL;
    }
}


#if 0
class SgFunctionType: public SgType{
        // Function Types have a returned value type
        // variant == T_FUNCTION
    public:
        SgFunctionType(SgType &return_val_type):SgType(T_FUNCTION)
          {
            SORRY;
          };
        SgType &returnedValue()
          {
            SORRY;
          };
        void changeReturnedValue(SgType &rv)
          {
            SORRY;
          };
        ~SgFunctionType(){RemoveFromTableType((void *) this);};
};


SgFunctionType * isSgFunctionType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_FUNCTION:
      return (SgFunctionType *) pt;
    default:
      return NULL;
    }
}

#endif



SgDerivedType * isSgDerivedType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_DERIVED_TYPE:
      return (SgDerivedType *) pt;
    default:
      return NULL;
    }
}

SgDerivedClassType * isSgDerivedClassType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_DERIVED_CLASS:
      return (SgDerivedClassType *) pt;
    default:
      return NULL;
    }
}


SgDescriptType * isSgDescriptType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_DESCRIPT:
      return (SgDescriptType *) pt;
    default:
      return NULL;
    }
}



SgDerivedCollectionType * isSgDerivedCollectionType (SgType *pt)
{

  if (!pt)
    return NULL;
  switch(TYPE_CODE(pt->thetype))
    {
    case T_DERIVED_COLLECTION:
      return (SgDerivedCollectionType *) pt;
    default:
      return NULL;
    }
}

// perhaps this function can use LlndMapping
SgExpression * SgSubscriptExp::lbound()
{
  PTR_LLND ll = NULL;
  SgExpression *pt = NULL;
  ll = NODE_OPERAND0(thellnd);
  if (ll && (NODE_CODE(ll) == DDOT))
    ll = NODE_OPERAND0(ll);
  return LlndMapping(ll);
}

SgExpression * SgSubscriptExp::ubound()
{
  PTR_LLND ll = NULL;

  ll = NODE_OPERAND0(thellnd);
  if (ll && (NODE_CODE(ll) == DDOT))
    ll = NODE_OPERAND1(ll);
  else
    ll = NODE_OPERAND1(thellnd);
  return LlndMapping(ll);
}

SgExpression * SgSubscriptExp::step()
{
  PTR_LLND ll = NULL;
  SgExpression *pt = NULL;
  ll = NODE_OPERAND0(thellnd);
  if (ll && (NODE_CODE(ll) == DDOT))
    ll = NODE_OPERAND1(thellnd);
  else
    ll = makeInt(1);
  return LlndMapping(ll);
}

//
// miscelleanous functions
//

// return a symbol with the name;
// if where is NULL the first symbol, whose name matches, found is returned;
// if where is non NULL the first symbol which scope included  where 
// is returned; as an example getSymbol("foo", GLOBAL_NODE)
// returns only the symbol named foo with scope = GLOBAL_NODE;

SgSymbol *getSymbol(char *name, SgStatement *where)
{
  if (where)
    return SymbMapping(getSymbolWithNameInScope(name, where->thebif));
  else
    return SymbMapping(getSymbolWithNameInScope(name,NULL));
}

SgVarDeclStmt *SgSymbol::makeVarDeclStmt()
{ return 
      isSgVarDeclStmt(BfndMapping(makeDeclStmt(thesymb)));} 

SgVarDeclStmt *SgSymbol::makeVarDeclStmtWithParamList
           (SgExpression &parlist)
{ return 
       isSgVarDeclStmt
         (BfndMapping(makeDeclStmtWPar(thesymb, parlist.thellnd)));} 


//
//
//
// Main file for debug purpose, check the routines in the
// in this file
//
//
//

#ifdef DEBUGLIB
main()
{
  SgProject project("test.proj");
  SgFile    file("simple.f");
  SgValueExp c1(1), c2(2), c3(3), c100(100);
  SgExpression *pt;
  SgVarRefExp  *e1, *e2, *e3, *e4;
  SgStatement *themain, *first, *firstex, *last;
  SgFuncHedrStmt *ptfunc;
  SgSymbol *ptsymb;
  SgSymbol *i1;
  SgSymbol *i2;
  SgSymbol *i3;
  SgSymbol *i4;
  SgSymbol *anarray;
  SgAssignStmt *stmt, *stmt1;
  SgIfStmt *anif;
  SgStatement *anotherif;
  SgWhileStmt *awhile;
  SgForStmt *afor;
  SgReturnStmt *areturn;
  SgCallStmt *afuncall;
  SgArrayType *typearray;
  SgType basetype(T_FLOAT);


  printf("There is %d files in that project\n",project.numberOfFiles());
  first = (file.firstStatement());
  themain = (file.mainProgram());

  ptfunc = new SgFuncHedrStmt("funct1");

  ptsymb = new SgVariableSymb("var1");
  pt = new SgVarRefExp(*ptsymb);
  ptfunc->AddArg(*pt);

  ptsymb = new SgVariableSymb("var2");
  pt = new SgVarRefExp(*ptsymb);
  ptfunc->AddArg(*pt);

  first->insertStmtAfter(*ptfunc);

  // lets add a statement to that function 
  i1 = new SgVariableSymb("i1");
  i1->declareTheSymbol(*ptfunc);
  e1 = new SgVarRefExp(*i1);

  i2 = new SgVariableSymb("i2");
  i2->declareTheSymbol(*ptfunc);
  e2 = new SgVarRefExp(*i2);

  i3 = new SgVariableSymb("i3");
  i3->declareTheSymbol(*ptfunc);
  e3 = new SgVarRefExp(*i3);
  
  i4 = new SgVariableSymb("i4");
  i4->declareTheSymbol(*ptfunc);
  e4 = new SgVarRefExp(*i4);

  firstex = (ptfunc->lastDeclaration());
  stmt = new SgAssignStmt((*e1), (*e2) + ((*e3) + c1) * (*e4));

  stmt1 = new SgAssignStmt(*e2,*e3);

  anif = new SgIfStmt(c1 > c2 , *stmt1, stmt->copy());
  anotherif = &(anif->copy());

  awhile = new SgWhileStmt( (*e4)< c2 , anif->copy());

  afor = new SgForStmt(* i1, c1, c2, c3, awhile->copy());
  areturn = new SgReturnStmt();

  afuncall = new SgCallStmt(*ptfunc->symbol());
  afuncall->addArg(c1.copy());
  afuncall->addArg(c2.copy());
  afuncall->addArg(c3.copy());
  
// let insert what we have created 
  firstex->insertStmtAfter(*anif);
  firstex->insertStmtAfter(stmt->copy());
  firstex->insertStmtAfter(*awhile);
  firstex->insertStmtAfter(*afor);

  last = (ptfunc->lastExecutable());
  last->insertStmtAfter(*areturn);

  
  themain->insertStmtAfter(*anotherif);
  themain->insertStmtAfter(*afuncall);

// Let's try array
   typearray = new SgArrayType(basetype);
   typearray->addRange(c1);
   typearray->addRange(c2);
   typearray->addRange(c3);
   anarray = new SgVariableSymb("Array1",*typearray);
   anarray->declareTheSymbol(*ptfunc);

// make an array expression
   pt =  new SgArrayRefExp(*anarray,*e1,*e2,*e3);
   stmt = new SgAssignStmt((*pt), (*e2) + ((*pt) + c1) * (*pt));
   firstex->insertStmtAfter(*stmt);

// unparse the file
  file.unparsestdout();
  file.saveDepFile("debug.dep");
  
}
#endif


// SgReturnStmt--inlines

SgReturnStmt::SgReturnStmt(SgExpression &returnValue):SgExecutableStatement(RETURN_NODE)
{
  BIF_LL1(thebif) = returnValue.thellnd;
  if (CurrentProject->Fortranlanguage())
    {
      Message("Fortran return does not have expression",0);
      BIF_CODE(thebif) = RETURN_STAT;
    }
}

SgReturnStmt::SgReturnStmt():SgExecutableStatement(RETURN_NODE)
{
  if (CurrentProject->Fortranlanguage())
    BIF_CODE(thebif) = RETURN_STAT;
}

SgExpression * SgObjectListExp::objectRef(int i)  //i'th symbol
{ 
  PTR_LLND pt;
  SgExpression *x;

  pt = getPositionInExprList(NODE_OPERAND0(thellnd),i);
  if (!pt ) return NULL;

  while (NODE_OPERAND0(pt)) {
      if (NODE_CODE(pt) == ARRAY_REF) {
	  break;
      }
      pt = NODE_OPERAND0(pt);
  }

  x = LlndMapping(pt);

  return x;
}
